(** Camelis API: managing a context with query and view histories, and many more... *)

(** Public interface. *)
module type T =
  sig
(** {2 Sub-modules} *)

    module Feat : Logcache.ROOT
(** Logic used for features and descriptors. *)
    module Ext : Logcache.EXT
(** Representation of extents, i.e. sets of objects ids. *)
    module Root : Root.T with module CRoot = Feat
(** Logical cache (see {!module: Root}). *)
    module Src : Source.T
(** Sources (see {!module: Source}). *)
    module Object : Object.T with module CRoot = Feat and module Root = Root and module Src = Src
(** Objects (see {!module: Object}). *)
    module Wll : Well.T
(** Wells (see {!module: Well}). *)

(** {2 LIS databases} *)

    val init : ?log:bool -> string -> unit
	(** [init dbname] initializes a LIS database by creating an empty file [dbname.lis].
	    If [dbname] is not the empty string, try and put a lock on this context.
	    If [log] is true (default is false), then a log file is created as [dbname.log]. *)
    val load : ?readonly:bool -> string -> unit
	(** [load dbname] loads the LIS database from the file [dbname.lis].
            Try and put a lock on this context, unless [readonly] is true (default is false).
	    An exception [Failure] is raised in case there is a version mismatch. *)
    val save : string -> unit
	(** [save dbname] saves all data about the current context in the LIS database file [dbname.lis].
            Fails if there is a lock on this context put by another user. *)

    val filename : unit -> string
	(** [filename ()] returns the filename of the current context, as specified by the last call to [init] or [load]. *)
    val readonly : unit -> bool
	(** [readonly ()] returns whether the current context has readonly access. *)

(** {2 Contexts import/export} *)

    type ctx_option = [`Axioms|`Features|`Sources|`Extrinsic|`Wells|`Rules|`Actions]

    val export : ctx_option list -> string -> unit
	(** [export options ctxname] exports context information in a readable and portable format. *)

    val import : ctx_option list -> string -> unit
	(** [import options ctxname] imports context information from a readable and portable format. *)

(** {2 Export as formal context} *)

    val export_as_binary : Ext.t -> Root.Query.t list -> string -> unit
	(** [export_as_binary oids props filename] writes in [filename] (extension should be .csv) the formal context derived from the current context, where [oids] is the set of objects, and [props] is the set of attributes. *)

(** {2 Logical operations} *)

    val entails : Feat.t -> Feat.t -> bool
	(** [entails f g] returns whether the formula [f] entails the formula [g] in the given logic. *)
    val entails_descr : Object.Descr.t -> Feat.t -> bool
	(** [entails_descr d x] returns whether the description [d] entails the feature [x]. *)
    val update : Object.Descr.t -> Root.Update.t -> Object.Descr.t * Root.Update.t
	(** [update d u] returns an update of description [d] by update [u], and the effective update. *)

(** {2 Axioms on terms} *)

    val add_axiom : ?src:Src.t -> Feat.t -> Feat.t -> unit
	(** [axiom f g] ensures that [f] entails [g] (when possible) by adding necessary and/or sufficient conditions on terms in [f] and [g]. *)

(** {2 Objects} *)

    type obj_update = Intr of Root.Update.t | Extr of Root.Update.t | Rem

    val valid_obj : Logcache.oid -> bool
	(** [valid_obj oid] returns whether [oid] is a valid object id. *)
    val origin_obj : Logcache.oid -> Object.origin
        (** [origin_obj oid] returns the origin (either source/ref or title/action_args) of object [oid]. *)
    val descr_obj : Logcache.oid -> Object.Descr.t
	(** [descr_obj oid] returns the logical description of the object [oid]. *)
    val make_obj : Source.preview -> Source.action_arg list -> Object.Descr.t -> Root.Update.t -> unit
	(** [make_obj pre args d e] creates a new object (that has no source) with preview [pre],
	   action arguments [args], and the initial description [d]; and applies to it the logical update [u]. *)
    val update_obj : Logcache.oid -> obj_update -> unit
	(** [update_obj oid up] applies the update [up] to object [oid]. *)

(** {2 Monitor} *)

    val home_monitor : unit -> Root.Query.t
	(** [home_monitor ()]returns the current state of the monitor. *)
    val update_monitor : unit -> unit
	(** [update_monitor ()] updates the context by generating timestamp objects in the current monitor state. *)
    val move_monitor : Root.Query.t -> unit
	(** [move_monitor q] sets [q] as the new monitor state. *)

(** {2 Sources} *)

    val home_source : Source.url -> Object.Descr.t (* raise Not_found *)
	(** [home_source url] returns the home of objects imported from the source [url]. *)
    val init_source : ?ss:Src.t -> Source.url -> Source.src_option list -> Object.Descr.t -> unit (* raise Failure *)
	(** [init_source url options e] sets up an object source at [url] by making all generated objects entail [e] (source home).
	    If the source already exists, it is updated according to new [options]. *)
    val update_source : Source.url -> unit (* raise Failure *)
	(** [update_source url] synchronizes the context with the source at [url]. *)
    val move_source : Source.url -> Root.Update.t -> unit (* raise Failure *)
	(** [move_source url u] applies the update [u] on the home of the source at [url]. *)
    val final_source : Source.url -> unit (* raise Failure *)
	(** [final_source url] removes the source at [url] from the context. This removes whatever has been added to the context by this source (objects, axioms, ...). *)
    val relocate_source : Source.url -> Source.url -> unit
	(** [relocate_source url new_url] relocates the given source at [url] at location [new_url]. *)
    val action_args : Logcache.oid -> Source.action_arg list
	(** [action_args oid] returns action arguments relevant to [oid]. *)
    val fold_sources_all : (Source.url -> 'a -> 'a) -> 'a -> 'a
	(** [get_sources_all f init] folds all existing sources in the context. *)
    val fold_sources_under : (Source.url -> 'a -> 'a) -> 'a -> Ext.t -> 'a
	(** [get_sources_under f init oids] folds all sources, whose all objects are in [oids]. *)

(** {2 Wells} *)

(** Wells provide a facility for exporting data from the context. This
data is the extent of a given query. Whenever this extent changes due
to updates, the exported data is updated as well. *)

    val init_well : Well.url -> Root.Query.t -> unit (* raise Failure *)
	(** [init_well url q] sets up the well at [url] on the source query [q]. *)
    val update_well : Well.url -> unit (* raise Failure *)
	(** [update_well url] synchronizes any data exported by the well at [url] with the context. *)
    val move_well : Well.url -> Root.Query.t -> unit (* raise Failure *)
	(** [move_well url q] sets [q] as the query of the well at [url]. *)
    val final_well : Well.url -> unit (* raise Failure *)
	(** [final_well url] removes the well at [url]. This also deletes any data exported by the well. *)
    val relocate_well : Well.url -> Well.url -> unit
	(** [relocate_well url new_url] relocates the given well at [url] at location [new_url]. *)
    val get_wells_all : unit -> (Root.Query.t * Well.url) list
	(** [get_wells_all ()] returns the list of all existing wells in the context. *)

(** {2 Update rules} *)

    type rule (** type of automatic update rules. *)

    val query_rule : rule -> Root.Query.t
	(** [query_rule rule] returns the query used as a the hypothesis part of the rule. *)
    val update_rule : rule -> obj_update
	(** [update_rule rule] returns the object update used as the conclusion part of the rule. *)
    val init_rule : Root.Query.t -> obj_update -> rule
	(** [init_rule q up] returns a rule [q => up] that will apply the logical update [up] to any object that enters the extent of [q]. *)
    val final_rule : rule -> unit
	(** [final_rule rule] removes the rule. *)

    val get_rules : unit -> rule list
	(** [get_rules ()] returns the list of all rules. *)

(** {2 Actions} *)

    type action = {
	mutable action_name : string; (** the name describing the action *)
	mutable action_cmd : string; (** the command line to be executed, parametized with variables *)
	mutable action_query : Root.Query.t; (** the query define the scope of the action *)
      }

    val get_actions : unit -> action list
	(** [get_actions ()] returns the list of all available actions. *)
	
    val add_action : action -> unit
	(** [add_action ac] registers the given action [ac] as a new action. *)

    val remove_action : action -> unit
	(** [remove_action ac_name] remove the action associated to the given name. *)

    val apply_action : action -> Source.action_arg list -> (string * (unit -> unit)) list
	(** [apply_action ac args ()] applies the action by replacing in the [ac.action_name] and in the [ac.action_cmd] the variables [$(X)]
	   by the values [arg.role] and [arg.location] associated to [X = arg.mime] in [args]. [X] is a MIME type.
	   Returns a list of pairs [(name,fun)] representing applicable concrete actions. *)

    val object_actions : Logcache.oid -> (string * (unit -> unit)) list
	(** [object_actions oid] returns all available actions on [oid] as pairs [(name,code)]. *)
    val well_actions : string -> (string * (unit -> unit)) list
	(** [well_actions url] returns all available actions on well [url] as pairs [(name,code)]. *)
    val temp_well_actions : string -> Well.contents -> (string * (unit -> unit)) list
	(** [temp_well_actions url cs] returns all available actions on temporary well [url] with contents [cs] as pairs [(name,code)]. *)

(** {2 Background updating thread} *)

    val nb_updates : unit -> int
	(** [nb_updates ()] returns the number of queueing updates. *)
    val wait_updates : unit -> unit
	(** [wait_updates ()] suspends the current thread until all updates are performed. *)

(** {2 The query history} *)

    val wq : unit -> Root.Query.t
	(** [wq ()] returns the working query. *)
    val wq_back : unit -> unit
	(** [wq_back ()] pops the query history. *)
    val wq_incr : (Root.Query.t -> Root.Query.t) -> unit
	(** [wq_incr f] pushes in the query history the new query resulting from the application of [f] to the working query. *)

(** {2 Features} *)

    val add_feature : Feat.t -> unit
	(** [add_feature f] adds the feature [f] to the logical cache, so that [f] can appear in navigation. *)
    val del_feature : Feat.t -> unit
	(** [del_feature f] hides the feature [f] so that it does not appear in navigation. *)
    val replace_feature : Feat.t -> Feat.t -> unit
	(** [replace_feature f f'] replaces all occurences of [f] by [f']. Only partially implemented. *)      

(** {2 Navigation and Knowledge Extraction} *)

(* get navigation links, local objects and number of objects in the working query
   and also views if the argument is true *)
    type propr = { (** Properties: features decorated by additional information. *)
	propr_card : int;  (** The coverage of a property in the context. *)
	propr_supp : int;  (** The support or coverage of a property in the query. *)
	propr_ext : Ext.t; (** The extent of the property. *)
	propr_logs : Root.Query.t list;  (** The set of equivalent features characterizing a property. *)
	propr_sups : (Ext.t * Root.Query.t) list;  (** Other formulas more general than a property. *)
      }

    type view = {
        nbobj : int;  (** Total number of objects. *)
	nbsel : int;  (** Number of objects in the current extent. *)
        extent : Ext.t;  (** Current extent. *)
        locals : Ext.t;  (** local objects. *)
        proprs : propr list;  (** increments. *)
      }
	  
    type links_option = [`NoObjs|`NoViews|`Implications|`MaxOnly|`MostSpec|`SuppMin of int|`Neg|`Pattern of string|`LexicalOrder|`ValOnly]

    val link : Root.Query.t -> propr
       (* returns the count of a individual link *)

    val links : links_option list -> Root.Query.t -> view (* propr list * Ext.t * int *)

	(** [links opts] returns a set of links (properties), a set of
	   local objects, and the support for the working query and the
	   working view. The working query is used to evaluate the
	   relevance of features as links, and the working view is used
	   to restrict the domain of features to be considered.

	   Various options are available:
	   - [`NoViews]: all links must strictly restrict the working query;
	   - [`Implications]: show contextual implications between links;
	   - [`MaxOnly]: show only contextually maximal links (conceptual vs. logical navigation);
	   - [`Mostspec]: for a given support, prefer more specific features;
	   - [`SuppMin n]: show only links, whose support is greater or equal to [n];
	   - [`Neg]: show negations of features as links;
	   - [`Pattern regexp]: restrict the domain of features to those whose syntactical form
	     matches the given regular expression [regexp]. *)

    val extent : unit  -> Ext.t
	(** [extent ()] returns the extent of the working query. *)
    val intent : Ext.t -> Feat.t list
	(** get the intent from some object set *)

(* TO BE REMADE *)
(* *******
(* get all concepts of the sub-context defined by working query and view *)
    type concept = Ext.t * Feat.t list * Ext.t
    type hyp = int * float * Feat.t list * Feat.t list list
    type cc = (bool * Feat.t) list  (* Conjunctive Clause *)
    type lr = int * Root.Query.t * cc  (* Logical Rule *)
    type confusion = {
      true_pos : Ext.t;
      false_neg : Ext.t;
      false_pos : Ext.t;
      true_neg : Ext.t}

    val input_rules : string -> lr list
    val output_rules : string -> lr list -> unit
    val rule_covered : Logcache.oid -> lr -> bool
    val eval_rule : lr -> Ext.t -> confusion

    val intent : Ext.t -> Ext.t * Feat.t list * cc list

    val concepts : int -> concept list
    val assocs : int -> Logcache.oid -> concept list
    val sufs : int -> float -> Ext.t -> hyp list
    val nec : Ext.t -> hyp
    val ind : int -> float -> Logcache.oid -> Ext.t -> hyp list
    val inds : int -> float -> Logcache.oid -> Root.Query.t -> (Feat.t * hyp list) list
**** *)

(** {2 Parsing} *)

(* parse a formula of Log intended to be a query atom *)
    val parse : Syntax.t_stream -> Feat.t
(* parse a formula of Feat encapsulated in a O modality *)
    val parse_descr : Syntax.t_stream -> Object.Descr.t
(* parse an update *)
    val parse_update : Syntax.t_stream -> Root.Update.t
(* parse a query *)
    val parse_query : Syntax.t_stream -> Root.Query.t
(* parse an query increment *)
    val parse_incr_query : Syntax.t_stream -> (Root.Query.t -> Root.Query.t)

(** {2 Printing} *)

(* print a formula of Log *)
    val print : Feat.t -> Syntax.t_list
(* print a description (without brackets) *)
    val print_descr : Object.Descr.t -> Syntax.t_list
(* print an update *)
    val print_update : Root.Update.t -> Syntax.t_list
(* print a query *)
    val print_query : Root.Query.t -> Syntax.t_list


(** {2 Converting between logical types} *)

   val descr_of_query : Root.Query.t -> Object.Descr.t
	(** Extract positive features from a query. *)
   val update_of_query : Root.Query.t -> Root.Update.t
        (** Convert a query to an update. *)
   val features_of_query : Root.Query.t -> Root.Query.t list * Root.Query.t list
        (** Split a query in its list of features (i.e. atomic parts): positives and negatives. *)

(** {2 Debugging utilities} *)

(* get some statistics *)
    type stat = {
	mutable nb_obj : int; (** number of objects *)
	mutable nb_node : int; (** number of nodes in the logical cache *)
	mutable nb_arc : int; (** number of edges in the logical cache *)
	mutable sum_ext : int;  (** sum of all node extent size *)
	mutable sum_ext_words : int;  (** sum of all node extent size in memory words *)
	mutable nb_entails : int}  (** number of entailment tests *)

    val get_stat : unit -> stat

(* print the nodes *)
    val print_nodes : unit -> unit
    val print_node : Feat.t -> unit

  end

module Make (CFeat : Logcache.T) (Src : Source.T) (Wll : Well.T) : T =
  struct
    module ObjectParam = struct end
    module Object = Object.Make(ObjectParam)(CFeat)(Src)
    module Src = Object.Src
    module Root = Object.Root
    module Ext = Logcache.Ext
    module Feat = Root.CRoot
    module Query = Root.Query

    module Wll = Wll

    let mutex = Threads_common.mutex

(* current context files *)

    type ctx = {
	mutable filename : string;
	mutable readonly : bool; (* if false, file 'filename.lck' exists *)
	mutable log : out_channel option;
      }

    let ctx = {
      filename = "";
      readonly = true;
      log = None;
    }

    let filename () = ctx.filename

    let readonly () = ctx.readonly

(* lock and log management *)

    module Lock =
      struct
	let lockname f = f ^ ".lck"

	let logname f = f ^ ".log"

	let check () = ()

	let check_file f =
	  if (if ctx.filename = f then ctx.readonly else Sys.file_exists (lockname f))
	  then failwith "You have no write access on this context.\nConsider saving a copy in another location."

	let set f r l =
	  if not ctx.readonly then Sys.remove (lockname ctx.filename);
	  if f = ""
	  then begin
	    ctx.filename <- "";
	    ctx.readonly <- true;
	    ctx.log <- None end
	  else begin
	    let lockname = lockname f in
	    let readonly = r || Sys.file_exists lockname in
	    if not readonly then begin let ch = open_out lockname in close_out ch end;
	    ctx.filename <- f;
	    ctx.readonly <- readonly;
	    ctx.log <-
	      if l
	      then Some (open_out_gen [Open_wronly; Open_append; Open_creat; Open_text] 0o660 (logname f))
	      else None
	  end

      end

(* working query *)

    let wq_histo : Query.t Stack.t = Stack.create ()

    let wq () = mutex (fun () ->
      try Stack.top wq_histo with Stack.Empty -> Query.top)

    let wq_back () = mutex (fun () ->
      try ignore (Stack.pop wq_histo) with Stack.Empty -> ())

    let wq_incr incr = mutex (fun () ->
      Stack.push (incr (wq ())) wq_histo) 

(* working view *)
(* suspended
    let wv_histo : Node.view Stack.t = Stack.create ()

    let get_wv () = mutex (fun () ->
      try Stack.top wv_histo with Stack.Empty -> Node.get_view Node.All)

    let wv () = mutex (fun () ->
      try Node.view_prop (Stack.top wv_histo) with Stack.Empty -> Node.All)

    let wv_back () = mutex (fun () ->
      try ignore (Stack.pop wv_histo) with Stack.Empty -> ())

    let wv_incr incr = mutex (fun () ->
      Stack.push (Node.get_view (incr (wv ()))) wv_histo)
*)

(* logical operations *)
(* ------------------ *)

    let entails f g = Feat.entails f g

    let entails_descr d x = Object.Descr.entails d x

    let update d u = Object.Descr.update d u


(* contextual operations *)
(* --------------------- *)

    open Source

(*** updates management ***)


(* period between 2 global source updates in seconds *)
    let update_period : int ref = ref 1000000

(* set of updates, ordered according to definition of Source.update *)
    module SetUpdates = Set.Make(struct type t = Source.update let compare = compare end)
    let set_updates : SetUpdates.t ref = ref SetUpdates.empty

(* get the number of remaining updates *)
    let nb_updates () = mutex (fun () ->
      SetUpdates.cardinal !set_updates)

    let wait_updates () = 
      Mutex.unlock Threads_common.m;
      while nb_updates () > 0 do
        Thread.delay 60.
      done;
      Mutex.lock Threads_common.m

    let push_updates updates = mutex (fun () ->
      Lock.check ();
      List.iter (fun update -> set_updates := SetUpdates.add update !set_updates) updates)


(*** features management ***)

    let add_feature f = mutex (fun () ->
      Lock.check ();
      !Root.cache # show f)
(*
      let n = Node.of_formula f in
      Node.add_features n)
*)

    let del_feature f = mutex (fun () ->
      Lock.check ();
      try !Root.cache # hide (!Root.cache # find f)
      with Not_found -> ())
(*
      let n = Node.of_formula f in
      Node.del_features n)
*)

(*
    let del_node f = mutex (fun () ->
      Node.free (Node.of_formula f))
*)


(*** axioms management ***)

    let l_axiom : (Feat.t * Feat.t * Src.t option) list ref = ref []

    let add_axiom ?src f g = mutex (fun () ->
      Lock.check ();
      if not (entails g f) then
	if entails f g then begin
	  let l = List.filter (fun (f',g',src') -> not (src' = None & entails f f' & entails g' g)) !l_axiom in
	  !Root.cache # axiom f (Some g);
	  l_axiom := (f,g,src)::l end
	else begin
	  let l = List.filter (fun (f',g',src') -> not (src' = None & entails f' f & entails g g')) !l_axiom in
	  !Root.cache # axiom f (Some g);
	  l_axiom := (f,g,src)::l end;
      push_updates [Show (Feat.print f); Show (Feat.print g)]) (* make [f] and [g] visible, but later on *)

    let replace_in_axioms f f' =
      let check_axiom a b = add_axiom a b; if not (entails a b) then failwith "Invalid replacement" in
      (* assumes f' is not already used *)
      let parents_of_f =
	Common.mapfilter
	  (fun (a,b,src_opt) ->
	    if src_opt = None && (a = f || a = f')
	    then Some b
	    else None)
	  !l_axiom in
      let children_of_f =
	Common.mapfilter
	  (fun (a,b,src_opt) ->
	    if src_opt = None && (b = f || b = f')
	    then Some a
	    else None)
	  !l_axiom in
      let new_parents_of_childrens =
	List.map
	  (fun c ->
	    let parents =
	      Common.mapfilter
		(fun (a,b,src_opt) ->
		  if src_opt = None && a = c
		  then Some (if b = f then f' else b)
		  else None)
		!l_axiom in
	    (c, parents))
	  children_of_f in
      ( match new_parents_of_childrens with
      | (c,_)::_ -> check_axiom c f'
      | _ ->
	  match parents_of_f with
	  | p::_ -> check_axiom f' p
	  | _ -> ());
      List.iter (fun p -> add_axiom f' p) parents_of_f;
      List.iter
	(fun (c,pc) ->
	  add_axiom c (Feat.top ());
	  List.iter (fun p -> add_axiom c p) pc)
	new_parents_of_childrens

(*** monitor management ***)

    type monitor = {
	mutable m_state : Root.Query.t;
	mutable m_last_move : float;
	mutable m_last_update : float;
      }

    let monitor : monitor =
      let t = Unix.time () in
      { m_state = Root.Query.top;
	m_last_move = t;
	m_last_update = t;
      }

    let init_monitor () =
      let t = Unix.time () in
      monitor.m_state <- Root.Query.top;
      monitor.m_last_move <- t;
      monitor.m_last_update <- t

    let load_monitor m =
      monitor.m_state <- m.m_state;
      monitor.m_last_move <- m.m_last_move;
      monitor.m_last_update <- m.m_last_update

    let monitor_period = 6. *. 60. (* tenths of hours *)

    type kind = State | Event
    let months = [|"jan"; "feb"; "mar"; "apr"; "may"; "jun"; "jul"; "aug"; "sep"; "oct"; "nov"; "dec"|]
    let get_pre_args_descr kind =
      let tm = Unix.localtime (match kind with State -> monitor.m_last_update | Event -> monitor.m_last_move) in
      let year = 1900 + tm.Unix.tm_year in
      let month = months.(tm.Unix.tm_mon) in
      let day = tm.Unix.tm_mday in
      let hour = tm.Unix.tm_hour in
      let min = tm.Unix.tm_min in
      let rank = int_of_float ((monitor.m_last_update -. monitor.m_last_move) /. monitor_period) in
      let d = Object.Descr.print (Object.Descr.of_query monitor.m_state) in
      let title = Printf.sprintf "%d %s %d  %d:%02d (%d)" day month year hour min rank in
      let title = if kind = Event then Syntax.stringizer d ^ "\n" ^ title else title in
      let d = Token.Ident "date" :: Token.PP_tilda :: Token.Equal :: Token.PP_tilda :: Token.Nat day :: Token.PP_tilda :: Token.Ident month :: Token.PP_tilda :: Token.Nat year :: Token.Comma :: d in
      let d = Token.Ident "time" :: Token.PP_tilda :: Token.Equal :: Token.PP_tilda :: Token.Nat hour :: Token.Colon :: Token.Nat min :: (* Token.Colon :: Token.Nat sec :: *) Token.Comma :: d in
      let d = Token.Ident (match kind with State -> "rank" | Event -> "duration") :: Token.PP_tilda :: Token.Equal :: Token.PP_tilda ::  Token.Nat rank :: Token.Comma :: d in
      let d = Token.Term (match kind with State -> "State" | Event -> "Event") :: Token.Comma :: d in
      Source.String title, [], d


    let home_monitor () = monitor.m_state

    let update_monitor () = mutex (fun () ->
      let now = Unix.time () in
      if monitor.m_state = Root.Query.top
      then monitor.m_last_update <- now
      else begin
	Lock.check ();
	while monitor.m_last_update < now do
	  let pre, args, d = get_pre_args_descr State in
	  push_updates [MakeExtr (pre, args, d)];
	  monitor.m_last_update <- monitor.m_last_update +. monitor_period
	done
      end)

    let move_monitor q = mutex (fun () ->
      update_monitor ();
      if monitor.m_state <> Root.Query.top then begin
	let pre, args, d = get_pre_args_descr Event in
	push_updates [MakeExtr (pre,args,d)]
      end;
      monitor.m_state <- q;
      monitor.m_last_move <- monitor.m_last_update;
      if q <> Root.Query.top then begin
	let pre, args, d = get_pre_args_descr State in
	push_updates [MakeExtr (pre, args, d)];
	monitor.m_last_update <- monitor.m_last_update +. monitor_period
      end)

(*** sources managment ***)

    type src = {
	s : Src.t;
	s_options : Source.src_option list;
	s_home : Object.Descr.t;
	s_s : Src.t option
      }

    let l_src : (Source.url, src) Hashtbl.t = Hashtbl.create 100
	(* hashtables of sources (src, options, home, s_opt) indexed by urls, where s_opt is the source of src, if any. *)
	(* and options are the source options *)
	(* and home is an update to be applied on created objects *)

    let get_src : Source.url -> src (* raise Not_found *) =
      fun url -> Hashtbl.find l_src url

    let home_source url = mutex (fun () ->
      (Hashtbl.find l_src url).s_home)

    let update_source_options url options = mutex (fun () ->
      Lock.check ();
      try
	let src = Hashtbl.find l_src url in
	Hashtbl.replace l_src url {src with s_options = options};
	push_updates (Src.update src.s options)
      with Not_found -> failwith ("Context.update_source_options: Unknown source: " ^ url))

    let update_source url = mutex (fun () ->
      Lock.check ();
      try
	let src = Hashtbl.find l_src url in
	push_updates (Src.update src.s src.s_options)
      with Not_found -> failwith ("Context.update_source: Unknown source: " ^ url))

    let init_source ?ss url options home = mutex (fun () ->
      Lock.check ();
      if not (Hashtbl.mem l_src url)
      then begin
	let s = Src.make url in
	Hashtbl.add l_src url {s = s; s_options = options; s_home = home; s_s = ss};
	push_updates (Src.init s options) end
      else update_source_options url options (*failwith ("Already known source: " ^ url)*) )

    let move_source url u = mutex (fun () ->
      Lock.check ();
      try
        let src = Hashtbl.find l_src url in
	let h', u' = Object.Descr.update src.s_home u in
	Hashtbl.replace l_src url {src with s_home = h'};
	(* retroactive effect on existing objects *)
	push_updates (List.map (fun r -> MoveExtrSrc (url, r, Root.Update.print u)) (Src.parts src.s))
      with Not_found -> failwith ("Context.move_source: Unknown source: " ^ url))

    let final_source url = mutex (fun () ->
      Lock.check ();
      try
	let src = Hashtbl.find l_src url in
	let updates = Src.final src.s in
	push_updates
	  (List.map
	     (function
	       | RemoveSrc (url,ref) -> Remove (Object.get_oid (src.s,ref))
	       | u -> u)
	     updates);
	Hashtbl.remove l_src url
      with Not_found -> failwith ("Context.final_source: Unknown source: " ^ url))

    let relocate_source url new_url = mutex (fun () ->
      Lock.check ();
      if new_url <> url then begin
	begin try final_source new_url
	with _ -> () end;
	try
	  let src = Hashtbl.find l_src url in
	  push_updates (Src.move src.s new_url);
	  Hashtbl.remove l_src url;
	  Hashtbl.add l_src (Src.url src.s) src;
	  push_updates (Src.update src.s src.s_options);
	with Not_found -> failwith ("Context.relocate_source: Unknown source: " ^ url) end)

    let action_args oid = mutex (fun () ->
      match Object.get_origin oid with
      | Object.External (s,r) -> Src.action_args s r
      | Object.Internal (pre,args) -> args)

    let fold_sources_all f init = mutex (fun () ->
      Hashtbl.fold (fun url _ res -> f url res) l_src init)

(* deprecated
    let get_sources_over oids = mutex (fun () ->
      Ext.fold
	(fun res oid ->
	  try
	    match Object.get_origin oid with
            | Object.External (s,_) ->
		let url = Src.url s in
		if List.mem url res
		then res
		else url::res
            | Object.Internal _ -> res
	  with Not_found -> res
	) [] oids)
*)

    let fold_sources_under f init oids = mutex (fun () ->
      Hashtbl.fold
	(fun url src res ->
	  let ext_s = List.fold_left (fun res r -> Ext.add (Object.get_oid (src.s,r)) res) Ext.empty (Src.parts src.s) in
	  if not (Ext.is_empty ext_s) & Ext.subset ext_s oids
	  then f url res
	  else res
	) l_src init)

(*** objects managment ***)

    type obj_update = Intr of Root.Update.t | Extr of Root.Update.t | Rem

    let valid_obj oid = mutex (fun () ->
      !Root.cache # valid oid)

    let origin_obj oid = mutex (fun () ->
      Object.get_origin oid)

    let descr_obj oid = mutex (fun () ->
      Object.get_descr oid)

    let make_obj : Source.preview -> Source.action_arg list -> Object.Descr.t -> Root.Update.t -> unit =
      fun pre args d u -> mutex (fun () -> Common.prof "make_obj" (fun () ->
	Lock.check ();
	let d', _ = Object.Descr.update d u in
	push_updates [MakeExtr (pre, args, Object.Descr.print d')]))

    let move_intr_obj : Logcache.oid -> Root.Update.t -> unit =
      fun oid u -> mutex (fun () ->
	Lock.check ();
        let updates = Object.move_intr oid u in
	push_updates updates)

    let move_obj : Logcache.oid -> Root.Update.t -> unit =
      fun oid u -> mutex (fun () ->
	Lock.check ();
        let updates = Object.move oid u in
	push_updates updates)

    let remove_obj : Logcache.oid -> unit =
      fun oid -> mutex (fun () ->
	Lock.check ();
        match Object.get_origin oid with
	| Object.Internal _ ->
            let updates = Object.remove oid in
            push_updates updates
        | Object.External _ ->
	    raise (Failure "This object must be remove through its source."))

    let update_obj : Logcache.oid -> obj_update -> unit =
      fun oid -> function
	| Intr u -> move_intr_obj oid u
	| Extr u -> move_obj oid u
	| Rem -> remove_obj oid

(*** wells management ***)

    type wll = (Wll.t * Root.Rules.elt)

    let l_wll : (Well.url, wll) Hashtbl.t = Hashtbl.create 100

(*
    let get_wll : Wll.t -> Wll.t * Root.Rules.elt (* raise Not_found *) =  (* argument wd is understand as a well description *)
      fun wd -> List.find (fun (w,_) -> Wll.eq wd w) !l_wll
*)

(*
    let string_of_wll : Wll.t -> string =
      fun w -> Syntax.stringizer (Wll.print w)
*)

    let contents_of_rule : Root.Rules.elt -> Well.contents =
      fun rule ->
	Ext.map action_args rule.Root.Rules.propext

(*
    let url_well wd = mutex (fun () ->
      Wll.url wd)

    let make_well url = mutex (fun () ->
      Wll.make url)
*)

    let init_well url q = mutex (fun () ->
      Lock.check ();
      if not (Hashtbl.mem l_wll url)
      then begin
	let up = UpdateWll (url) in
	let rule = Root.Rules.add q (Some up) (Some up) in
	let cs = contents_of_rule rule in
	let wll = Wll.make url in
	Wll.init wll cs;	
	Hashtbl.add l_wll url (wll,rule) end
      else failwith ("Already known well: " ^ url))

    let update_well url = mutex (fun () ->
      Lock.check ();
      try
	let w,rule = Hashtbl.find l_wll url in
	let cs = contents_of_rule rule in
	Wll.update w cs
      with Not_found -> failwith ("Unknown well: " ^ url))

    let move_well url q' = mutex (fun () ->
      Lock.check ();
      try
        let w,rule = Hashtbl.find l_wll url in
	Root.Rules.remove rule;
	let rule' = Root.Rules.add q' rule.Root.Rules.addup rule.Root.Rules.remup in
	Hashtbl.remove l_wll url;
	Hashtbl.add l_wll url (w,rule');
	let cs' = contents_of_rule rule' in
	Wll.update w cs'
      with Not_found -> failwith ("Unknown well: " ^ url))

    let final_well url = mutex (fun () ->
      Lock.check ();
      try
	let w,rule = Hashtbl.find l_wll url in
	Root.Rules.remove rule;
	Hashtbl.remove l_wll url;
	let cs = contents_of_rule rule in
	Wll.final w cs
      with Not_found -> failwith ("Unknown well: " ^ url))

    let relocate_well url new_url = mutex (fun () ->
      Lock.check ();
      try
	let w,rule = Hashtbl.find l_wll url in
	Wll.move w new_url;
	Hashtbl.remove l_wll url;
	Hashtbl.add l_wll (Wll.url w) (w, rule)
      with Not_found -> failwith ("Unknown well: " ^ url))

    let get_wells_all () = mutex (fun () ->
      Hashtbl.fold (fun url (w,rule) res -> (rule.Root.Rules.prop,url)::res) l_wll [])


(*** rules management ***)

    type rule = {
	rule_query : Query.t;         (* premisse of the rule *)
	rule_update : Source.update;  (* conclusion of the rule as an update *)
      }

    let query_rule rule = mutex (fun () -> rule.rule_query)

    let update_rule rule = mutex (fun () ->
      Lock.check ();
      match rule.rule_update with
      | MoveIntr (_,u) -> Intr (Root.Update.parse (Syntax.of_list u))
      | MoveExtr (_,u) -> Extr (Root.Update.parse (Syntax.of_list u))
      | Remove (_) -> Rem
      | _ -> assert false)

    let init_rule : Query.t -> obj_update -> rule =
      fun p up -> mutex (fun () ->
	Lock.check ();
	match up with
	| Intr c ->
	    let addup = MoveIntr (0, Root.Update.print c) in
            let rule = Root.Rules.add p (Some addup) None in
            push_updates (Ext.map (fun oid -> Source.oid_update oid addup) rule.Root.Rules.propext);
	    { rule_query = p; rule_update = addup}
	| Extr c -> 
	    let addup = MoveExtr (0, Root.Update.print c) in
            let rule = Root.Rules.add p (Some addup) None in
            push_updates (Ext.map (fun oid -> Source.oid_update oid addup) rule.Root.Rules.propext);
	    { rule_query = p; rule_update = addup}
	| Rem ->
	    let addup = Remove (0) in
	    let rule = Root.Rules.add p (Some addup) None in
            push_updates (Ext.map (fun oid -> Source.oid_update oid addup) rule.Root.Rules.propext);
	    { rule_query = p; rule_update = addup})

    let final_rule : rule -> unit =
      fun rule -> mutex (fun () ->
	Lock.check ();
	let r_opt =
	  Root.Rules.fold
	    (fun res r ->
	      if r.Root.Rules.prop == rule.rule_query && r.Root.Rules.addup = Some rule.rule_update
	      then Some r
	      else res)
	    None in
	match r_opt with
	| Some r -> Root.Rules.remove r
	| None -> print_endline "Context.final_rule: rule not found")

    let get_rules () = mutex (fun () ->
      Root.Rules.fold
	(fun res rule ->
	  match rule.Root.Rules.addup with
	  | Some (MoveIntr _ as addup) -> { rule_query = rule.Root.Rules.prop; rule_update = addup} :: res
	  | Some (MoveExtr _ as addup) -> { rule_query = rule.Root.Rules.prop; rule_update = addup} :: res
	  | Some (Remove _ as addup) -> { rule_query = rule.Root.Rules.prop; rule_update = addup} :: res
	  | _ -> res)
	[])

(*** actions ***)

    type action = {
	mutable action_name : string; (* the name describing the action, parameterized with mime variables *)
	  (* mime variables are in the form $(mime) *)
	mutable action_cmd : string; (* the command line to be executed, parametized with mime variables (a superset of those in name) *)
          (* mime variables are in the form $(mime) or $(mime:type), where type indicates a transformation to be applied on the value *)
	  (* for instance ":win" will transform a Cygwin path into Windows path *)
	mutable action_query : Query.t; (* the query defines the scope of the action *)
      }

    let l_actions : action list ref = ref []

    let get_actions () = mutex (fun () ->
      !l_actions)

    let add_action ac = mutex (fun () ->
      Lock.check ();
      if not (List.mem ac !l_actions) (* no duplicate *)
      then l_actions := ac::!l_actions)

    let remove_action ac = mutex (fun () ->
      Lock.check ();
      l_actions := List.filter (fun ac' -> ac' != ac) !l_actions)

    let rec apply_action_aux (name,cmd) args mimes : (string * string) list =
      match mimes with
      | [] -> [(name,cmd)]
      | (m,t)::ms ->
	  let vs = (* values to be substituted to parameter m *)
	    Common.mapfilter
	      (fun arg ->
		if Source.mime_match m arg.Source.mime
		then Some (arg.Source.role, arg.Source.location)
		else None)
	      args in
	  List.fold_left
	    (fun res (r,l) ->
	      let l' = (* l modified according to t *)
		match t with
		| ":win" -> Common.cyg2win l
		| _ -> l in
	      let new_name = Str.global_substitute (Str.regexp_string ("$(" ^ m ^ ")")) (fun s -> r) name in
	      let new_cmd = Str.global_substitute (Str.regexp_string ("$(" ^ m ^ t ^ ")")) (fun s -> l') cmd in
	      apply_action_aux (new_name,new_cmd) args ms @ res)
	    []
	    vs

    let apply_action ac args =
      (* looking for mime parameters in action command *)
      (* the action name should contain a subset of them *)
      let re = Str.regexp "[^$]*\\$(\\([^:)]*\\)\\(:[a-z]+\\)?)" in
      let cmd = ac.action_cmd in
      let mimes, _ =
	Common.fold_while
	  (fun (res,pos) ->
	    if Str.string_match re cmd pos
	    then
	      let m = Str.matched_group 1 cmd in
	      let t = try Str.matched_group 2 cmd with Not_found -> "" in
	      Some ((if List.mem_assoc m res then res else (m,t)::res), Str.match_end ())
	    else None)
	  ([], 0) in
      let name_cmd_l = apply_action_aux (ac.action_name, cmd) args mimes in
      List.map
	(fun (name,cmd) -> (name, (fun () -> print_endline cmd; ignore (Sys.command cmd))))
	name_cmd_l


    let object_actions oid = mutex (fun () ->
      List.fold_left
	(fun res ac ->
	  if Ext.mem oid (Query.extent ac.action_query)
	  then
	    let args = action_args oid in
	    res @ apply_action ac args
	  else [])
	[]
	(get_actions ()))

    let well_actions url = mutex (fun () ->
      try
	let wll, rule = Hashtbl.find l_wll url in
	let args = Wll.action_args wll in
	let actions =
	  List.fold_left
	  (fun res ac ->
	    res @ apply_action ac args)
	  []
	  (get_actions ()) in
	actions
      with Not_found -> failwith ("unknown well: " ^ url))

    let temp_well_actions url cs = mutex (fun () ->
      let wll = Wll.make url in
      let args = Wll.action_args wll in
      let actions = 
	List.fold_left
	  (fun res ac ->
	    res @ apply_action ac args)
	  []
	  (get_actions ()) in
      List.map
	(fun (name,code) ->
	  (name, (fun () -> Wll.init wll cs; code ())))
	actions)
      
(*	    
      let cmd =
	List.fold_left
	  (fun res (attr,v) ->
	    Str.global_replace (Str.regexp ("\\$(" ^ attr ^ ")")) v res)
	  ac.action_cmd
	  args in
      if Str.string_match (Str.regexp ".*\\$(.+)") cmd 0
      then failwith "This action is not applicable to this object"
      else (fun () -> ignore (Sys.command cmd))
*)

(*** performing updates ***)

    let rec do_update : update -> unit =
      fun update ->
	try match update with
	| Init (url_source,url,options) ->
            let src = Hashtbl.find l_src url_source in
	    init_source ~ss:src.s url options src.s_home
	| Update (url,options) ->
	    update_source_options url options
	| Final (url) ->
	    final_source url
	| Move (url,new_url) ->
	    relocate_source url new_url
	| InitWll (url,query) ->
	    let q = Query.parse (Syntax.of_list query) in
	    init_well url q
	| UpdateWll (url) ->
	    update_well url
	| FinalWll (url) ->
	    final_well url
	| MoveWll (url,new_url) ->
	    relocate_well url new_url
	| Make (url,r,d) -> mutex (fun () -> Common.prof "do_update.Make" (fun () ->
            let src = Hashtbl.find l_src url in (* get_src (Src.parse (Syntax.of_list source)) in *)
	    let fd = Object.Descr.parse (Syntax.of_list d) in
	    let updates = Object.make (Object.External (src.s,r), fd, src.s_home) in
            push_updates updates))
	| UpdateIntrSrc (url,ref,d) -> mutex (fun () ->
	    let src = Hashtbl.find l_src url in
	    let oid = Object.get_oid (src.s,ref) in
	    let updates = Object.update (oid, Object.Descr.parse (Syntax.of_list d)) in
	    push_updates updates)
        | MakeExtr (pre,args,d) -> mutex (fun () ->
	    let updates = Object.make (Object.Internal (pre,args), Object.Descr.empty, Object.Descr.parse (Syntax.of_list d)) in
            push_updates updates)
	| MoveIntr (oid,u) ->
	    move_intr_obj oid (Root.Update.parse (Syntax.of_list u))
	| MoveExtr (oid,u) ->
            move_obj oid (Root.Update.parse (Syntax.of_list u))
	| MoveExtrSrc(url,ref,u) ->
	    let src = Hashtbl.find l_src url in (* get_src (Src.parse (Syntax.of_list src)) in *)
	    let oid = Object.get_oid (src.s,ref) in
	    move_obj oid (Root.Update.parse (Syntax.of_list u))
	| Remove oid -> mutex (fun () ->
	    let updates = Object.remove oid in
            push_updates updates)
	| RemoveSrc (url,ref) -> mutex (fun () ->
	    let src = Hashtbl.find l_src url in
	    let oid = Object.get_oid (src.s,ref) in
	    let updates = Object.remove oid in
            push_updates updates)
	| Axiom (f,g,url) ->
	    let s_opt =
	      if url=""
	      then None
	      else Some (Hashtbl.find l_src url).s in
	    add_axiom ?src:s_opt (Feat.parse (Syntax.of_list f)) (Feat.parse (Syntax.of_list g))
	| Show f ->
	    add_feature (Feat.parse (Syntax.of_list f))
	| Hide f ->
            del_feature (Feat.parse (Syntax.of_list f))
	| InitRuleIntr (p,c) ->
	    ignore (init_rule (Query.parse (Syntax.of_list p)) (Intr (Root.Update.parse (Syntax.of_list c))))
(*
	| FinalRuleIntr (p,c) ->
	    final_rule (Query.parse (Syntax.of_list p)) (Intr (Root.Update.parse (Syntax.of_list c)))
*)
	| InitRuleExtr (p,c) ->
	    ignore (init_rule (Query.parse (Syntax.of_list p)) (Extr (Root.Update.parse (Syntax.of_list c))))
(*
	| FinalRuleExtr (p,c) ->
	    final_rule (Query.parse (Syntax.of_list p)) (Extr (Root.Update.parse (Syntax.of_list c)))
*)
	| InitRuleRem (p) ->
	    ignore (init_rule (Query.parse (Syntax.of_list p)) Rem)
(*
	| FinalRuleRem (p) ->
	    final_rule (Query.parse (Syntax.of_list p)) Rem
*)
	| Action (name,cmd,query) ->
	    add_action {action_name=name; action_cmd=cmd; action_query = Query.parse (Syntax.of_list query)}
	with e ->
	  print_endline (Syntax.stringizer (Source.print_update update));
	  print_string ("Context.do_update: " ^ Printexc.to_string e ^ "\n")

    let p_updates : Thread.t =
      Thread.create
	(fun () ->
	  while true do
	    try mutex (fun () ->
	      let update = SetUpdates.min_elt !set_updates in
	      set_updates := SetUpdates.remove update !set_updates;
	      do_update update;
	      Thread.yield ())
	    with Not_found ->
	      Thread.delay 1.
	  done
	) ()

    let replace_feature f f' = mutex (fun () ->
      Lock.check ();
      (* axioms *)
      replace_in_axioms f f';
      let u = Root.Update.Replace (f,f') in
      (* sources *)
      Hashtbl.iter
	(fun url src ->
	  let h', _ = Object.Descr.update src.s_home u in
	  Hashtbl.replace l_src url {src with s_home = h'})
	l_src;
      (* objects *)
      Ext.iter
	(fun oid ->
	  let updates = Object.move oid u in
	  push_updates updates)
	(!Root.cache # ext f);
      (* visibility *)
      del_feature f;
      add_feature f')

(*** querying and navigation ***)

(* get the extent of the working query: a list of (oid, description) *)
    let extent () = mutex (fun () ->
      Query.extent (wq ()))


(* deprecated by cache *)
(* ******************

(* get increments of the working query *)
(* ar_ext: array indexed on node idents, such that ar_ext.(oid) <> 0 if oid is in working extent *)
(* l: size of the working extent *)
(* view: working view *)
(* subview_opt: true if subviews must be returned, false if only non-views must be *)

(* get an extent array and size from an extent *)

    (* bool is true is ext is all objects *)
    let ar_ext_l_from_ext : Ext.t -> (bool * int array) * int =
      fun ext -> mutex (fun () ->
	let ar_ext = Array.create (!Oid.cpt + 1) 0 in
	let l = Ext.fold (fun l oid -> ar_ext.(oid) <- 1; l+1) 0 ext in
	(l = Ext.cardinal (!Root.cache # all), ar_ext), l)

(* membership predicate in an extent array *)
    let mem_ar_ext : bool * int array -> int -> bool =
      fun (full, ar_ext) oid ->
        full or ar_ext.(oid) <> 0

(* get intersection and support between an extent array and a node *)
    let inter_ar_ext : bool * int array -> Ext.t -> int * Ext.t =
      fun (full,ar_ext) ext ->
	if full
	then Ext.cardinal ext, ext
	else Ext.fold
	    (fun (supp,inter) oid ->
	      if ar_ext.(oid) <> 0
	      then supp+1, Ext.add oid inter
	      else supp, inter)
	    (0,Ext.empty) ext

(* get subtraction and support between an extent array and a node : node \ extent array *)
    let subtract_ar_ext : bool * int array -> Ext.t -> int * Ext.t =
      fun (full,ar_ext) ext ->
	if full
	then 0, Ext.empty
	else Ext.fold
	    (fun (supp,inter) oid ->
	      if ar_ext.(oid) = 0
	      then supp+1, Ext.add oid inter
	      else supp, inter)
	    (0,Ext.empty) ext

    let rec incrs : Query.t * bool * int array * int -> Feat.elt -> (Feat.elt -> Node.view_state) -> bool -> (Feat.t -> bool) (*Str.regexp option*) ->  bool -> int -> (int * Ext.t * Feat.elt) list =
      fun (wq,full,ar_ext,l) start view_pred subview_opt log_pred (*pattern_opt*) mostspec_opt supp_min -> mutex (fun () ->
	let vis n =
	  (Node.pred_of_view !Node.features n).Node.vis &
	  (view_pred n).Node.vis &
(*	  not (subview_opt & Query.entails wq (Node.f n)) &*)
(*	  Node.obj n = None &*)
	  log_pred (Node.f n) in
(*
	  ( match pattern_opt with
	  | None -> true
	  | Some p ->
	      try ignore (Str.search_forward p (Syntax.stringizer (Feat.print (Node.f n))) 0); true
	      with Not_found -> false) in *)
(*
	let more_vis n =
	  (Node.pred_of_view !Node.features n).Node.more_vis &
	  (view_pred n).Node.more_vis in
*)
	let incrs0 = Node.first_topo
	    ?prev:(if Node.id start = Node.id (!Root.cache # top) then Some (fun n -> LSet.cardinal (Node.sup n)) else None)
	    ~pred:(fun n ->
	      if vis n
	      then
		let l',inter = inter_ar_ext (full,ar_ext) (!Root.cache # ext_elt n) in
		(if
		  l' >= supp_min &
		  (l' < l or subview_opt) 
		  (* & (List.length (incrs ar_ext l (Node.inter_view view (Node.Atom n)) false supp_min)) >= 2))*)
		then begin
		  Ext.iter (fun oid -> ar_ext.(oid) <- 2) inter;
		  Some (l',inter,n) end
		else
		  None),
		(Some l')
	      else None, None)
	    ~next:(fun n -> function
	      | Some (Some supp) -> if supp > supp_min (* & more_vis n *) then Node.sub n else []
	      | _ -> Node.sub n)
	    ~start:(Node.from_sub (Node.sub start (*(!Root.cache # top)*) )) in
(*
	let incrs0 = Node.afirst
	    ~pred_next:(fun n ->
	      if vis n
	      then
		let l',inter = inter_ar_ext (full,ar_ext) (!Root.cache # ext_elt n) in
		(if
		  l' >= supp_min &
		  (l' < l or subview_opt) 
		  (* & (List.length (incrs ar_ext l (Node.inter_view view (Node.Atom n)) false supp_min)) >= 2))*)
		then begin
		  List.iter (fun oid -> ar_ext.(oid) <- 2) inter;
		  Some (l',inter,n) end
		else None),
		(if l' > supp_min & more_vis n then Node.sub n else [])
	      else None, Node.sub n)
	    ~start:(Node.sub (!Root.cache # top))
	    ~prev:Node.sup in
*)
	let post_process incrs0 =
	  if mostspec_opt
	  then
	    List.fold_left
	      (fun incrs (l',inter,n) ->
                (*if mostspec_opt
		then*)
                  let ns' = Node.last
		    ~pred_next:(fun n' ->
		      if n' == n
		      then Some n', Node.sub n'
		      else
			if vis n'
			then
			  if fst (inter_ar_ext (full,ar_ext) (!Root.cache # ext_elt n'))=l'
			  then Some n', Node.sub n'
			  else None, []
			else None, Node.sub n')
		    ~start:n in
		  List.fold_left (fun incrs n' -> LSet.add (l',inter,n') incrs) incrs ns'
                (*else
                  LSet.add (l',inter,n) incrs*)
              )
	      []
	      incrs0
	  else incrs0 in
	match incrs0 with
	| [l',_,n] (* when l' = l *) ->
	    let p = Node.pred_of_view (Node.get_view (Node.Atom (Node.Down,n))) in
	    let incrs1 = incrs 
		(wq,full,ar_ext,l)
		(if Node.id start = Node.id (!Root.cache # top) then (!Root.cache # top) else n)
		(fun n -> Node.view_state_and
		    (view_pred n)
		    (p n)) 
		subview_opt log_pred (*pattern_opt*) mostspec_opt supp_min in
	    post_process (if incrs1 = [] then incrs0 else incrs1)
	| _ -> post_process incrs0
)

    let rec incrs_links : Query.t * bool * Ext.t * int -> Feat.elt -> (Feat.elt -> bool) -> bool ->  bool -> int -> (int * Ext.t * Feat.elt) list =
      fun (wq,full,ext,l) start view_pred subview_opt mostspec_opt supp_min -> mutex (fun () ->
	let cardinal_inter_full =
	  if full
	  then (fun n ->
	    let inter = !Root.cache # ext_elt n in
	    Ext.cardinal inter, inter)
	  else (fun n ->
	    let inter = Ext.inter ext (!Root.cache # ext_elt n) in
	    Ext.cardinal inter, inter) in
	let vis n =
	  (Node.pred_of_view !Node.features n).Node.vis &
	  (view_pred n) in
(*	  log_pred (Node.f n) in*)
	let incrs0 = Node.first_topo
	    ?prev:(if Node.id start = Node.id (!Root.cache # top) then Some (fun n -> LSet.cardinal (Node.sup n)) else None)
	    ~pred:(fun n ->
	      if vis n
	      then
		let l', inter = cardinal_inter_full n in
		(if
		  l' >= supp_min &
		  (l' < l or subview_opt)
		then Some (l',inter,n)
		else None),
		(Some l')
	      else
		None, None)
	    ~next:(fun n -> function
	      | Some (Some supp) -> if supp > supp_min then Node.sub n else []
	      | _ -> Node.sub n)
	    ~start:(Node.from_sub (Node.sub start)) in
	let post_process incrs0 =
	  if mostspec_opt
	  then
	    List.fold_left
	      (fun incrs (l',inter,n) ->
                let ns' = Node.last
		    ~pred_next:(fun n' ->
		      if n' == n
		      then Some n', Node.sub n'
		      else
			if vis n'
			then
			  if fst (cardinal_inter_full n') = l'
			  then Some n', Node.sub n'
			  else None, []
			else None, Node.sub n')
		    ~start:n in
		List.fold_left (fun incrs n' -> LSet.add (l',inter,n') incrs) incrs ns')
	      []
	      incrs0
	  else incrs0 in
	(*match incrs0 with
	| [l',_,n] (* when l' = l *) ->
(*	    let p = Node.pred_of_view (Node.get_view (Node.Atom (Node.Down,n))) in*)
	    let incrs1 = incrs_links
		(wq,full,ext,l)
		n (*if Node.id start = Node.id (!Root.cache # top) then (!Root.cache # top) else n*)
		view_pred
		subview_opt (*log_pred*) mostspec_opt supp_min in
	    post_process (if incrs1 = [] then incrs0 else incrs1)
	| _ ->*) post_process incrs0)

*) (* deprecated by cache *)


    let feat_of_node : Feat.elt -> Feat.t =
      fun n -> !Root.cache # formula n

    type propr = {
	propr_card : int;
	propr_supp : int;
	propr_ext : Ext.t;
	propr_logs : Query.t list;
	propr_sups : (Ext.t * Query.t) list;
      }

    type view = {
        nbobj : int;  (** Total number of objects. *)
	nbsel : int;  (** Number of objects in the current extent. *)
        extent : Ext.t;  (** Current extent. *)
	locals : Ext.t; (** Local objects. *)
        proprs : propr list;  (** increments. *)
      }

    let sort_log : (int * Ext.t * Query.t * Ext.t) list -> propr list =
      fun incrs -> List.map
	  (fun (i,inter,n,ext_n) -> {
	    propr_card = Ext.cardinal ext_n; (* for efficiency, to be moved down to logic caches, *)
	    propr_supp = i;
	    propr_ext = inter;
	    propr_logs = [n];
	    propr_sups = []
	  })
	  incrs
	  (* List.sort (fun (l1,_,_) (l2,_,_) -> compare l1 l2) incrs *) (* sort is let to the interface *)

(*
    let rec sort_ctx : (int * Ext.t * Query.t) list -> propr list =
      fun incrs -> sort_ctx2 [] (List.sort (fun (l1,_,_) (l2,_,_) -> compare l2 l1) incrs)
    and sort_ctx2 res = function (* the list is sorted in decreasing order of the support of increments *)
	[] -> res
      | ((l,ext,n) as incr)::incrs ->
	  try sort_ctx2 (merge_with_equiv incr res) incrs
	  with Not_found -> sort_ctx2 (extend_with_sups (l,ext,[n],[]) res::res) incrs
    and merge_with_equiv (l,ext,n) = function
	[] -> raise Not_found
      |	(l',ext',xs,sups)::res -> (* we have here l'>= l because of above sort *)
	  if l'>l then raise Not_found
	  else (* l'=l *) if ext'=ext then (l',ext',n::xs,sups)::res
	  else (l',ext',xs,sups)::merge_with_equiv (l,ext,n) res
    and extend_with_sups (l,ext,xs,sups) = function
	[] -> (l,ext,xs,sups)
      |	(l',ext',xs',sups')::res ->
	  if l'>l & Ext.subset ext ext' &
	    not (List.exists (fun (ext1,_) -> Ext.subset ext1 ext') sups)
	  then extend_with_sups (l,ext,xs,(ext',List.hd xs')::sups) res
	  else extend_with_sups (l,ext,xs,sups) res
*)

    let rec sort_max : int -> (int * Ext.t * Query.t * Ext.t) list -> propr list =
      fun lmax incrs -> sort_max2 lmax [] (List.sort (fun (l1,_,_,_) (l2,_,_,_) -> compare l2 l1) incrs)
    and sort_max2 lmax res = function (* the list is sorted in decreasing order of the support of increments *)
      | [] -> res
      | (i,inter,n,ext_n)::incrs ->
	  let incr = {
	    propr_card = Ext.cardinal ext_n;
	    propr_supp = i;
	    propr_ext = inter;
	    propr_logs = [n];
	    propr_sups = []
	  } in
	  sort_max2 lmax (merge_or_delete lmax incr res) incrs
    and merge_or_delete imax incr = function
      | [] -> [incr] (* inserted *)
      |	p::res -> (* we have here p.propr_supp >= l because of above sort *)
	  if p.propr_supp > incr.propr_supp then
	    if imax > p.propr_supp & Ext.subset incr.propr_ext p.propr_ext
	    then p::res (* deleted *)
	    else p::merge_or_delete imax incr res
	  else
	    if incr.propr_ext = p.propr_ext
	    then incr::p::res (* merged *)
	    else p::merge_or_delete imax incr res

    let get_locals : int array -> Ext.t =
      fun ar_ext ->
	let ref_locals = ref Ext.empty in
	Array.iteri (fun oid v -> if v = 1 then ref_locals := Ext.add oid !ref_locals) ar_ext;
	!ref_locals

    let get_locals_links : Ext.t -> (int * Ext.t * Query.t * Ext.t) list -> Ext.t =
      fun ext incrs ->
	List.fold_left
	  (fun res (_,inter,_,_) ->
	    if Ext.subset ext inter
	    then res (* we do not use views for computing locals *)
	    else Ext.diff res inter)
	  ext
	  incrs

(* return the count of an increment *)
    let link x = mutex (fun () ->
      let i, inter, ext_x = Root.Query.increment (wq ()) x in
      {propr_card = Ext.cardinal ext_x;
       propr_supp = i;
       propr_ext = inter;
       propr_logs = [x];
       propr_sups = []
     })

(* returns increments for working query and view, local objects, and size of working extent *)
(* opt: true if subviews must be returned *)
    type links_option = [`NoObjs|`NoViews|`Implications|`MaxOnly|`MostSpec|`SuppMin of int|`Neg|`Pattern of string|`LexicalOrder|`ValOnly]

    let links (options : links_option list) q_as_view = mutex (fun () -> Common.prof "Context.links" (fun () ->
      let wq = wq () in
      let ext, incrs =
	Root.Query.increments
	  ~suppmin:(List.fold_left (fun suppmin -> function `SuppMin x -> x | _ -> suppmin) 1 options)
	  wq
	  q_as_view
(*
	  (List.fold_left
	     (fun view_pred -> function
	       | `ValOnly -> (fun n -> Node.obj n > 0 & view_pred n)
	       | `Pattern s -> (fun n ->
		   (try ignore (Str.search_forward (Str.regexp s) (Syntax.stringizer (Feat.print (Node.f n))) 0); true
		   with Not_found -> false)
		     & view_pred n)
	       | _ -> view_pred)
	     (fun n -> (pred_wv n).Node.vis)
	     options)
	  (not (List.mem `NoViews options (*or List.mem `MaxOnly options*)))
          (List.mem `MostSpec options)
*)
      in
      let l = Ext.cardinal ext in
      let proprs =
(* deprecated
	if List.mem `Implications options
	then sort_ctx incrs
*)
	if List.mem `MaxOnly options
	then List.rev (sort_max l incrs)
	else sort_log incrs in
      let local_objs = 
	if List.mem `NoObjs options
	then Ext.empty
	else get_locals_links ext incrs in
      {nbobj = Ext.cardinal (Query.extent Query.top);
       nbsel = l;
       extent = ext;
       proprs = proprs;
       locals = local_objs}))

    let intent oids = List.map (!Root.cache # formula) (!Root.cache # int oids)


(* !!! TO BE REMADE ENTIRELY LATER *)
(* ***********

(* Conceptual operations *)
(* --------------------- *)

    type concept = Ext.t * Feat.t list * Ext.t
    type hyp = int * float * Feat.t list * Feat.t list list  (* support, confidence, formula, equivalents *)
    type cc = (bool * Feat.t) list  (* Conjunctive Clause *)
    type lr = int * Query.t * cc  (* Logical Rule (id, head, body) *)
    type confusion = {
      true_pos : Ext.t;
      false_neg : Ext.t;
      false_pos : Ext.t;
      true_neg : Ext.t}

    type c = Ext.t * Feat.elt list list * Ext.t

    let refine_nss : Feat.elt list list -> Feat.elt list -> Feat.elt list list =
      fun nss ns -> ns::nss

    let refine_int : Feat.t list -> Feat.elt list -> Feat.t list =
      fun int xs -> mutex (fun () ->
        Object.Descr.weak_conj int (List.map feat_of_node xs))

    let int_of_nss : Feat.elt list list -> Feat.t list =
      fun nss -> List.fold_right (fun ns res -> refine_int res ns) nss []

    let refine_eqs : (Ext.t * Feat.t list) list -> (Ext.t * Feat.elt) list -> (Ext.t * Feat.t list) list = 
      fun eqs xs ->
	let xs', eqs' =
	  List.fold_right
	    (fun (ext',fs') (xs,eqs) ->
	      let xs' = List.filter (fun (ext,x) -> Ext.subset ext ext') xs in
              let xs'' = List.filter (fun (ext,x) -> Ext.cardinal ext = Ext.cardinal ext') xs' in
	      if xs'' <> [] then ([],(ext',fs')::eqs)
              else if xs' <> [] then (xs',eqs)
              else (xs,(ext',fs')::eqs)
	    ) eqs (xs,[]) in
	if xs' = []
        then eqs'
        else (Ext.union_r (List.map fst xs'),mutex (fun () -> List.map (fun n -> feat_of_node (snd n)) xs'))::eqs'

    let eqs_of_nss : (Feat.elt -> Ext.t) -> Feat.elt list list -> Feat.t list list =
      fun get_ext nss ->
        let nss' = List.map (fun ns -> List.map (fun n -> get_ext n, n) ns) nss in
        List.map snd (List.fold_right (fun ns res -> refine_eqs res ns) nss' [])

    let greedy_max_of_int (dom,wq) (full,ar_wq) (t_ext : Ext.t) (t_int : ((bool * Feat.elt) * float) list) : cc list = mutex (fun () ->
      let t_e = Ext.cardinal t_ext in
      let xs =
        List.map
          (fun ((p,n),v) ->
	    let n_e, n_ext =
	      if p
	      then inter_ar_ext (full,ar_wq) (!Root.cache # ext_elt n)
	      else
		let n_ext = Ext.diff wq (!Root.cache # ext_elt n) in
		Ext.cardinal n_ext, n_ext in
            let f = feat_of_node n in
	    (n_e, v, n_ext, (p, f), List.length (Feat.print f)))
	  t_int in
      let compare = fun (e1,v1,_,_,s1) (e2,v2,_,_,s2) -> compare (e1,-.v1,s1) (e2,-.v2,s2) in
      let xs = List.sort compare xs in
      let _, _, f, _ = Common.fold_while
        (fun (xs,ext,f_xs,s_xs) -> 
           if ext = t_ext
           then None
           else
             match xs with
             | [] -> None
             | (_,_,ext',f_x,s_x)::xs_rest ->
               let xs' =
                 List.sort
                   compare
                   (List.map
                      (fun (e,v,ext_x,f,s) ->
                         let ext_x' = Ext.inter ext' ext_x in
                         (Ext.cardinal ext_x', v, ext_x', f,s))
		      xs_rest) in
               Some (xs',ext',f_x::f_xs,s_x+s_xs+1))
        (xs, wq, [], 0) in
      [f])

    let rec max_of_int (dom,wq) (full,ar_wq) t_ext t_int : Query.t list = mutex (fun () ->
      let t_e = Ext.cardinal t_ext in
      let xs = List.map (fun n -> 
          let n_e, n_ext = inter_ar_ext (full,ar_wq) (!Root.cache # ext_elt n) in
          (n_e, n_ext, Node.Atom (Node.Down, n))
        ) t_int in
      let xs =
	List.sort
	  (fun (e1,_,f1) (e2,_,f2) ->
            let c = compare e1 e2 in
            if c <> 0 then c
	    else if Query.entails f1 f2 then -1
	    else if Query.entails f2 f1 then 1
	    else 0)
	  xs in
      max_of_int_loop t_e [(dom,wq,[],xs)])
    and max_of_int_loop t_e l =
      if l = []
      then []
      else
       let lmax = List.fold_right (fun (e,_,max,_) res ->
           if e = t_e
           then
             let f = List.fold_left (fun res (_,_,f) -> Node.And (res,f)) Node.All max in
             if List.exists (fun f' -> Query.entails f f') res
             then res
             else f::res
           else res
         ) l [] in
       if lmax <> []
       then lmax
       else
        let l' =
          List.fold_right (fun (e,ext,ys,xs) res ->
            snd (List.fold_right (fun (x_e,x_ext,x_f as x) (xs',res') -> 
              if List.exists (fun (y_e,y_ext,y_f) -> x_e >= y_e & Ext.subset y_ext x_ext) ys
              then x::xs', res'
              else
                let ext' = Ext.inter ext x_ext in
                let e' = Ext.cardinal ext' in
                x::xs', (e',ext',x::ys,xs')::res'
            ) xs ([],res))
          ) l [] in
        max_of_int_loop t_e l'

    let max_of_nss (dom,wq) (full,ar_wq) t_ext : Feat.elt list list -> Query.t list =
      fun nss ->
        max_of_int (dom,wq) (full,ar_wq) t_ext (List.fold_left (@) [] nss)

    let conf dom prem concl inter =
       float_of_int(inter) /. float_of_int(prem)

    let conf_gui dom prem concl inter =
       if concl = dom
       then 1.0 (* to avoid further division by 0. *)
       else
         let wq_inter = dom * inter in
         let prem_concl = prem * concl in
         if wq_inter >= prem_concl
         then float_of_int(wq_inter - prem_concl) /. float_of_int(dom * prem - prem_concl)
         else float_of_int(wq_inter - prem_concl) /. float_of_int(prem_concl)

(* input of a set of rules from a file .lrs *)
    let input_rules filename = mutex (fun () ->
      let ch = open_in_bin (filename ^ ".lrs") in
      let res = ref [] in
      begin try
        while true do
          res := (input_value ch : lr)::!res
        done
      with _ -> () end;
      List.rev !res)

(* output of a set of rules to a file *)
    let output_rules filename lrs = mutex (fun () ->
      let ch = open_out_bin (filename ^ ".lrs") in
      List.iter
        (fun lr -> output_value ch (lr : lr))
        lrs;
      close_out ch)

(* test if an object is covered by some rule *)
    let rule_covered oid (id,head,body : lr) = mutex (fun () ->
      let d = Object.get_descr oid in
      List.for_all (fun (sign,x) -> sign = Object.Descr.entails d x) body)

(* get the confusion matrix of some rule over some extent *)
    let eval_rule (id,head,body : lr) (ext : Ext.t) : confusion = mutex (fun () ->
      let pos = Ext.inter ext (Query.extent head) in (* (!Root.cache # ext_elt (Node.of_formula head)) in *)
      let cov =
        List.fold_left
          (fun res (sign,x) ->
             let x_ext = !Root.cache # ext_elt (Node.of_formula x) in
             if sign
             then Ext.inter res x_ext
             else Ext.diff res x_ext)
          ext
          body in
      { true_pos = Ext.inter cov pos;
        false_neg = Ext.diff pos cov;
        false_pos = Ext.diff cov pos;
        true_neg = Ext.diff ext (Ext.union pos cov)})

(* get the closed extent, the intent and the max labels from some object set *)
    let intent oids = mutex (fun () ->
      let wq = extent () in
      let (full, ar_wq), dom = ar_ext_l_from_ext wq in
      let _, oids = inter_ar_ext (full,ar_wq) oids in
      if Ext.is_empty oids then raise (Failure "Error: empty set of objects.");
      let p_view =
	let p1 = Node.pred_of_view !Node.features in
	let p2 = Node.pred_of_view (get_wv ()) in
	fun n -> Node.view_state_and (p1 n) (p2 n) in
      let lubs = Node.lubs
        ~pred:(fun n _ -> if (p_view n).Node.vis then Some n else None)
        ~suppmin:(Ext.cardinal oids)
         (Ext.map (fun oid -> Node.from_sup (Node.links oid)) oids) in
      let _, ext = inter_ar_ext (full,ar_wq) (Ext.inter_r (List.map !Root.cache # ext_elt lubs)) in
      let int = List.map feat_of_node lubs in
      let maj =
        Node.fold
          (Node.from_sup lubs)
          (fun n res ->
             let vst = p_view n in
             (if vst.Node.vis then ((true,n),0.)::res else res),
             (Node.sup n))
          (fun n res -> res)
          [] in
      let max = greedy_max_of_int (dom,wq) (full,ar_wq) ext maj in
      ext, int, max)


    let rec synth_concepts : bool -> bool -> Ext.t -> (Feat.elt -> Node.view_state) -> int -> (c -> 'h -> 'h * bool) -> (c -> 'h -> 'a list -> bool) -> (c -> 'h -> 'a list -> 'a) -> 'h -> 'a -> 'a =
      fun opt mostspec ext view_pred suppmin herit next synth h0 s0 ->
        let int =
          if mostspec
          then Node.last
            ~pred_next:(fun n ->
               let vst = view_pred n in
               (if vst.Node.vis & Ext.subset ext (!Root.cache # ext_elt n) then Some n else None),
               (if vst.Node.more_vis then Node.sub n else []))
            ~start:(!Root.cache # top)
          else [] in
        synth_concepts2 opt mostspec (Hashtbl.create 1000) ext [int] view_pred suppmin herit next synth h0 s0
    and synth_concepts2 =
      fun opt mostspec tbl_ext_res ext int view_pred suppmin herit next synth h s0 ->
        let (full, ar_ext), l = ar_ext_l_from_ext ext in
	let proprs = sort_max l (incrs (Query.top,full,ar_ext,l) (!Root.cache # top) view_pred false (fun f -> true) (*None*) mostspec suppmin) in
        let c = (ext,int,get_locals ar_ext) in
        let h', cont = herit c h in
        if not cont
        then s0
	else begin
print_int (Ext.cardinal ext); print_string "<"; flush stdout;
         let rec f c h' l_res = function
           | [] -> l_res
           | (_,inter,xs,_)::proprs ->
                let cont = next c h' l_res in
                if cont
                then
                  let res =
                    try Hashtbl.find tbl_ext_res inter
	            with Not_found ->
	              synth_concepts2 opt mostspec tbl_ext_res inter (refine_nss int xs) view_pred suppmin herit next synth h' s0 in
                  f c h' (res::l_res) proprs
                else l_res in
         let l_res = f c h' [] proprs in
         let res = synth c h' l_res in
         Hashtbl.add tbl_ext_res ext (if opt then s0 else res);
print_string ">";
         res end

    let rec fold_concepts : Ext.t -> (Feat.elt -> Node.view_state) -> int -> (c -> 'a -> 'a * bool) -> 'a -> 'a =
      fun ext view_pred suppmin f init ->
        let size_view = Node.fold
          (!Root.cache # top) 
          (fun n size -> (if (view_pred n).Node.vis then size+1 else size), Node.sub n)
          (fun n size -> size)
          0 in
        print_string "fold_concepts: number of features is "; print_int size_view; print_newline (); flush stdout;
	fold_concepts2 (Hashtbl.create 1000) ext [] view_pred suppmin f init
    and fold_concepts2 =
      fun tbl_ext ext int view_pred suppmin f res0 ->
	Hashtbl.add tbl_ext ext ();
        let (full, ar_ext), l = ar_ext_l_from_ext ext in
	let proprs = sort_max l (incrs (Query.top,full,ar_ext,l) (!Root.cache # top) view_pred false (fun f -> true) (*None*) true suppmin) in
        let res1, cont = f (ext,int,get_locals ar_ext) res0 in
        if not cont
        then res1
	else begin
print_int (Ext.cardinal ext); print_string "<"; flush stdout;
	 let res = List.fold_left
	  (fun res (_,inter,xs,_) ->
	    if Hashtbl.mem tbl_ext inter
	    then res
	    else fold_concepts2 tbl_ext inter (refine_nss int xs) view_pred suppmin f res)
	  res1
	  proprs in
print_string ">"; res end

(* get first concepts of the sub-context defined by ext and view that satisfy pred *)
(* caution: pred must not rely on the set of local objects of concepts *)
    let rec first_concepts : Ext.t -> Feat.elt list list -> (Feat.elt -> Node.view_state) -> int -> (c -> 'a option * bool) -> 'a list =
      fun ext int view_pred supp_min pred_cont ->
	snd (List.split (first_concepts2 (Hashtbl.create 1000) ext int view_pred supp_min pred_cont [] 1))
    and first_concepts2 =
      fun tbl_ext ext int view_pred supp_min pred_cont res0 i ->
	Hashtbl.add tbl_ext ext ();
	let (full, ar_ext), l = ar_ext_l_from_ext ext in
	match pred_cont (ext,int,Ext.empty) with
	| Some x, _ -> (* new first concept *)
print_string "+"; flush stdout;
	    insert_first_concept (ext,x) res0
	| None, true -> (* continue below *)
print_int i; print_string "<"; flush stdout;
	    let proprs = sort_max l (incrs (Query.top,full,ar_ext,l) (!Root.cache # top) view_pred false (fun f -> true) (*None*) false supp_min) in
	    let res = List.fold_left
	      (fun res (_,inter,xs,_) ->
		if Hashtbl.mem tbl_ext inter
		then res
		else first_concepts2 tbl_ext inter (refine_nss int xs) view_pred supp_min pred_cont res (i+1))
	      res0
	      proprs in
	    print_string ">"; flush stdout; res
	| None, false -> (* stop *)
print_string "-"; flush stdout;
	    res0
    and insert_first_concept (ext,x) = (* add concept (ext,x) in list by removing smaller concepts *)
      function
	| [] -> [(ext,x)]
	| (ext',x')::l ->
	    if Ext.subset ext ext'
	    then (ext',x')::l
	    else if Ext.subset ext' ext
	    then insert_first_concept (ext,x) l
	    else (ext',x')::insert_first_concept (ext,x) l


    let concepts suppmin = mutex (fun () ->
      synth_concepts true true
	(Query.extent (wq ())) 
	(Node.pred_of_view (get_wv ()))
	suppmin
        (fun c () -> (), true)
        (fun c h l_res -> true)
        (fun (ext,nss,loc) h l_res -> (ext,int_of_nss nss,loc) :: List.fold_left (fun res res' -> res' @ res) [] l_res)
        () []
    )

    let assocs suppmin oid = mutex (fun () ->
      let p_view = Node.pred_of_view (get_wv ()) in
      let p_obj = Node.pred_of_obj (Ext.singleton oid) 1 in
      fold_concepts 
	(Ext.remove oid (Query.extent (wq ())))
	(fun node -> Node.view_state_and (p_obj node) (p_view node))
	suppmin
	(fun (ext,nss,loc) l -> (ext,int_of_nss nss,loc)::l, true) [])

    let sufs suppmin confmin target = mutex (fun () ->
      let wq = Query.extent (wq ()) in
      let dom = Ext.cardinal wq in
      let target' = Ext.inter wq target in (* restricting the target to the wq *)
      let t = Ext.cardinal target' in
      let p_view = Node.pred_of_view (get_wv ()) in
      let p_target_nec = Node.pred_of_obj target' t in
      let p_target_sufs = Node.pred_of_obj target' (truncate (ceil (float_of_int suppmin *. confmin))) in
      let (ext,nss) = synth_concepts true false
        wq (fun n -> Node.view_state_and (p_target_nec n) (p_view n)) t
        (fun (_,_,_) () -> (), true)
        (fun (_,_,_) () l_res -> l_res = [])
        (fun (ext,nss,_) () l_res ->
           if l_res = []
           then (ext,nss)
           else List.hd l_res)
        () (Ext.empty,[]) in
      first_concepts 
	ext nss
	(fun node -> Node.view_state_and (p_target_sufs node) (p_view node)) 
	suppmin
	(fun (ext,nss,_) ->
	  let e = Ext.cardinal ext in
	  let s = Ext.cardinal (Ext.inter target' ext) in
	  let c = conf_gui dom e t s in (*(float_of_int s) /. (float_of_int e) in*)
	  if e >= suppmin &  c >= confmin
	  then Some (e,c,int_of_nss nss,eqs_of_nss (fun n -> Ext.diff (Ext.inter (!Root.cache # ext_elt n) wq) target') nss), false
	  else None, e > suppmin & conf_gui dom suppmin t s >= confmin
	))

    let nec target = mutex (fun () ->
      let wq = Query.extent (wq ()) in
      let dom = Ext.cardinal wq in
      let target' = Ext.inter wq target in (* restricting the target to the wq *)
      let t = Ext.cardinal target' in
      let p_view = Node.pred_of_view (get_wv ()) in
      let p_target = Node.pred_of_obj target' t in
      synth_concepts true true
        wq (fun n -> Node.view_state_and (p_target n) (p_view n)) t
        (fun (_,_,_) () -> (), true)
        (fun (_,_,_) () l_res -> l_res = [])
        (fun (ext,nss,_) () l_res ->
           if l_res = []
           then (Ext.cardinal ext,1.0,int_of_nss nss,eqs_of_nss (fun n -> Ext.diff (Ext.inter (!Root.cache # ext_elt n) wq) target') nss)
           else List.hd l_res)
        () (0,0.,[],[]))

    let ind suppmin confmin oid target = mutex (fun () ->
      let wq = Ext.remove oid (Query.extent (wq ())) in
      let dom = Ext.cardinal wq in
      let target' = Ext.inter wq target in  (* restricting the target to the wq *)
      let (full, ar_target), t = ar_ext_l_from_ext target' in
      let p_view = Node.pred_of_view (get_wv ()) in
      let p_obj = Node.pred_of_obj (Ext.singleton oid) 1 in
      let p_target = Node.pred_of_obj target' (truncate (ceil (float_of_int suppmin *. confmin))) in
      let p_target_2 = Node.pred_of_obj target' 2 in
      let p_obj_target_view = (fun node -> Node.view_state_and (p_obj node) (Node.view_state_and (p_target node) (p_view node))) in
      first_concepts
	wq []
	p_obj_target_view
	suppmin
	(fun (ext,nss,_) ->
	  let e = Ext.cardinal ext in
	  let s = Ext.cardinal (Ext.inter target' ext) in
	  let c = conf_gui dom e t s in
	  if e >= suppmin &  c >= confmin
	  then Some (e,c,int_of_nss nss,eqs_of_nss !Root.cache # ext_elt nss), false
	  else None, e > suppmin & conf_gui dom suppmin t s >= confmin
	))

    let rec inds suppmin confmin oid target_view = mutex (fun () ->
      let wq = Ext.remove oid (Query.extent (wq ())) in
      let p_wq = Node.pred_of_obj wq (truncate (ceil (float_of_int suppmin *. confmin))) in
      let p_view = Node.pred_of_view (get_wv ()) in
      let p_obj = Node.pred_of_obj (Ext.singleton oid) 1 in
      let p_view_obj n = Node.view_state_and (Node.view_state_and (p_obj n) (p_wq n)) (p_view n) in
      let p_features = Node.pred_of_view !Node.features in
      let p_target = Node.pred_of_view (Node.get_view target_view) in
      let p_feat_target n = Node.view_state_and (p_target n) (p_features n) in
      let (full, ar_wq), dom = ar_ext_l_from_ext wq in
      let h_t = Hashtbl.create 1000 in
      Node.fold  (* recording in hashtable t, the sizes of ext(wq) inter ext(t) *)
        (!Root.cache # top)
        (fun n () ->
          (if (p_feat_target n).Node.vis
           then Hashtbl.add h_t (Node.id n) (fst (inter_ar_ext (full,ar_wq) (!Root.cache # ext_elt n)))),
          (if (p_feat_target n).Node.more_vis
           then Node.sub n
           else []))
        (fun n () -> ())
        ();
      let h = Hashtbl.create 1000 in
      fold_concepts 
        wq
	p_view_obj
	suppmin
	(fun (ext,nss,_) () ->
          let (full, ar_ext), e = ar_ext_l_from_ext ext in
          let c_opt_max = ref 0.0 in
          let fs =

(*            Node.lubs
             ~pred:(fun node s ->
                if (p_feat_target node).Node.vis & not (p_view_obj node).Node.vis
                then Some (Node.id node, Node.f node, (float_of_int s) /. (float_of_int e))
                else None)
             ~suppmin:(truncate (ceil (float_of_int e *. confmin)))
             (List.map Node.get_obj ext) in
*)
            Node.last
             ~pred_next:(fun node ->
                let id = Node.id node in
                let view_state = p_feat_target node in
                if view_state.Node.vis & not (p_view_obj node).Node.vis then  (* this node is a target feature unknown for oid *)
                   let target = !Root.cache # ext_elt node in
                   let t = Hashtbl.find h_t id in
                   let s, inter = inter_ar_ext (full,ar_ext) target in
                   let c = conf_gui dom e t s in
                   ( if c >= confmin  (* this target feature is induced from current concept *)
                     then Some (id, Node.f node, c) 
                     else begin
                       c_opt_max := max !c_opt_max (conf_gui dom suppmin t s);
                       None end),
                   ( if view_state.Node.more_vis & conf_gui dom e s s >= confmin
                     then Node.sub node
                     else [])
                else None, if view_state.Node.more_vis then Node.sub node else [])
             ~start:(!Root.cache # top) in

          List.iter
            (fun (id,x,c) ->
               try
                 let _x, hyps = Hashtbl.find h id in
                 Hashtbl.replace h id (x, insert_hyp (ext,e,c,int_of_nss nss,eqs_of_nss !Root.cache # ext_elt nss) hyps)
               with Not_found ->
                 Hashtbl.add h id (x, [(ext,e,c,int_of_nss nss,eqs_of_nss !Root.cache # ext_elt nss)]))
            fs,
          !c_opt_max >= confmin
        )
        ();
        Hashtbl.fold (fun id (x,hyps) l -> (x, List.map (fun (_,e,c,int,eqs) -> (e,c,int,eqs)) hyps)::l) h [])
    and insert_hyp (ext,e,c,int,eqs) = function
      | [] -> [(ext,e,c,int,eqs)]
      | (ext',e',c',int',eqs')::l ->
	    if e' >= e & Ext.subset ext ext'
	    then (ext',e',c',int',eqs')::l
	    else if e >= e' & Ext.subset ext' ext
	    then insert_hyp (ext,e,c,int,eqs) l
	    else (ext',e',c',int',eqs')::insert_hyp (ext,e,c,int,eqs) l

****************** *)

(* Debugging operations *)
(* -------------------- *)

(* get some statistics *)
    type stat = {
	mutable nb_obj : int;
	mutable nb_node : int;
	mutable nb_arc : int;
	mutable sum_ext : int;
	mutable sum_ext_words : int; (* in memory words *)
	mutable nb_entails : int; (* suspended *)
      }

    let get_stat () = mutex (fun () ->
      let st = {nb_obj = 0; nb_node = 0; nb_arc = 0; sum_ext = 0; sum_ext_words = 0; nb_entails = 0} in
      st.nb_obj <- Ext.cardinal (!Root.cache # all);
      !Root.cache # fold Logcache.Down
	(fun e _ ->
	  let ext = !Root.cache # ext_elt e in
	  st.nb_node <- st.nb_node + 1;
	  st.nb_arc <- st.nb_arc + (List.length (!Root.cache # children e));
	  st.sum_ext <- st.sum_ext + (Ext.cardinal ext);
	  st.sum_ext_words <- st.sum_ext_words + (Ext.memory_size ext))
	[!Root.cache # top]
	();
      st)

    let print_elt e =
      print_string (Syntax.stringizer (Feat.print (!Root.cache # formula e)))

(* print all nodes, for debugging *)
    let print_nodes () = mutex (fun () ->
      !Root.cache # fold Logcache.Down
	(fun e _ ->
	  print_elt e; print_string " ";
	  Ext.iter (fun oid -> print_int oid; print_string ",") (!Root.cache # ext_elt e);
	  print_newline ())
	[!Root.cache # top]
	())

(* print node refered by a formula *)
    let print_node f = mutex (fun () ->
      try
	let e = !Root.cache # find f in
	print_string "ext = "; Ext.iter (fun oid -> print_int oid; print_string ", ") (!Root.cache # ext_elt e); print_newline ();
	List.iter
	  (fun e' ->
	    print_string "-> ";
	    print_elt e';
	    print_newline ())
	  (!Root.cache # parents e);
	List.iter
	  (fun e' ->
	    print_string "<- ";
	    print_elt e';
	    print_newline ())
	  (!Root.cache # children e)
      with Not_found ->
	print_string (Syntax.stringizer (Feat.print f)); print_endline " does not exists as an elt")


    let parse s = mutex (fun () -> Feat.parse s)

    let parse_descr s = mutex (fun () -> Object.Descr.parse s)

    let parse_update s = mutex (fun () -> Root.Update.parse s)

    let parse_query s = mutex (fun () -> Query.parse s)

    let parse_incr_query s = mutex (fun () -> Query.parse_incr s)

(*
    let parse_view s = mutex (fun () -> Query.parse_view s)

    let parse_incr_view s = mutex (fun () -> Query.parse_incr_view s)
*)
(*    let print_source = Src.print *)

(*    let print_well = Wll.print *)

    let print = Feat.print

    let print_descr = Object.Descr.print

    let print_update = Root.Update.print

    let print_query q = mutex (fun () -> Query.print q)

(*
    let print_view v = mutex (fun () -> Query.print_view v)
*)

    let descr_of_query : Query.t -> Feat.t list =
      fun q -> Object.Descr.of_query q

    let update_of_query : Query.t -> Root.Update.t =
      fun q -> Root.Update.of_query q

    let features_of_query : Query.t -> Query.t list * Query.t list =
      fun q -> Root.Query.features q

(* import and export contexts *)
(* -------------------------- *)

    open Token
    open Syntax
    open Format

    type ctx_option = [`Axioms|`Features|`Sources|`Extrinsic|`Wells|`Rules|`Actions]
    let all_options = [`Axioms; `Features; `Sources; `Extrinsic; `Wells; `Rules; `Actions]

    let export_src_options = Source.print_src_options

    let export_preview = Source.print_preview

    let export_arg_options = Source.print_arg_options

    let export options ctxname = mutex (fun () ->
      Lock.check_file ctxname;
      let dirname =
	let d = Filename.dirname ctxname in
	if d = "."
	then Sys.getcwd ()
	else d in
      let options = if options = [] then all_options else options in
      Common.safe_file_update (ctxname ^ ".ctx")
	(fun tmp ->
	  let ch = open_out_bin tmp in
	  let ff = formatter_of_out_channel ch in
	  (* sources *)
	  if List.mem `Sources options then begin
	    let l =
	      Hashtbl.fold
		(fun url src res -> (url,src.s_options,src.s_home)::res)
		l_src [] in
	    List.iter
	      (fun (url,options,h) ->
		pp_print_string ff (Syntax.stringizer (Ident "src" :: PP_tilda :: export_src_options options @ PP_tilda :: String (Source.rel dirname url) :: PP_tilda :: Object.Descr.print h));
		pp_print_newline ff ())
	      (List.sort compare l) end;
	  (* axioms *)
	  if List.mem `Axioms options then
	    List.iter
	      (fun (f,g,src) ->
		if src=None then begin (* only extrinsic axioms are exported *)
		  pp_print_string ff (Syntax.stringizer (Ident "axiom" :: PP_tilda :: Feat.print f @ Comma :: PP_tilda :: Feat.print g));
		  pp_print_newline ff () end)
	      !l_axiom;
	  (* extrinsic descriptions *)
	  if List.mem `Extrinsic options then
	    Ext.iter
	      (fun oid ->
		match Object.get_origin oid with
		| Object.Internal (pre,args) ->
		    pp_print_string ff (Syntax.stringizer (Ident "mk" :: export_arg_options args @ PP_tilda :: export_preview dirname pre @ PP_tilda :: Object.Descr.print (Object.get_extr oid)));
		    pp_print_newline ff ()
		| Object.External (src,r) ->
		    let e = Object.get_extr oid in
		    if e <> Object.Descr.empty
		    then begin
		      pp_print_string ff (Syntax.stringizer (Ident "mv" :: PP_tilda :: String (Source.rel dirname (Src.url src)) :: PP_tilda :: String r :: PP_tilda :: Object.Descr.print e));
		      pp_print_newline ff ()
		    end)
	      (extent ());
	  (* rules and holes *)
	  if List.mem `Rules options then
	    List.iter
	      (fun rule ->
		match query_rule rule, update_rule rule with
		| (q,Intr u) ->
		    pp_print_string ff (Syntax.stringizer (Ident "rule_intr" :: PP_tilda :: Query.print q @ PP_tilda :: Root.Update.print u));
		    pp_print_newline ff ()
		| (q,Extr u) ->
		    pp_print_string ff (Syntax.stringizer (Ident "rule_extr" :: PP_tilda :: Query.print q @ PP_tilda :: Root.Update.print u));
		    pp_print_newline ff ()
		| (q,Rem) ->
		    pp_print_string ff (Syntax.stringizer (Ident "rule_rem" :: PP_tilda :: Query.print q));
		    pp_print_newline ff ())
	      (get_rules ());
	  (* features *)
	  if List.mem `Features options then begin
	    List.iter
	      (fun e ->
		pp_print_string ff (Syntax.stringizer (Ident "show" :: PP_tilda :: Feat.print (!Root.cache # formula e)));
		pp_print_newline ff ())
	      !Root.cache#shown;
	    List.iter
	      (fun f ->
		pp_print_string ff (Syntax.stringizer (Ident "hide" :: PP_tilda :: Feat.print f));
		pp_print_newline ff ())
	      !Root.cache#hidden
	  end;
	  (* wells *)
	  if List.mem `Wells options then begin
	    Hashtbl.iter
	      (fun url (w,rule) ->
		pp_print_string ff (Syntax.stringizer (Ident "wll" :: PP_tilda :: String (Source.rel dirname url) :: (* Wll.print w @ *) PP_tilda :: Query.print rule.Root.Rules.prop));
		pp_print_newline ff ())
	      l_wll end;
	  if List.mem `Actions options then begin
	    List.iter
	      (fun ac ->
		pp_print_string ff (Syntax.stringizer (Ident "action" :: PP_tilda :: String ac.action_name :: PP_tilda :: String ac.action_cmd :: PP_tilda :: Query.print ac.action_query));
		pp_print_newline ff ())
	      !l_actions end;
	  close_out ch))

    let rec import options ctxname = mutex (fun () ->
      Lock.check ();
      let dirname =
	let d = Filename.dirname ctxname in
	if d = "."
	then Sys.getcwd ()
	else d in
      let options = if options = [] then all_options else options in
      let ch = open_in_bin (ctxname ^ ".ctx") in
      let axioms, features, sources, extrinsic, wells, rules, actions =
	let b x = List.mem x options in
	b `Axioms, b `Features, b `Sources, b `Extrinsic, b `Wells, b `Rules, b `Actions in
      try while true do
	let line = input_line ch in
	try
	  match Syntax.from_string line with parser
	  | [<'Ident "set">] -> () (* handled in gui.ml *)
	  | [<'Ident "include"; 'String url>] ->
	      import options (Filename.chop_extension (Source.abs dirname url))
	  | [<'Ident "axiom"; f = Feat.parse; str>] ->
	      if axioms then
		( match str with parser
		| [<'Comma; g = Feat.parse>] ->
		    push_updates [Axiom(Feat.print f, Feat.print g, "")]
		| [<g = Feat.parse>] -> (* for backward compatibility with v1.1 *)
		    push_updates [Axiom(Feat.print f, Feat.print g, "")]
		| [<>] -> (* for backward compatibility with v1.1*)
		    match Syntax.of_list (Feat.print f) with parser
		    | [<'Term a; 'Term b>] -> 
			push_updates [Axiom([Term a], [Term b], "")]
		    | [<>] -> failwith "Context.import: invalid axiom")
	  | [<'Ident "show"; f = Feat.parse>] ->
	      if features then push_updates [Show(Feat.print f)]
	  | [<'Ident "hide"; f = Feat.parse>] ->
	      if features then push_updates [Hide(Feat.print f)]
	  | [<'Ident "src"; options = Source.parse_src_options; 'String url; str>] ->
	      let h =
		match str with parser
		| [<d = Object.Descr.parse>] -> d
		| [<u = Root.Update.parse>] -> Root.Update.to_descr u in (* backward compatibility *)
	      if sources then init_source (Source.abs dirname url) options h  (* don't use Source.Init, because it is for sources generated by another source *)
	  | [<'Ident "mk"; str>] ->
	      if extrinsic then
		let pre, args, d =
		  match str with parser
		  | [<args = Source.parse_action_args; pre = Source.parse_preview dirname; d = Object.Descr.parse>] -> pre, args, d
		  | [<d = Object.Descr.parse>] -> Source.String "", [], d in
		push_updates [MakeExtr(pre,args,Object.Descr.print d)]
	  | [<'Ident "mv"; 'String url (* src = Src.parse *); 'String r; d = Object.Descr.parse>] ->
	      if extrinsic then push_updates [MoveExtrSrc(Source.abs dirname url,r,Root.Update.print (Root.Update.of_descr d))]
	  | [<'Ident "wll"; 'String url (*w = Wll.parse*); q = Query.parse>] ->
	      if wells then push_updates [InitWll(Source.abs dirname url,Query.print q)]
	  | [<'Ident "rule_intr"; p = Query.parse; c = Root.Update.parse>] ->
	      if rules then push_updates [InitRuleIntr(Query.print p,Root.Update.print c)]
	  | [<'Ident "rule_extr"; p = Query.parse; c = Root.Update.parse>] ->
	      if rules then push_updates [InitRuleExtr(Query.print p,Root.Update.print c)]
	  | [<'Ident "rule_rem"; p = Query.parse>] ->
	      if rules then push_updates [InitRuleRem(Query.print p)]
	  | [<'Ident "action"; 'String name; 'String cmd; query = Query.parse>] ->
	      if actions then push_updates [Action(name,cmd,Query.print query)]
	  | [<'Percent>] -> () (* comment line *)
	  | [<_ = Stream.empty>] -> () (* empty line *)
	with
	| End_of_file -> raise End_of_file
	| e ->
	    print_string "In line: "; print_endline line;
	    print_endline (Printexc.to_string e)
      done
      with _ -> ())

(* export as binary formal context *)
(* ------------------------------- *)

    let string_of_oid oid =
      let pre =
	match Object.get_origin oid with
	| Object.Internal (pre,args) -> pre
	| Object.External (s,r) -> Src.preview s r in
      match pre with
      | Source.String s -> s
      | Source.Picture (_,s) -> s

    let export_as_binary_csv oids props filename = mutex (fun () ->
      let ff = formatter_of_out_channel (open_out filename) in
      (* line of properties *)
      List.iter
	(fun prop -> pp_print_string ff "; "; pp_print_string ff (Syntax.stringizer (Query.print prop)))
	props;
      pp_print_newline ff ();
      (* a line for each object *)
      let exts = List.map Query.extent props in (* the extents of props *)
      Ext.iter
	(fun oid ->
	  pp_print_string ff (string_of_oid oid);
	  List.iter
	    (fun ext ->
	      if Ext.mem oid ext
	      then pp_print_string ff ";1"
	      else pp_print_string ff ";0")
	    exts;
	  pp_print_newline ff ())
	oids)

    let export_as_binary_cxt oids props filename = mutex (fun () ->
      let ff = formatter_of_out_channel (open_out filename) in
      (* head of file *)
      pp_print_string ff "B\n";
      pp_print_newline ff ();
      pp_print_int ff (Ext.cardinal oids); pp_print_newline ff ();
      pp_print_int ff (List.length props); pp_print_newline ff ();
      pp_print_newline ff ();
      (* list of object names *)
      Ext.iter
	(fun oid -> pp_print_string ff (string_of_oid oid); pp_print_newline ff ())
	oids;
      (* list of attributes *)
      List.iter
	(fun prop -> pp_print_string ff (Syntax.stringizer (Query.print prop)); pp_print_newline ff ())
	props;
      (* incidence relation *)
      let exts = List.map Query.extent props in (* the extents of props *)
      Ext.iter
	(fun oid ->
	  List.iter
	    (fun ext ->
	      if Ext.mem oid ext
	      then pp_print_string ff "X"
	      else pp_print_string ff ".")
	    exts;
	  pp_print_newline ff ())
	oids)

    let export_as_binary = export_as_binary_cxt

(* load and save *)
(* ------------- *)

    type data =
	Object.data *
	  (Feat.t * Feat.t * Src.t option) list *
	  monitor *
	  (Source.url, src) Hashtbl.t *
	  (Well.url, wll) Hashtbl.t *
	  action list *
	  SetUpdates.t

    let init ?(log : bool = false) filename = mutex (fun () ->
      Object.init ();
      Threads_common.owner := None;
      Stack.clear wq_histo;
      l_axiom := [];
      init_monitor ();
      Hashtbl.clear l_src;
      Hashtbl.clear l_wll;
      l_actions := [];
      set_updates := SetUpdates.empty;
      Gc.major ();
      Lock.set filename false log)

    let lis_version = "lis-1.4"
	(* this string is put in .lis files for checking compatibility with Camelis version. *)

    let load ?(readonly : bool = false) filename = mutex (fun () ->
      let ch = open_in_bin (filename ^ ".lis") in
      let v = (input_value ch : string) in
      if v <> lis_version
      then failwith "The version of the .lis file does not match this version of Camelis."
      else begin
	let (d, l_ax, m, l_s, l_w, l_ac, s_up) = (input_value ch : data) in
	Object.load d;
	Threads_common.owner := None;
	Stack.clear wq_histo;
	l_axiom := l_ax;
	load_monitor m;
	Hashtbl.clear l_src;
	Hashtbl.iter (fun url src -> Hashtbl.add l_src url src) l_s;
	Hashtbl.clear l_wll;
	Hashtbl.iter (fun url wll -> Hashtbl.add l_wll url wll) l_w;
	l_actions := l_ac;
	set_updates := s_up;
	close_in ch;
	Gc.major ();
	Lock.set filename readonly (Sys.file_exists (Lock.logname filename))
      end)

    let save filename = mutex (fun () ->
      Lock.check_file filename;
      (* saving as .ctx *)
      wq_incr (fun _ -> Query.top); (* in order to save everything with export *)
      export [] filename;
      wq_back (); (* restore current query *)
      (* saving as .lis *)
      Common.safe_file_update (filename ^ ".lis")
	(fun tmp ->
	  let ch = open_out_bin tmp in
	  output_value ch (lis_version : string);
	  let data : data = (Object.save (), !l_axiom, monitor, l_src, l_wll, !l_actions, !set_updates) in
	  output_value ch data;
	  close_out ch))

  end
