module type EXT = Intset.T
(*
  sig
    type t
    val empty : t
    val is_empty : t -> bool
    val cardinal : t -> int
    val mem : int -> t -> bool
    val singleton : int -> t
    val add : int -> t -> t
    val remove : int -> t -> t
    val subset : t -> t -> bool
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val union_r : t list -> t
    val inter_r : t list -> t
    val fold : ('a -> int -> 'a) -> 'a -> t -> 'a
    val iter : (int -> unit) -> t -> unit
    val map : (int -> 'a) -> t -> 'a list
    val filter : (int -> bool) -> t -> t
    val elements : t -> int LSet.t
    val memory_size : t -> int (* in words *)
  end
*)

module Ext : EXT =
  (* Intset.Cis *)
  Intset.Bitmap (Intset.Bitmap (Intset.Bitmap (Intset.Bitmap31)))
    (* bound is 923,521 objects *)

(* Cis implementation of extents *)
(*
  struct
    type t = Cis.t
    let empty = Cis.empty
    let is_empty = Cis.is_empty
    let cardinal = Cis.cardinal
    let mem = Cis.mem
    let singleton = Cis.singleton
    let add = Cis.add
    let remove = Cis.remove
    let subset = Cis.subset
    let union = Cis.union
    let inter = Cis.inter
    let diff = Cis.diff
    let union_r l = List.fold_left (fun res set -> union res set) empty l
    let inter_r = function
      | [] -> raise (Invalid_argument "Node.Ext.inter_r : empty list of sets")
      | set::sets -> List.fold_right (fun set res -> inter set res) sets set
    let fold = Cis.fold_left
    let iter = Cis.iter
    let map f ext = Cis.fold_left (fun res x -> f x::res) [] ext
    let filter p ext = List.fold_left (fun res x -> if p x then Cis.add x res else res) Cis.empty (Cis.elements ext)
    let elements = Cis.elements
    let memory_size = Cis.memory_size
  end
*)

exception No_more_id

module Genid =
  struct
    type t = {
	mutable cpt : int;
	mutable all_free : int list;
      }

    let create () = { cpt = 0; all_free = []}

    let alloc g =
      match g.all_free with
      | [] ->
	  if g.cpt < max_int
	  then begin g.cpt <- g.cpt+1; g.cpt end
	  else raise No_more_id
      | id::l ->
	  g.all_free <- l;
	  id

    let free g id =
      g.all_free <- id::g.all_free

    let valid g id =
      id <= g.cpt && id > 0 && not (List.mem id g.all_free)
  end

type oid = int      (* object id *)
type oidpart = int  (* object or part id (objects are also parts *)
type part = int     (* strict part ids *)

type node = {
    mutable parts : Ext.t;
    whole : oidpart;
    oid : oid
  }
type level = {
    genid : Genid.t;
    nodes : (oidpart, node) Hashtbl.t;
  }

type kernel_data = Ext.t * (int, level) Hashtbl.t

class kernel =
  object (self)
    val mutable v_oids : Ext.t = Ext.empty (* extent of all oids *)
    val mutable levels : (int, level) Hashtbl.t = Hashtbl.create 10

    method init =
      v_oids <- Ext.empty;
      levels <- Hashtbl.create 10

    method save = (v_oids, levels)

    method load (v, l) =
      v_oids <- v;
      levels <- l

    method private level l =
      try Hashtbl.find levels l
      with Not_found ->
	let level = {
	  genid = Genid.create ();
	  nodes = Hashtbl.create 100;
	} in
	Hashtbl.add levels l level;
	level

    method private node l op =
      let level = self # level l in
      try Hashtbl.find level.nodes op
      with Not_found -> (* happens only at level 0 *)
	let node = {
	  parts = Ext.empty;
	  whole = 0;
	  oid = op;
	} in
	Hashtbl.add level.nodes op node;
	node

    method private find l (op : oidpart) =
      let level = self # level l in
      Hashtbl.find level.nodes op

    method oids : Ext.t = v_oids

    method parts l (op : oidpart) : Ext.t =
      try (self # find l op).parts
      with Not_found -> Ext.empty

    method whole l (p : part) : oidpart = (* raise Not_found if [p] does not exist in level [l] *)
      if l = 0
      then raise Not_found (* should not be used this way *)
      else (self # find l p).whole

    method oid l (op : oidpart) : oid = (* raise Not_found if [op] does not exist in level [l] *)
      if l = 0
      then op
      else (self # find l op).oid

    method ext l ext_op : Ext.t = (* raise Not_found if some element of [ext_op] does not exist in level [l] *)
      if l = 0
      then ext_op
      else
	let level = self # level l in
	List.fold_left
	  (fun res op ->
	    let o = (Hashtbl.find level.nodes op).oid in
	    Ext.add o res)
	  Ext.empty
	  (Ext.elements ext_op)

    method alloc_oid : oid =
      let level0 = self # level 0 in
      let oid = Genid.alloc level0.genid in
      v_oids <- Ext.add oid v_oids;
      oid

    method free_oid (oid : oid) : unit =
      self # free_part 0 oid

    method valid_oid (oid : oid) : bool =
      Ext.mem oid v_oids

    method alloc_part l (op : oidpart) : part =
      let node = self # node l op in
      let level1 = self # level (l+1) in
      let p = Genid.alloc level1.genid in
      node.parts <- Ext.add p node.parts;
      Hashtbl.add level1.nodes p
	{ parts = Ext.empty;
	  whole = op;
	  oid = self # oid l op;
	};
      p

    method free_part l (op : oidpart) : unit =
      let level = self # level l in
      (try
	let node = Hashtbl.find level.nodes op in
	Ext.iter (fun p -> self # free_part (l+1) p) node.parts;
	if l > 0 then
	  let whole_node = self # find (l-1) node.whole in
	  whole_node.parts <- Ext.remove op whole_node.parts;
	  Hashtbl.remove level.nodes op
      with Not_found -> ());
      Genid.free level.genid op;
      if l = 0 then v_oids <- Ext.remove op v_oids

  end


type fold_mode = Down | Up

  (* invariants
     - one must ensure that no bottom-formula is present in cache, ever (simplifies subsumption, and so ideal, filter, insertion)
     - e in (max) iff (parents e) = []
     - forall e,e' in (links oid): e and e' are incomparable (no subsumption relation)
     - link should apply only on descriptors
     - object descriptors are always present as elt
     - int is not supposed to be applied on empty extents
   *)

(* for logics as classes *)
(*
class type ['formula] logic =
  object
    method compare : 'formula -> 'formula -> int
    method desc : 'formula -> bool
    method feat : 'formula -> bool
    method entails : 'formula -> 'formula -> bool
    method add : 'formula -> 'formula -> 'formula
    method sub : 'formula -> 'formula -> 'formula
    method parse : Syntax.t_stream -> 'formula
    method print : 'formula -> Syntax.t_list
  end

module LogicAsClass (L : Logic.T) =
  struct
    type t = L.t

    class clogic : [L.t] logic =
      object
	method compare = L.compare
	method desc = L.desc
	method feat = L.feat
	method entails = L.entails
	method add = L.add
	method sub = L.sub
	method parse = L.parse
	method print = L.print
      end
  end
*)

class type ['formula,'elt,'data] common =
  object
(*    inherit ['formula] logic*)

    method save : 'data (* out_channel -> unit *)
    method load : 'data -> unit  (* in_channel -> unit *)

    method formula : 'elt -> 'formula
    method mem : 'formula -> bool
    method find : 'formula -> 'elt (* raise Not_found *)

    method axiom : 'formula -> 'formula option -> unit
    method show : 'formula -> unit
    method hide : 'elt -> unit
    method shown : 'elt list
    method hidden : 'formula list

    method top : 'elt (* raise Not_found *)
    method max : 'elt list
    method children : 'elt -> 'elt list
    method parents : 'elt -> 'elt list
    method ideal : 'elt -> ('elt -> bool)
    method filter : 'elt -> ('elt -> bool)
    method fold : 'a . fold_mode -> ('elt -> 'a -> 'a) -> 'elt list -> 'a -> 'a

    method ext : 'formula -> Ext.t
    method ext_elt : 'elt -> Ext.t (* room for more efficient implementation *)
    method int : Ext.t -> 'elt list (* raise Not_found *)
    method increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list

    method forward : 'formula -> Ext.t -> Ext.t
    method forward_elt : 'elt -> Ext.t -> Ext.t
    method forward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list

    method backward : 'formula -> Ext.t -> Ext.t
    method backward_elt : 'elt -> Ext.t -> Ext.t
    method backward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list
  end

class type ['formula,'elt,'data] logcache =
  object
    inherit ['formula,'elt,'data] common

    method links : oidpart -> 'elt list
    method link : oidpart -> 'formula -> 'elt (* raise Not_found when linking is not defined, or has no effect *)
    method unlink : oidpart -> 'formula -> unit
    method unlink_elt : oidpart -> 'elt -> unit (* room for more efficient implementation *)
    method replace : oidpart -> 'formula -> 'formula -> unit
    method remove : oidpart -> unit

    method extpart : 'formula -> Ext.t
    method extpart_elt : 'elt -> Ext.t (* room for more efficient implementation *)

    method map : 'formula -> oidpart -> oidpart -> 'elt (* raise Not_found when undefined *)
    method unmap : 'formula -> oidpart -> oidpart -> unit
    method unmap_elt : 'elt -> oidpart -> oidpart -> unit

    method forwardpart : 'formula -> Ext.t -> Ext.t (* raise Not_found when undefined *)
    method forwardpart_elt : 'elt -> Ext.t -> Ext.t

    method backwardpart : 'formula -> Ext.t -> Ext.t (* raise Not_found when undefined *)
    method backwardpart_elt : 'elt -> Ext.t -> Ext.t
  end


class virtual ['formula,'elt,'data] default =
  fun kernel l -> object (self)
(*
    method mem : 'formula -> bool = fun f -> false
    method find : 'formula -> 'elt = fun f -> raise Not_found
*)
    method virtual formula : 'elt -> 'formula

    method show : 'formula -> unit = fun f -> ()
    method hide : 'elt -> unit = fun e -> ()
    method shown : 'elt list = []
    method hidden : 'formula list = []

    method top : 'elt = raise Not_found
    method max : 'elt list = []
    method children : 'elt -> 'elt list = fun e -> []
    method parents : 'elt -> 'elt list = fun e -> []
(*
    method ideal : 'elt -> ('elt -> bool) = fun e1 -> (fun e2 -> false)
    method filter : 'elt -> ('elt -> bool) = fun e1 -> (fun e2 -> false)
*)
    method virtual unlink : oidpart -> 'formula -> unit
    method unlink_elt : oidpart -> 'elt -> unit = fun oid e -> self # unlink oid (self # formula e)
    method replace : oidpart -> 'formula -> 'formula -> unit = fun oid f f' -> ()
    method remove : oidpart -> unit = fun oid -> ()
    method axiom : 'formula -> 'formula option -> unit = fun f g_opt -> ()

    method virtual extpart : 'formula -> Ext.t
    method extpart_elt : 'elt -> Ext.t = fun e -> self # extpart (self # formula e)
    method ext : 'formula -> Ext.t = fun f -> kernel # ext l (self # extpart f)
    method ext_elt : 'elt -> Ext.t = fun e -> kernel # ext l (self # extpart_elt e)
(*    method int : Ext.t -> 'elt list = fun ext -> [] *)
(*    method increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list = fun ~suppmin ext start -> [] *)

    method map : 'formula -> oidpart -> oidpart -> 'elt = fun f oid1 oid2 -> raise Not_found
    method unmap : 'formula -> oidpart -> oidpart -> unit = fun f oid1 oid2 -> ()
    method unmap_elt : 'elt -> oidpart -> oidpart -> unit = fun e oid1 oid2 -> self # unmap (self # formula e) oid1 oid2

    method forwardpart : 'formula -> Ext.t -> Ext.t = fun f oids -> raise Not_found
    method forwardpart_elt : 'elt -> Ext.t -> Ext.t = fun e oids -> self # forwardpart (self # formula e) oids
    method forward : 'formula -> Ext.t -> Ext.t = fun f oids -> kernel # ext l (self # forwardpart f oids)
    method forward_elt : 'elt -> Ext.t -> Ext.t = fun e oids -> kernel # ext l (self # forwardpart_elt e oids)
    method forward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list = fun ~suppmin ext start -> []

    method backwardpart : 'formula -> Ext.t -> Ext.t = fun f oids -> raise Not_found
    method backwardpart_elt : 'elt -> Ext.t -> Ext.t = fun e oids -> self # backwardpart (self # formula e) oids
    method backward : 'formula -> Ext.t -> Ext.t = fun f oids -> kernel # ext l (self # backwardpart f oids)
    method backward_elt : 'elt -> Ext.t -> Ext.t = fun e oids -> kernel # ext l (self # backwardpart_elt e oids)
    method backward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list = fun ~suppmin ext start -> []
  end


module type T =
  sig
    include Logic.T
	(** module of the logic on which the cache is based *)

    type elt
	  (** type of the cache elements, i.e., visible formulas *)

    type data
	  (** type of all persistent internal data structures *)

    class c : kernel -> int -> [t,elt,data] logcache
	  (** class handling logic caches, where the parameter if the level *)
  end

(** The functor identity. *)
module Id (X : T) =
  struct
  (* manual including of (Logic.Id X), does not work with 'include' *)

    let props =
      Logic.fixpoint
	(fun () ->
	  let l = X.props () in
	  l)

    type t = X.t

    let desc f = X.desc f
    let feat f = X.feat f
    let role f = X.role f
    let isvalue f = X.isvalue f
    let top () = X.top ()
    let bot () = X.bot ()
    let entails f g = X.entails f g
    let share f g = X.share f g
    let conj f g = X.conj f g
    let disj f g = X.disj f g
    let le_min f g = X.le_min f g
    let le_max f g = X.le_max f g
    let add f g = X.add f g
    let sub f g = X.sub f g
    let features f = X.features f
    let terms f = X.terms f
    let axiom f g_opt = X.axiom f g_opt
    let gen f g hs = X.gen f g hs
    let parse str = X.parse str
    let print f = X.print f
    let simpl f = X.simpl f
    let compare f g = X.compare f g
    let parse_compact str = X.parse_compact str

    type elt = X.elt

    type data = X.data

    class c kernel l =
      object (self)
	val c = new X.c kernel l

	method save : data = c # save
	method load : data -> unit = fun data -> c # load data

	method formula : 'elt -> 'formula = fun e -> c # formula e
	method mem : 'formula -> bool = fun f -> c # mem f
	method find : 'formula -> 'elt = fun f -> c # find f
	    
	method axiom : 'formula -> 'formula option -> unit = fun f g_opt -> c # axiom f g_opt
	method show : 'formula -> unit = fun f -> c # show f
	method hide : 'elt -> unit = fun e -> c # hide e
	method shown : 'elt list = c # shown
	method hidden : 'formula list = c # hidden

	method top : 'elt = c # top
	method max : 'elt list = c # max
	method children : 'elt -> 'elt list = fun e -> c # children e
	method parents : 'elt -> 'elt list = fun e -> c # parents e
	method ideal : 'elt -> ('elt -> bool) = fun e -> c # ideal e
	method filter : 'elt -> ('elt -> bool) = fun e -> c # filter e
	method fold : 'a . fold_mode -> ('elt -> 'a -> 'a) -> 'elt list -> 'a -> 'a = fun mode f l e -> c # fold mode f l e
	    
	method links : oidpart -> 'elt list = fun op -> c # links op
	method link : oidpart -> 'formula -> 'elt = fun op d -> c # link op d
	method unlink : oidpart -> 'formula -> unit = fun op f -> c # unlink op f
	method unlink_elt : oidpart -> 'elt -> unit = fun op e -> c # unlink_elt op e
	method replace : oidpart -> 'formula -> 'formula -> unit = fun op f f' -> c # replace op f f'
	method remove : oidpart -> unit = fun op -> c # remove op
	    
	method extpart : 'formula -> Ext.t = fun f -> c # extpart f
	method extpart_elt : 'elt -> Ext.t = fun e -> c # extpart_elt e
	method ext : 'formula -> Ext.t = fun f -> c # ext f
	method ext_elt : 'elt -> Ext.t = fun e -> c # ext_elt e
	method int : Ext.t -> 'elt list = fun ext -> c # int ext
	method increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list =
	  fun ~suppmin ext start -> c # increments ~suppmin ext start

	method map : 'formula -> oidpart -> oidpart -> 'elt = fun d op1 op2 -> c # map d op1 op2
	method unmap : 'formula -> oidpart -> oidpart -> unit = fun f op1 op2 -> c # unmap f op1 op2
	method unmap_elt : 'elt -> oidpart -> oidpart -> unit = fun e op1 op2 -> c # unmap_elt e op1 op2

	method forwardpart : 'formula -> Ext.t -> Ext.t = fun f -> c # forwardpart f
	method forwardpart_elt : 'elt -> Ext.t -> Ext.t = fun e -> c # forwardpart_elt e
	method forward : 'formula -> Ext.t -> Ext.t = fun f -> c # forward f
	method forward_elt : 'elt -> Ext.t -> Ext.t = fun e -> c # forward_elt e
	method forward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list =
	  fun ~suppmin ext start -> c # forward_increments ~suppmin ext start

	method backwardpart : 'formula -> Ext.t -> Ext.t = fun f -> c # backwardpart f
	method backwardpart_elt : 'elt -> Ext.t -> Ext.t = fun e -> c # backwardpart_elt e
	method backward : 'formula -> Ext.t -> Ext.t = fun f -> c # backward f
	method backward_elt : 'elt -> Ext.t -> Ext.t = fun e -> c # backward_elt e
	method backward_increments : suppmin:int -> Ext.t option -> 'elt option -> (int * Ext.t * 'elt) list =
	  fun ~suppmin ext start -> c # backward_increments ~suppmin ext start
      end
  end


(* Root functor for logic caches, in order to switch between oid and oidpart *)
class type ['formula,'elt,'data] root =
  object
    inherit ['formula,'elt,'data] common

    method alloc : oid
    method free : oid -> unit
    method valid : oid -> bool
    method all : Ext.t

    method links : oid -> 'elt list
    method link : oid -> 'formula -> 'elt
    method unlink : oid -> 'formula -> unit
    method unlink_elt : oid -> 'elt -> unit
    method replace : oid -> 'formula -> 'formula -> unit
    method remove : oid -> unit

    method map : 'formula -> oid -> oid -> 'elt
    method unmap : 'formula -> oid -> oid -> unit
    method unmap_elt : 'elt -> oid -> oid -> unit
  end

module type ROOT =
  sig
    include Logic.T

    type elt

    type data

    class c : [t,elt,data] root
  end

module Root (C : T) : ROOT =
  struct
    include (C : Logic.T with type t = C.t)

    type elt = C.elt

    type data = kernel_data * C.data

    class c =
      let k0 = new kernel in
      object (self)
	val k = k0
	val c = new C.c k0 0

	method save = (k # save, c # save)

	method load (dk, dc) =
	  k # load dk;
	  c # load dc

	method alloc = k # alloc_oid
	method free = k # free_oid
	method valid = k # valid_oid
	method all = k # oids

	method formula = c # formula
	method mem = c # mem
	method find = c # find

	method show = c # show
	method hide = c # hide
	method shown = c # shown
	method hidden = c # hidden

	method top = c # top
	method max = c # max
	method children = c # children
	method parents = c # parents
	method ideal = c # ideal
	method filter = c # filter
	method fold : 'a . fold_mode -> (elt -> 'a -> 'a) -> elt list -> 'a -> 'a = c # fold

	method links = c # links
	method link = c # link
	method unlink = c # unlink
	method unlink_elt = c # unlink_elt
	method replace = c # replace
	method remove = c # remove
	method axiom = c # axiom

	method ext = c # ext
	method ext_elt = c # ext_elt
	method int = c # int
	method increments = c # increments

	method map = c # map
	method unmap = c # unmap
	method unmap_elt = c # unmap_elt

	method forward = c # forward
	method forward_elt = c # forward_elt
	method forward_increments = c # forward_increments

	method backward = c # backward
	method backward_elt = c # backward_elt
	method backward_increments = c # backward_increments
      end
  end


open Token

module Tester (C : T) =
  struct
    module R = Root (C)

    let rec parse_ext = parser
      | [<'LeftAcc; ext = parse_ext_suite>] -> ext
    and parse_ext_suite = parser
      | [<'RightAcc>] -> Ext.empty
      | [<'Nat oid; ext = parse_ext_suite>] -> Ext.add oid ext
      | [<'Comma; 'Nat oid; ext = parse_ext_suite>] -> Ext.add oid ext

    let print_ext ext =
      print_string "{"; Ext.iter (fun oid -> print_int oid; print_string " ") ext; print_endline "}"

    let print_bool b =
      print_endline (string_of_bool b)


    let main () =
      let c = new R.c in
      let parse_f = R.parse in
      let print_f f = print_endline (Syntax.stringizer (R.print f)) in
      let print_fl fl = List.iter print_f fl in
      let parse_e = parser [<f = parse_f>] -> try c # find f with Not_found -> failwith "the element does not exist" in
      let print_e e = print_f (c # formula e) in
      let print_el el = List.iter print_e el in

      while true do
	print_string "> ";
	let cmd = read_line () in
	try match Syntax.from_string cmd with parser
	| [<'Ident "exit">] -> exit 0

	| [<'Ident "formula"; e = parse_e>] -> print_f (c # formula e)
	| [<'Ident "mem"; f = parse_f>] ->  print_bool (c # mem f)
	| [<'Ident "find"; f = parse_f>] -> print_e (c # find f)

	| [<'Ident "show"; f = parse_f>] -> c # show f
	| [<'Ident "hide"; e = parse_e>] -> c # hide e
	| [<'Ident "shown">] -> print_el (c # shown)
	| [<'Ident "hidden">] -> print_fl (c # hidden)

	| [<'Ident "top">] -> print_e (c # top)
	| [<'Ident "children"; e = parse_e>] -> print_el (c # children e)
	| [<'Ident "parents"; e = parse_e>] -> print_el (c # parents e)

	| [<'Ident "links"; 'Nat oid>] -> print_el (c # links oid)
	| [<'Ident "link"; 'Nat oid; f = parse_f>] -> print_e (c # link oid f)
	| [<'Ident "unlink"; 'Nat oid; f = parse_f>] -> c # unlink oid f
	| [<'Ident "replace"; 'Nat oid; f = parse_f; f' = parse_f>] -> c # replace oid f f'
	| [<'Ident "remove"; 'Nat oid>] -> c # remove oid
	| [<'Ident "axiom"; f = parse_f; str>] ->
	    ( match str with parser
	    | [<g = parse_f>] -> c # axiom f (Some g)
	    | [<>] -> c # axiom f None)

	| [<'Ident "ext"; f = parse_f>] -> print_ext (c # ext f)
	| [<'Ident "int"; ext = parse_ext>] -> print_el (c # int ext)
	| [<'Ident "increments"; 'Nat suppmin; 
	    ext = Syntax.parse_opt (parser [<ext = parse_ext>] -> Some ext) None;
	    start = Syntax.parse_opt (parser [<e = parse_e>] -> Some e) None>] ->
	      let l = c # increments ~suppmin ext start in
	      List.iter (fun (i,_,e) -> print_int i; print_string "\t"; print_e e) l

	| [<'Ident "map"; f = parse_f; 'Nat oid1; 'Nat oid2>] -> print_e (c # map f oid1 oid2)
	| [<'Ident "unmap"; f = parse_f; 'Nat oid1; 'Nat oid2>] -> c # unmap f oid1 oid2

	| [<'Ident "forward"; f = parse_f; ext = parse_ext>] -> print_ext (c # forward f ext)
	| [<'Ident "backward"; f = parse_f; ext = parse_ext>] -> print_ext (c # backward f ext)

	| [<>] -> print_endline "Syntax error"
	with
	| Stream.Error msg -> print_endline msg
	| Not_found -> print_endline "undefined"
(*	| exn -> print_endline (Printexc.to_string exn) *)
      done

  end
