(** Graphical interface for Camelis. *)

module Make (Descr : Logcache.T) (Src : Source.T) (Wll : Well.T) =
  struct
    open Format
    open StdLabels
    open GdkKeysyms
    open Gobject.Data

    module Context = Context.Make(Descr)(Src)(Wll)
    module Ext = Context.Ext
    module Query = Context.Root.Query

(* command line arguments *)

    let readonly = ref false
    let log = ref false
    let lis_file = ref ""
    let ctx_files = ref []

(* utilities *)

    let do_while_idle : (unit -> unit) Queue.t -> Glib.Idle.id =
      fun q ->
	Glib.Idle.add (fun () ->
          try Queue.take q () ; true
          with Queue.Empty -> false)

    let utf0 s = s (* the conversion is now performed in source/Fields *)

    let utf s =
      if Glib.Utf8.validate s
      then s
      else try Glib.Convert.locale_to_utf8 s with _ -> s

    let loc0 s = s

    let loc s =
      try Glib.Convert.locale_from_utf8 s with _ -> s


    let file_of_string : string -> string =
      fun name ->
	if Filename.is_implicit name
	then
	  try Filename.concat (Sys.getenv "CAMELISDATA") name
	  with Not_found -> name
	else name

    let parse_of_string p s =
      let str = Syntax.from_string s in
      try
	let x = p str in
	try
	  let _ = Stream.empty str in
	  x
	with Stream.Failure -> raise (Stream.Error "Syntax error: unexpected end")
      with Stream.Failure -> raise (Stream.Error "Syntax error: unexpected beginning")

(* CAUSE: camlp4 interferes with the parsing of {< ...; ... >} expressions
      match Syntax.from_string s with parser
      | [<x = p; _ = Stream.empty ?? "Syntax error: unexpected end">] -> x
      | [<>] -> raise (Stream.Error "Syntax error: unexpected beginning")
*)

    let log_of_string = parse_of_string Context.parse

    let query_of_string = parse_of_string Context.parse_query

    let update_of_string = parse_of_string Context.parse_update

    let string_of_feat f =
      Syntax.stringizer (Context.print f)

    let string_of_descr d =
      Syntax.stringizer (Context.Object.Descr.print d)

    let string_of_update u =
      Syntax.stringizer (Context.Root.Update.print u)

    let string_of_query q =
      Syntax.stringizer (Context.print_query q)

    let string_of_feats : Context.Feat.t list -> string =
      function
	| [] -> ""
	| x::xs ->
	    let s0 = string_of_feat x in
	    List.fold_left ~f:(fun s x -> s ^ "/" ^ string_of_feat x) ~init:s0 xs

    let descr_of_oid oid =
      match Context.origin_obj oid with
      | Context.Object.Internal (pre,args) -> pre
      | Context.Object.External (s,r) -> Context.Src.preview s r

    let contents_of_extent ext =
      Context.Ext.map Context.action_args ext

    let dir_sep =
      match Sys.os_type with
      | "Unix" -> "/"
      | "Cygwin" -> "/"
      | "Win32" -> "\\"
      | _ -> "/"

    let camelisdata = try Some (Sys.getenv "CAMELISDATA" ^ dir_sep) with _ -> None

    let last_dir = ref (Sys.getcwd () ^ dir_sep)

(* Application specific code *)

    let filter_lis = "*.lis"
    let filter_ctx = "*.ctx"
    let filter_bin = "*.cxt" (* for binary contexts *)
    let filter_jpg = "*.jpg"

    let str_white = "white"
    let str_black = "black"
    let str_lightgrey = "lightgrey"
    let str_darkgrey = "darkgrey"
    let str_darkorange = "darkorange"
    let str_darkred = "darkred"
    let str_darkgreen = "darkgreen"
    let str_default_color = "lightgrey"

    let str_color_of_rgb (r,g,b) = Printf.sprintf "#%02x%02x%02x" r g b

(*
    let colormap = Gdk.Color.get_system_colormap ()
    let white = Gdk.Color.alloc ~colormap `WHITE
    let black = Gdk.Color.alloc ~colormap `BLACK
    let default_color = Gdk.Color.alloc ~colormap (`NAME "lightgrey")
*)

    type lift = {
	lift_supp : int;
	lift_scale : float;
	lift_color : string;
      }

    type incr = {
	incr_lift : lift;    (* lift of feature w.r.t. query *)
	incr_supp : int;    (* support of feature in query *)
	incr_scale : float; (* scale representing support/query *)
	incr_color : string; (* grey-level equivalent to scale *)
	incr_ext : Ext.t;   (* coverage of feature in query *)
      }

    type links_sort_fun = (Context.Root.Query.t * incr) -> (Context.Root.Query.t * incr) -> int
    let sort_by_count (s,i) (s',i') =
      Common.compare_pair (compare,Context.Root.Query.compare) (i'.incr_supp,s) (i.incr_supp,s')
    let sort_by_feature (s,i) (s',i') =
      Common.compare_pair (Context.Root.Query.compare,compare) (s,i'.incr_supp) (s',i.incr_supp)

    type cb_copy =
      | Nothing
      | Features of Context.Feat.t list
      | Objects of Ext.t
      | Query of Query.t
      | Sources of Source.url list
      | Wells of Well.url list

    let cb_copy : cb_copy ref = ref Nothing (* clipboard of object ids for copy, paste, ... *)

    type extent_tab = String | Picture

    class state_extent_tab (type_tab : extent_tab) =
      object (self)
	val mutable nbobj : int = 0;
	val mutable page_start : int = 1;
	val mutable page_size : int = 12;
	val mutable page_scroll : float = 0.;

	method nbobj = nbobj
	method page_start = page_start
	method page_size = page_size
	method page_scroll = page_scroll

	method set_nbobj n = nbobj <- n
	method set_page_start v = page_start <- v
	method set_page_size v = page_size <- v
	method set_page_scroll v = page_scroll <- v

	method copy = {< page_start = 1; page_scroll = 0. >}
      end

    class state_extent =
      object (self)
	val mutable nbobj : int = 0
	val mutable ext : Ext.t = Ext.empty
	val mutable locals_only : bool = false
	val tabs : (extent_tab,state_extent_tab) Hashtbl.t = Hashtbl.create 3

	method nbobj = nbobj
	method ext = ext
	method locals_only = locals_only
	method tab t = Hashtbl.find tabs t

	method set_nbobj n = nbobj <- n
	method set_ext e = ext <- e
	method set_locals_only v = locals_only <- v
	method private add_tab t = Hashtbl.replace tabs t (new state_extent_tab t)
	method private rem_tab t = Hashtbl.remove tabs t

	method refresh =
	  let e = Context.extent () in
	  let n = Ext.cardinal e in
	  nbobj <- n;
	  ext <- e

	method copy =
	  {< tabs =
	     let h = Hashtbl.create 3 in
	     Hashtbl.iter
	       (fun t st -> Hashtbl.add h t st#copy)
	       tabs;
	     h >}

	initializer
	  self # add_tab String;
	  self # add_tab Picture
      end


    let all_ext_colors, last_ext_color =
(*      let splits = [| 0; 16; 8; 24; 4; 12; 20; 28; 2; 6; 10; 14; 18; 22; 26; 30;
		       1; 3; 5; 7; 9; 11; 13; 15; 17; 19; 21; 23; 25; 27; 29; 31 |] in *)
(*      let splits = [| 0; 8; 4; 12; 2; 6; 10; 14; 1; 3; 5; 7; 9; 11; 13; 15 |] in *)
      let splits = [| 0; 4; 2; 6; 1; 3; 5; 7|] in
      let nbsplits = Array.length splits in
      let base = 30000 in
      let levels = [| 20000; 40000|] in (* should not include base so as to avoid doublons *)
      let nblevels = Array.length levels in
      let chunk = (65535 - base) / nbsplits in
      let nbcolors = 3 * nblevels * nbsplits + 1 in
      let t = Array.make nbcolors str_default_color in
      for i = 0 to nbsplits - 1 do (* for all split between 2 colors *)
	let x1 = splits.(i) in
	let x2 = nbsplits - x1 in
	for l = 0 to nblevels - 1 do (* for all levels *)
	  let x, y, z = base + x1*chunk, base + x2*chunk, levels.(l) in 
	  for j = 0 to 2 do (* for every combination of 2 colors *)
	    let rgb =
	      match j with
	      | 0 -> z, x, y (* green-blue *)
	      | 1 -> y, z, x (* blue-red *)
	      | _ -> x, y, z (* red-green *) in 
	    t.(3*nblevels*i + 3*l + j) <- str_color_of_rgb rgb (* Gdk.Color.alloc ~colormap (`RGB rgb) *)
	  done
	done
      done;
      t, nbcolors-1

    type ext_data = { mutable count : int; mutable color : string option}

    class state_ext_colors =
      object (self)
	val mutable ext_colors : (Ext.t,ext_data) Hashtbl.t = Hashtbl.create 13
	val mutable next_color : int = 0

	method get_ext_color (e : Ext.t) =
	  try
	    let d = Hashtbl.find ext_colors e in
	    if d.count <= 1 then str_white else match d.color with Some c -> c | None -> str_white
	  with _ -> str_white

	method incr_ext_count (e : Ext.t) =
	  try
	    let d = Hashtbl.find ext_colors e in
	    d.count <- d.count + 1;
	    if d.count >= 2 && d.color = None
	    then begin
	      d.color <- Some all_ext_colors.(next_color);
	      next_color <- min (next_color + 1) last_ext_color
	    end
	  with _ ->
	    Hashtbl.replace ext_colors e {count = 1; color = None}

	method decr_ext_count (e : Ext.t) =
	  try
	    let d = Hashtbl.find ext_colors e in
	    d.count <- d.count - 1
	  with _ -> ()

	method copy = new state_ext_colors (* all colors can be reused in new state *)
(*
	  (* next_color is unchanged *)
	  {< ext_colors =
	       let ht = Hashtbl.create 13 in
	       Hashtbl.iter (fun e d -> Hashtbl.add ht e {d with count = 0}) ext_colors;
	       ht;>}
*)
      end

    let default_links_page_size = 15

    class state_view ext_colors0 =
      object (self)
	val mutable ext_colors : state_ext_colors = ext_colors0
	val mutable to_be_expanded : bool = false
	val mutable expanded : bool = false
	val mutable links_opt : (string * incr) list option = None
	val mutable locals_opt : Ext.t option = None
	val mutable links_options : Context.links_option list = []
	val mutable links_sort_fun : links_sort_fun = sort_by_feature
	val mutable links_page_start : int = 1 (* first page *)
	val mutable links_page_size : int = default_links_page_size (*  use max_int for no limit *)
 
	method expanded = expanded || to_be_expanded
	method links_opt = links_opt
	method locals_opt = locals_opt
	method links_options = links_options
	method links_sort_fun = links_sort_fun
	method links_page_start = links_page_start
	method links_page_size = links_page_size

        method set_to_be_expanded b = to_be_expanded <- b

	method set_expanded b =
	  if not expanded && b
	  then begin
	    match links_opt with
	    | None -> assert false
	    | Some l ->
		List.iter (fun (_,i) -> ext_colors # incr_ext_count i.incr_ext) l end
	  else if expanded && not b
	  then begin
	    match links_opt with
	    | None -> assert false
	    | Some l ->
		List.iter (fun (_,i) -> ext_colors # decr_ext_count i.incr_ext) l
	  end;
	  expanded <- b

	method set_locals_opt ext = locals_opt <- ext
	method set_links_options options = links_options <- options
	method set_links_sort_fun f = links_sort_fun <- f
	method set_links_page_start s = links_page_start <- s
	method set_links_page_size s = links_page_size <- s

	method set_links_conceptual_ordering b =
	  if b
	  then
	    if not (List.mem `MaxOnly links_options)
	    then links_options <- `MaxOnly :: links_options
	    else ()
	  else
	    links_options <- List.filter ((<>) `MaxOnly) links_options

	method set_links_minimum_support n_opt =
	  links_options <- List.filter (function `SuppMin _ -> false | _ -> true) links_options;
	  (match n_opt with None -> () | Some n -> links_options <- `SuppMin n ::  links_options)

	method set_links_opt l =
	  if expanded then begin
	    ( match links_opt with
	    | None -> ()
	    | Some l_old ->
		List.iter (fun (_,i) -> ext_colors # decr_ext_count i.incr_ext) l_old);
	    (match l with
	    | None -> ()
	    | Some l_new ->
		List.iter (fun (_,i) -> ext_colors # incr_ext_count i.incr_ext) l_new)
	  end;
	  links_opt <- l

	method copy ext_colors0 = {< ext_colors = ext_colors0; links_opt = None; locals_opt = None; links_page_start = 1>}
      end

    class state = 
      let get_query_features () : string Bintree.t * string Bintree.t =
	let bintree_of_list l =
	  List.fold_left
	    ~f:(fun res q -> Bintree.add (string_of_query q) res)
	    ~init:Bintree.empty
	    l
	in
	let lp, ln = Context.features_of_query (Context.wq ()) in
	bintree_of_list lp, bintree_of_list ln
      in
      object (self)
	val mutable query : Query.t = Context.wq ()
	val mutable query_features : string Bintree.t * string Bintree.t = get_query_features ()
	val mutable extent : state_extent = new state_extent
	val mutable views : (string,state_view) Hashtbl.t = Hashtbl.create 13
	val mutable ext_colors : state_ext_colors = new state_ext_colors
	val mutable ext_preselect : Ext.t option = None
	val mutable int_preselect : Ext.t option = None (* set of selected objects *)

	method query = query
	method query_features = query_features
	method extent = extent
	method views = views
	method ext_preselect = ext_preselect
	method int_preselect = int_preselect

	method new_view = new state_view ext_colors

	method copy_view (sv : state_view) = sv # copy ext_colors

	method get_view v =
	  try Hashtbl.find views v
	  with Not_found ->
	    let sv = new state_view ext_colors in
	    Hashtbl.add views v sv;
	    sv

	method ext_colors = ext_colors

	method replace_view : string -> string -> unit =
	  fun v v' ->
	    try
	      Hashtbl.replace views v' (Hashtbl.find views v);
	      Hashtbl.remove views v
	    with _ -> ()

	method fold_views : 'a.(string -> state_view -> 'a -> 'a) -> 'a -> 'a = fun f e -> Hashtbl.fold f views e

	method set_ext_preselect e_opt = ext_preselect <- e_opt
	method set_int_preselect oids = int_preselect <- oids

	method locals : Ext.t =
	  Ext.inter_r
	    (self # fold_views
	       (fun v sv res ->
		 if sv # expanded
		 then Common.fold_option (fun x -> x :: res) res (sv # locals_opt)
		 else res)
	       [extent#ext])

	method copy =
	  let ext_colors_copy = ext_colors # copy in
	  {< query = Context.wq ();
	     query_features = get_query_features ();
	     extent = extent # copy;
	     ext_colors = ext_colors_copy;
	     ext_preselect = None;
	     int_preselect = None;
	     views =
               let ht = Hashtbl.create (Hashtbl.length views) in
	       Hashtbl.iter (fun v sv -> Hashtbl.add ht v (sv # copy ext_colors_copy)) views;
	       ht;>}

	initializer
	  Hashtbl.add views "all" (new state_view ext_colors)
      end

    class history =
      object (self)
	val st_home : state = new state
	val st_backward : state Stack.t = Stack.create ()
	val st_forward : state Stack.t = Stack.create ()

	method home = st_home

	method current = Stack.top st_backward

	method init =
	  Stack.clear st_backward;
	  Stack.clear st_forward;
	  Stack.push st_home#copy st_backward

	method after_cd =
	  Stack.clear st_forward;
	  self # current # set_ext_preselect None;
	  self # current # set_int_preselect None;
	  let st = self # current # copy in
	  Stack.push st st_backward

	method after_cd_home =
	  Stack.clear st_forward;
	  Stack.push st_home#copy st_backward

	method backward =
	  if Stack.length st_backward > 1
	  then begin
	    Context.wq_back ();
	    let s = Stack.pop st_backward in
	    s # set_ext_preselect None;
	    s # set_int_preselect None;
	    Stack.push s st_forward
	  end

	method forward =
	  if Stack.length st_forward > 0
	  then begin
	    self # current # set_ext_preselect None;
	    self # current # set_int_preselect None;
	    let s = Stack.pop st_forward in
	    Context.wq_incr (fun _ -> s # query);
	    Stack.push s st_backward
	  end

	initializer
	  Stack.push st_home#copy st_backward
      end

    let history = new history

    let cd (path : string) : unit =
      Context.wq_incr (parse_of_string Context.parse_incr_query path)

    let cd_abs (path : string) : unit =
      Context.wq_incr (fun _ -> query_of_string path)

    let cd_intent oids =
      let fs = Context.intent oids in
      let q = 
	match fs with
	| [] -> "all"
	| _ -> String.concat " and " (List.map string_of_feat fs) in
      cd_abs q

(*
    let locals ext : Ext.t =
      Ext.inter_r
	(history # current # fold_views
	   (fun v sv res ->
	     if sv # expanded
	     then Common.fold_option (fun x -> x) ext (sv # locals_opt) :: res
	     else res)
	   [ext])
*)

    let expected_supp nbo nbq nbx = (nbx * nbq) / nbo

    let scale_of_ratio n m =
      let max_scale = 1.5 in
      let min_scale = 0.8 in
      min_scale +.
	(max_scale -. min_scale) *.
	log10 (float_of_int (2 + n)) /. log10 (float_of_int (2 + m))
    
    let color_of_ratio n m = (* assuming 0 <= n <= m *)
      let max_intensity = 65535. in
      let min_intensity = 16000. in
      let rgb = int_of_float (
	if m = 0 then max_intensity
	else if n = 0 then min_intensity
	else
	  min_intensity +.
	    (max_intensity -. min_intensity) *.
	    log10 (float_of_int (2 * n)) /. log10 (float_of_int (2 * m))) in
      str_color_of_rgb (rgb,rgb,rgb) (* Gdk.Color.alloc ~colormap (`RGB (rgb,rgb,rgb)) *)


    let lift nbo nbq nbx nbqx = 
      (* float_of_int nbo *. float_of_int nbqx /. float_of_int nbq /. float_of_int nbx *) (* lift *)
      (*
	 if nbx = 0
	 then 1.
	 else float_of_int nbqx /. float_of_int nbx
       *) (* anti-proportion *)
      { lift_supp = nbx;
	lift_scale = scale_of_ratio nbqx nbx;
	lift_color = color_of_ratio nbqx nbx
	  (* begin
	    let nbo, nbq, nbx, nbqx = float_of_int nbo, float_of_int nbq, float_of_int nbx, float_of_int nbqx in
	    let base = 10000. in
	    let marge = 65535. -. base in
	    if nbqx /. nbx >= nbq /. nbo
	    then (* positive lift -> green *)
	      let p = (nbo *. nbqx -. nbx *. nbq) /. nbx /. (nbo -. nbq) in (* in 0..1 *)
	      let rb = int_of_float (base +. marge *. (1. -. p)) in
	      str_color_of_rgb (rb,65535,rb) (* Gdk.Color.alloc ~colormap (`RGB (rb,65535,rb)) *)
	    else (* negative lift -> red *)
	      let p = (nbx *. nbq -. nbo *. nbqx) /. nbx /. nbq in
	      let gb = int_of_float (base +. marge *. (1. -. p)) in
	      str_color_of_rgb (65535,gb,gb) (* Gdk.Color.alloc ~colormap (`RGB (65535,gb,gb)) *)
	  end *)
	  (* begin
	    let base = 20000. in
	    let marge = 65535. -. base in
	    let p = if nbx = 0 then 0. else float_of_int nbqx /. float_of_int nbx in
	    let rgb = int_of_float (base +. marge *. p) in
	    str_color_of_rgb (rgb,rgb,rgb) (* Gdk.Color.alloc ~colormap (`RGB (rgb,rgb,rgb)) *)
	  end *)
      }


    let get_incr (path : string) (i0 : incr) : incr =
      let p = Context.link (query_of_string path) in
      {i0 with
       incr_lift = { i0.incr_lift with
		     lift_supp = p.Context.propr_card};
       incr_supp = p.Context.propr_supp;
       incr_ext = p.Context.propr_ext}

    let rec ls root sv (path, i : string * incr) : (string * incr) list * Ext.t = Common.prof "ls" (fun () ->
      let view =
	let query =
	  if path = "all"
	  then Context.Root.Query.top
	  else query_of_string path in
	let options =
	  if root && not (List.exists (function `SuppMin _ -> true | _ -> false) sv#links_options) then
	    `SuppMin 0 :: sv#links_options
	  else if i.incr_supp = 0 then
	    `SuppMin 0 :: List.filter (function `SuppMin _ -> false | _ -> true) sv#links_options
(*
	  else if not (List.exists (function `SuppMin _ -> true | _ -> false) sv#links_options) then
	    `SuppMin 1 :: sv#links_options
*)
	  else sv#links_options in
	Context.links options query in
      let incrs =
	List.fold_left
	  ~f:(fun res p ->
	    let lift = lift view.Context.nbobj view.Context.nbsel p.Context.propr_card p.Context.propr_supp in
	    let scale = scale_of_ratio p.Context.propr_supp view.Context.nbsel in
	    let color = color_of_ratio p.Context.propr_supp view.Context.nbsel in
	    List.fold_left
	      ~f:(fun res' log ->
		let i = {incr_lift = lift;
			 incr_supp = p.Context.propr_supp;
			 incr_scale = scale;
			 incr_color = color;
			 incr_ext = p.Context.propr_ext} in
		(log,i)::res')
	      ~init:res
	      p.Context.propr_logs
	  )
	  ~init:[]
	  view.Context.proprs in
      let incrs1 =
	if List.exists (function `SuppMin _ -> true | _ -> false) sv # links_options
	then incrs
	else
	  let incrs =
	    List.sort
	      ~cmp:(fun (_,i1) (_,i2) -> Pervasives.compare (i2.incr_supp,i2.incr_lift.lift_supp) (i1.incr_supp,i1.incr_lift.lift_supp))
	      incrs in
	  let _, _, incrs1, _ =
	    Common.fold_while
	      (fun (incrs,n1,incrs1,supp_card) ->
		match incrs with
		| [] -> None
		| (log,i)::incrs' ->
		    let supp_card' = (i.incr_supp,i.incr_lift.lift_supp) in
		    if n1 < sv # links_page_size || supp_card' >= supp_card
		    then Some (incrs',n1+1,(log,i)::incrs1,supp_card')
		    else None)
	      (incrs,0,[],(max_int,max_int)) in
	  incrs1 in
      let incrs2 = List.sort ~cmp:sv # links_sort_fun incrs1 in
      let incrs3 = List.map (fun (log,i) -> (string_of_query log, i)) incrs2 in
      (incrs3, view.Context.locals))

    let ls_opt force root sv (v,i) =
      if force or sv # links_opt = None
      then begin
	let links, locals = ls root sv (v,i) in
	sv # set_links_opt (Some links);
	sv # set_locals_opt (Some locals);
	links end
      else match sv # links_opt with
      | Some links -> links
      | None -> assert false

(* graphical utilities *)

    let pixbuf_of_picture =
      let ht = Hashtbl.create 100 in
      fun ~maxside filename ->
	try Hashtbl.find ht filename
	with Not_found ->
	  let pixbuf = GdkPixbuf.from_file_at_size filename ~width:maxside ~height:maxside in
	  Hashtbl.add ht filename pixbuf;
	  pixbuf

    let file_dialog ~title ?filename ?filter ~callback () =
      let sel = GWindow.file_selection ~title ~modal:true ?filename () in
      sel#cancel_button#connect#clicked ~callback:sel#destroy;
      sel#ok_button#connect#clicked
	~callback:(fun () ->
	  let name = loc sel#filename in
	  sel#destroy ();
	  callback name);
      sel#show ();
      ( match filter with
      | Some s -> sel#complete ~filter:s
      | None -> ())

    let error_message msg =
      let md = GWindow.message_dialog
	  ~message:(utf msg)
	  ~message_type:`ERROR
	  ~buttons:GWindow.Buttons.ok
	  ~modal:true () in
      md#run ();
      md#destroy ()

    let info_message ?title ?icon msg =
      let md = GWindow.message_dialog
	  ~message:(utf msg)
	  ~message_type:`INFO
	  ?title
	  ?icon
	  ~buttons:GWindow.Buttons.ok
	  ~modal:true () in
      md#run ();
      md#destroy ()

    let question msg =
      let md = GWindow.message_dialog
	  ~message:(utf msg)
	  ~message_type:`QUESTION
	  ~buttons:GWindow.Buttons.yes_no
	  ~modal:true () in
      let res = md#run () in
      md#destroy ();
      res = `YES

    let confirm_overwrite filename =
      not (Sys.file_exists filename) or question "This file already exists. Are you sure you want to overwrite it ?"

    let syntax_error_logic msg =
      error_message (if msg="" then "Syntax error" else msg)

    let parse_dialog ~title ?text ~(parse:string -> 'a) ~(callback : 'a -> unit) =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~modal:true
	  ~title
	  () in
      let vbox = w#vbox in
      let entry = GEdit.entry ?text ~width_chars:80 ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      let action () =
	try
	  let res = parse entry#text in
	  w#destroy ();
	  callback res
	with
	| Stream.Error msg
	| Failure msg -> syntax_error_logic msg
	| _ -> syntax_error_logic "Error..." in
      entry#connect#activate ~callback:action;
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:action;
      w#show ()

(* building the interface *)

    let accel_group = GtkData.AccelGroup.create ()

    let tooltips = GData.tooltips ()

    let title () =
      let name = Context.filename () in
      let readonly = Context.readonly () in
      "Camelis" ^
      (if name="" then "" else " - " ^ Filename.basename name) ^
      (if name <> "" && readonly then " (read only)" else "")

    let window = GWindow.window ~title:(title ()) ()

    let vbox = GPack.vbox ~border_width:8 ~spacing:8 ~packing:window#add ()

    let menubar = GMenu.menu_bar ~packing:vbox#pack ()
    let menubar_factory = new GMenu.factory menubar

    let file_menu = menubar_factory#add_submenu "File"
    let file_menu_factory = new GMenu.factory file_menu
    let cmd_new = file_menu_factory#add_item "New..."
    let cmd_open = file_menu_factory#add_item "Open..."
    let cmd_save = file_menu_factory#add_item ~key:_s "Save"
    let cmd_saveas = file_menu_factory#add_item "Save as..."
    let cmd_close = file_menu_factory#add_item "Close"
    let _ = file_menu_factory#add_separator ()
    let import_file_menu = file_menu_factory#add_submenu "Import"
    let import_file_factory = new GMenu.factory import_file_menu
    let cmd_import_from_ctx = import_file_factory#add_item "Import context..."
    let cmd_import_from_file = import_file_factory#add_item "Import file..."
    let cmd_import_list = import_file_factory#add_item "List sources..."
    let _ = import_file_factory#add_separator ()
    let cmd_import_update = import_file_factory#add_item ~key:_u "Update all"
    let export_menu = file_menu_factory#add_submenu "Export"
    let export_factory = new GMenu.factory export_menu
    let cmd_export_to_ctx = export_factory#add_item "Export context..."
    let cmd_export_to_bin = export_factory#add_item "Export context in Burmeister format..."
    let cmd_export_file = export_factory#add_item "Export file..."
    let cmd_export_list = export_factory#add_item "List wells..."
    let _ = file_menu_factory#add_separator ()
    let cmd_stat = file_menu_factory#add_item "Statistics"
    let _ = file_menu_factory#add_separator ()
    let cmd_quit = file_menu_factory#add_item "Quit"

    let logic_menu = menubar_factory#add_submenu "Logic"
    let logic_menu_factory = new GMenu.factory logic_menu
    let cmd_show = logic_menu_factory#add_item ~key:_a "Add feature..."
    let cmd_join = logic_menu_factory#add_item "Join as..."
    let cmd_copy_features = logic_menu_factory#add_item "Copy features"
    let cmd_paste_features = logic_menu_factory#add_item "Paste"
    let cmd_hide = logic_menu_factory#add_item ~key:_h "Delete features"
    let _ = logic_menu_factory#add_separator ()
    let cmd_copy_query = logic_menu_factory#add_item "Copy query"
    let cmd_copy_paste_query = logic_menu_factory#add_item "Copy and paste query to..."
    let cmd_delete_query = logic_menu_factory#add_item "Delete query"
    let cmd_rule_list = logic_menu_factory#add_item "List rules..."

    let view_menu = menubar_factory#add_submenu "Browsing"
    let view_menu_factory = new GMenu.factory view_menu
    let cmd_home = view_menu_factory#add_item "Home"
    let cmd_current_state = view_menu_factory#add_item "Current state"
    let cmd_back = view_menu_factory#add_item ~key:_b "Back"
    let cmd_fwd = view_menu_factory#add_item ~key:_f "Forward"
    let cmd_refresh = view_menu_factory#add_item ~key:_r "Refresh"
    let _ = view_menu_factory#add_separator ()
    let cmd_complement = view_menu_factory#add_item "NOT <query>"
    let cmd_if = view_menu_factory#add_item "<query> IF All"
    let cmd_unif = view_menu_factory#add_item "IF -> AND"
    let cmd_intent = view_menu_factory#add_item "Intent <objects>"
    let _ = view_menu_factory#add_separator ()
    let check_scale = view_menu_factory#add_check_item ~active:true "Variable font size"
    let check_feat_color = view_menu_factory#add_check_item ~active:true "Foreground color"
    let check_lift = view_menu_factory#add_check_item ~active:false "Total count"
    let check_lift_color = view_menu_factory#add_check_item ~active:true "Total count grey level"
    let check_ext_color = view_menu_factory#add_check_item ~active:true "Equal extent colors"
    let _ = view_menu_factory#add_separator ()
    let cmd_expand = view_menu_factory#add_item "Expand..."
    let cmd_collapse = view_menu_factory#add_item "Collapse"
    let cmd_extent_options = view_menu_factory#add_item "Extent options..."

    let objects_menu = menubar_factory#add_submenu "Updating"
    let objects_menu_factory = new GMenu.factory objects_menu
    let cmd_new_state = objects_menu_factory#add_item ~key:_e "New state"
    let _ = objects_menu_factory#add_separator ()
    let cmd_make = objects_menu_factory#add_item ~key:_o "Add object"
    let cmd_copy = objects_menu_factory#add_item "Copy objects"
    let cmd_paste_extr = objects_menu_factory#add_item ~key:_p "Paste"
    let cmd_copy_paste_to = objects_menu_factory#add_item ~key:_t "Copy and paste objects to..."
(*    let cmd_paste_intr = objects_menu_factory#add_item "Paste intrinsic"*)
    let cmd_delete = objects_menu_factory#add_item "Delete objects"

    let actions_menu = menubar_factory#add_submenu "Actions"
    let actions_menu_factory = new GMenu.factory actions_menu
    let cmd_add_action = actions_menu_factory#add_item "Add action..."
    let cmd_list_action = actions_menu_factory#add_item "List actions..."

    let help_menu = menubar_factory#add_submenu "Help"
    let help_menu_factory = new GMenu.factory help_menu
    let cmd_about = help_menu_factory#add_item "About Camelis..."

    let hbox_buttons = GPack.hbox ~spacing:8 ~height:30 ~packing:vbox#pack ()
    let button_home = GButton.button ~stock:`HOME ~packing:(hbox_buttons#pack) ()
    let button_back = GButton.button ~stock:`GO_BACK ~packing:(hbox_buttons#pack) ()
    let button_fwd = GButton.button ~stock:`GO_FORWARD ~packing:(hbox_buttons#pack) ()
    let button_refresh = GButton.button ~stock:`REFRESH ~packing:(hbox_buttons#pack) ()
    let button_save = GButton.button ~stock:`SAVE ~packing:(hbox_buttons#pack) ()
(*    let button_make = GButton.button ~stock:(`STOCK "Add object") ~packing:(hbox_buttons#pack) ()*)
    let button_paste = GButton.button ~stock:`PASTE ~packing:(hbox_buttons#pack) ()
    let button_update = GButton.button ~stock:(`STOCK " Update ") ~packing:(hbox_buttons#pack) ()
    let entry_nbupdates = GEdit.entry ~width_chars:8 ~packing:(hbox_buttons#pack) ()

    let hbox_checks = GPack.hbox ~spacing:8 ~packing:vbox#pack ()
    let check_int_preselect = GButton.check_button ~label:"Features preselection" ~active:false ~packing:hbox_checks#pack ()
    let check_ext_preselect = GButton.check_button ~label:"Objects preselection" ~active:false ~packing:hbox_checks#pack ()
(*
    let check_int_preselect = view_menu_factory#add_check_item ~active:false "Features preselection"
    let check_ext_preselect = view_menu_factory#add_check_item ~active:false "Objects preselection"
*)
(* deprecated
    class entry_query ~width_chars ~pack =
      let entry = GEdit.entry ~width_chars ~packing:pack () in
      object (self)
	val mutable callback_activated = fun query -> ()
	val mutable callback_ctx_menu = fun query menu -> ()

	method connect_activated ~callback = callback_activated <- callback
	method connect_ctx_menu ~callback = callback_ctx_menu <- callback

	method activated () =
	  let query = loc0 entry#text in
	  callback_activated query

	method button_press ev =
	  let button = GdkEvent.Button.button ev in
	  if button = 3
	  then begin
	    let query = loc0 entry#text in
	    let menu = GMenu.menu () in
	    callback_ctx_menu query menu;
	    menu#popup ~button:button ~time:(GdkEvent.Button.time ev);
	    true end
	  else false

	method copy_query =
	  let q = Context.wq () in
	  cb_copy := Query q

	method refresh =
	  let wq = Context.wq () in
	  entry # set_text (utf0 (Syntax.stringizer (Query.print wq)))

	initializer
	  entry # connect # activate ~callback:self # activated;
	  entry # event # connect # button_press ~callback:self # button_press;
	  tooltips#set_tip entry#coerce
	    ~text:"This is the query area that you can edit directly or by activating features in the navigation window";
      end
*)

    class text_query ~width_chars ~pack =
      let tag_table = GText.tag_table () in
      let buffer = GText.buffer ~tag_table:tag_table () in
      let tag_scale_up = buffer # create_tag [`SCALE (`CUSTOM 1.5)] in
      let hbox = GPack.hbox ~spacing:8 ~packing:pack () in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN
	  ~hpolicy:`NEVER ~vpolicy:`AUTOMATIC ~packing:(fun w -> hbox#pack ~expand:true w) () in
      let view = GText.view ~buffer:buffer
	  ~editable:true
	  ~cursor_visible:true
	  ~wrap_mode:`WORD
	  ~width:width_chars
	  ~height:2
	  ~packing:sw#add () in
      let button = GButton.button ~stock:`APPLY ~packing:hbox#pack () in
      object (self)
	val mutable callback_activated = fun query -> ()
	val mutable callback_ctx_menu = fun query menu -> ()

	method connect_activated ~callback = callback_activated <- callback
	method connect_ctx_menu ~callback = callback_ctx_menu <- callback

	method activated () =
	  let query = loc0 (buffer # get_text ()) in
	  callback_activated query

	method button_press ev =
	  let button = GdkEvent.Button.button ev in
	  if button = 3
	  then begin
	    let query = loc0 (buffer # get_text ()) in
	    let menu = GMenu.menu () in
	    callback_ctx_menu query menu;
	    menu#popup ~button:button ~time:(GdkEvent.Button.time ev);
	    true end
	  else false

	method copy_query =
	  let q = Context.wq () in
	  cb_copy := Query q

	method refresh =
	  let wq = Context.wq () in
	  buffer # set_text (utf0 (Syntax.stringizer (Query.print wq) ^ " "));
	  buffer # apply_tag tag_scale_up ~start:buffer#start_iter ~stop:buffer#end_iter

	initializer
	  button # connect # clicked ~callback:self#activated;
	  view # event # connect # button_press ~callback:self#button_press;
	  tooltips#set_tip view#coerce
	    ~text:"This is the query area that you can edit directly or by activating features in the navigation window";
      end

    type navigation_mode = Zoom | Pivot

    class navigation_link_callbacks =
      object
	method cd ~(mode:navigation_mode) ~(negation:bool) ~(comparison:string) = ()
      end

    class navigation_link ~pack =
      let hbox = GPack.hbox ~spacing:8 ~packing:pack () in
      let check_not = GButton.check_button ~label:"NOT" ~active:false ~packing:hbox#pack () in
      let radio_eq = GButton.radio_button ~label:"=" ~active:true ~packing:hbox#pack () in
      let radio_ge = GButton.radio_button ~group:radio_eq#group ~label:">=" ~packing:hbox#pack () in
      let radio_le = GButton.radio_button ~group:radio_eq#group ~label:"<=" ~packing:hbox#pack () in
      let button_pivot = GButton.button ~label:"Pivot" ~packing:(hbox#pack ~from:`END) () in
      let button_zoom = GButton.button ~label:"Zoom" ~packing:(hbox#pack ~from:`END) () in
      object (self)
	val mutable callbacks = new navigation_link_callbacks

	method connect c = callbacks <- c

	method cd mode () =
	  let negation = check_not#active in
	  let comparison =
	    if radio_eq#active then "="
	    else if radio_ge#active then ">="
	    else "<=" in
	  callbacks#cd ~mode ~negation ~comparison

	initializer
	  button_zoom # connect # clicked ~callback:(self#cd Zoom);
	  button_pivot # connect # clicked ~callback:(self#cd Pivot);
	  tooltips#set_tip check_not#coerce
	    ~text:"Apply negation (NOT) on the current selection";
	  tooltips#set_tip button_zoom#coerce
	    ~text:"Specialize/generalize the current query with the current selection";
	  tooltips#set_tip button_pivot#coerce
	    ~text:"Replace the current query by the current selection";
	  tooltips#set_tip radio_eq#coerce
	    ~text:"Keep = in the current selection";
	  tooltips#set_tip radio_ge#coerce
	    ~text:"Replace = by >= in the current selection";
	  tooltips#set_tip radio_le#coerce
	    ~text:"Replace = by <= in the current selection"
      end

    class tree_features_callbacks =
      object
	method row_activated : Gtk.tree_path -> unit = fun path -> ()
	method row_expanded : Gtk.tree_path -> unit = fun path -> ()
	method row_collapsed : Gtk.tree_path -> unit = fun path -> ()
	method refresh_preselect : bool -> unit = fun _ -> ()
	method ctx_menu : int -> GMenu.menu -> unit = fun nb menu -> ()
	    (* the int is the number of selected features *)
	method data_received : unit -> unit = fun () -> ()
      end

    class tree_features ~pack v =
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN
	  ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:pack () in
      let cols = new GTree.column_list in
      let feature = cols#add string in
      let incr = cols#add caml in
      let count = cols#add string in
      let lift_label = cols#add string in
(*      let lift_scale = cols#add float in *)
(*      let lift_fg_color = cols#add string in *)
      let lift_color = cols#add string in
(*      let ext = cols#add caml in *)
      let row_color = cols#add string in
      let row_scale = cols#add float in
      let bg_color = cols#add string in
      let model = GTree.tree_store cols in
      let view = GTree.view ~model:model ~headers_visible:false ~packing:sw#add () in
      let cell_count =  GTree.cell_renderer_text [`XALIGN 1.] in
      let col_count =
	let col =
	  GTree.view_column ~title:"Count" ()
	    ~renderer:(cell_count, ["text",count]) in
	col # add_attribute cell_count "scale" row_scale;
	col # set_sizing `GROW_ONLY;
	col in
      let cell_lift = GTree.cell_renderer_text [`XALIGN 1.] in
      let col_lift =
	let col =
	  GTree.view_column ~title:"Lift" ()
	    ~renderer:(cell_lift, ["text",lift_label]) in
(*	col # add_attribute cell_lift "scale" lift_scale; *)
(*	col # add_attribute cell_lift "foreground" lift_fg_color; *)
	col # add_attribute cell_lift "background" lift_color;
	col # set_sizing `GROW_ONLY;
	col in
      let cell_feature = GTree.cell_renderer_text [] in
      let col_feature =
	let col =
	  GTree.view_column ~title:"Feature" ()
	    ~renderer:(cell_feature, ["text",feature]) in
	col # add_attribute cell_feature "scale" row_scale;
	col # add_attribute cell_feature "foreground" row_color;
	col # add_attribute cell_feature "background" bg_color;
	col # set_sizing `GROW_ONLY;
	col in
      let function_top = "<top>" in
      let function_up = "<up>" in
      let function_down = "<down>" in
      let function_bottom = "<bottom>" in
      object (self)
	val mutable callbacks = new tree_features_callbacks

	method connect c = callbacks <- c

	method widget = (sw :> GObj.widget)

	method root = v

	val mutable ignore_selection_changed = false

	method selection_changed () =
	  if ignore_selection_changed
	  then ignore_selection_changed <- false
	  else begin
	    let paths = view # selection # get_selected_rows in
	    let new_preselection =
	      match paths (*if check_ext_preselect#active then paths else []*) with
	      | [] -> None
	      | lp ->
		  let ext =
		    Ext.union_r
		      (Common.mapfilter
			 (fun path ->
			   try Some (self # ext_from_path path)
			   with _ -> None)
			 lp) in
		  if Ext.is_empty ext
		  then None
		  else Some ext in
	    let st = history # current in
	    if new_preselection <> st # ext_preselect
	    then begin
	      st # set_ext_preselect new_preselection;
	      st # set_int_preselect new_preselection; (* *)
	      if check_int_preselect#active then self # refresh_preselect false;
	      if check_ext_preselect#active then callbacks # refresh_preselect true
	    end
	  end

	method row_activated path vcol =
	  let iter = model # get_iter path in
	  let count = model#get ~row:iter ~column:count in
	  if count <> ""
	  then callbacks # row_activated path
	  else begin (* a function has been activated *)
	    let sf = self # feature_from_path path in
	    if sf = function_top then self # page_top (model#iter_parent iter)
	    else if sf = function_up then self # page_up (model#iter_parent iter)
	    else if sf = function_down then self # page_down (model#iter_parent iter)
	    else if sf = function_bottom then self # page_bottom (model#iter_parent iter)
	    else ()
	  end

	method button_press ev =
	  let button = GdkEvent.Button.button ev in
	  if button = 3 then begin
	    let selection = view#selection in
	    let nb = selection#count_selected_rows in
	    if nb = 0 then begin
	      let x = int_of_float (GdkEvent.Button.x ev) in
	      let y = int_of_float (GdkEvent.Button.y ev) in
	      match view#get_path_at_pos ~x ~y with
	      | None -> ()
	      | Some (path, _, _, _) ->
		  selection#unselect_all ();
		  selection#select_path path end;
	    let menu = GMenu.menu () in
	    callbacks # ctx_menu nb menu;
	    menu#popup ~button:(GdkEvent.Button.button ev) ~time:(GdkEvent.Button.time ev);
	    true end
	  else false

	method private feature_from_tree_iter tree_iter =
	  loc0 (model#get ~row:tree_iter ~column:feature)

	method ext_from_path path : Ext.t =
	  (model#get ~row:(model#get_iter path) ~column:incr).incr_ext

	method feature_from_path path : string =
	  loc0 (model#get ~row:(model#get_iter path) ~column:feature)

	method incr_from_path path : incr =
	  model#get ~row:(model#get_iter path) ~column:incr

	method state_view = history # current # get_view v

	(* get the state_view of the path, and create it if necessary *)
	method state_view_from_path path =
	  let st = history # current in
	  let views = st # views in
	  let v = self # feature_from_path path in
	  try
	    Hashtbl.find views v
	  with _ ->
	    let sv =
	      try
		match model#iter_parent (model#get_iter path) with
		| None -> st # copy_view (self # state_view)
		| Some iter -> st # copy_view (Hashtbl.find views (self # feature_from_tree_iter iter))
	      with _ -> (* should not happen *)
		st # new_view in
	    Hashtbl.add views v sv;
	    sv

(*
	method is_strict_empty path =
	  let c = model#get ~row:(model#get_iter path) ~column:count in
	  c < history # current # extent # nbobj, c = 0
*)

	method private row_color_from_incr (s,i) =
	  if not check_feat_color#active
	  then str_black
	  else
	    let pos_feats, neg_feats = history # current # query_features in
	    let nbobj = history # current # extent # nbobj in
	    let name =
	      if i.incr_supp = 0 && Bintree.mem s neg_feats then str_darkgreen
	      else if i.incr_supp = nbobj && Bintree.mem s pos_feats then str_darkred
	      else if i.incr_supp = 0 then str_darkgrey
	      else if i.incr_supp = nbobj then str_darkorange
	      else str_black in
	    name

	method private row_scale_from_incr (s,i) =
	  if not check_scale#active
	  then 1.
	  else i.incr_scale

(*
	method private count_color_from_incr (s,i) =
	  if not check_count_color#active
	  then str_white
	  else i.incr_color
*)

	method private lift_label_from_incr (s,i) =
	  (* Printf.sprintf "%d%%" (int_of_float (i.incr_lift *. 100.)) *)
	  if not check_lift#active
	  then " "
	  else string_of_int i.incr_lift.lift_supp

	method private lift_scale_from_incr (s,i) =
	  if not check_scale#active
	  then 1.
	  else i.incr_lift.lift_scale

	method private lift_color_from_incr (s,i) =
	  if not check_lift_color#active
	  then str_white
	  else i.incr_lift.lift_color

	method private lift_fg_color_from_incr (s,i) =
	  if true || not check_lift_color#active || i.incr_scale >= i.incr_lift.lift_scale
	  then str_black
	  else str_white

(*
	method private row_bg_color_from_incr (s,n,e) =
	  if check_int_preselect#active
	  then Common.fold_option (fun oids -> if not (Ext.is_empty (Ext.inter oids e)) then str_white else str_lightgrey) str_white history#current#int_preselect
	  else str_white
*)
	      
	method private bg_color_from_ext int_preselect e =
	  if not check_int_preselect#active
	  then str_white
	  else
	    match int_preselect with
	    | None -> str_white
	    | Some oids ->
		let inter = Ext.inter oids e in
		color_of_ratio (Ext.cardinal inter) (Ext.cardinal oids)
(*
	  if not check_int_preselect#active || Common.fold_option (fun oids -> not (Ext.is_empty (Ext.inter oids e))) true int_preselect
	  then str_white
	  else str_lightgrey
*)

	method selected_features = view # selection # get_selected_rows

	method selected_update ?(interactive = false) ?(neg = false) ~callback () =
	  let paths = view#selection#get_selected_rows in
(*	  view#selection#unselect_all (); *)
	  let init_update =
	    if paths = []
	    then
	      string_of_update (Context.update_of_query (Context.wq ()))
	    else
	      let pos_neg s = if neg then "not " ^ s else s in
	      let conj_feats = String.concat " and " (List.map (fun path -> pos_neg (self # feature_from_path path)) paths) in
	      conj_feats in
	  try
	    if interactive then failwith "must call dialog box";
	    callback (update_of_string init_update, paths)
	  with _ ->
	    parse_dialog ~title:"Enter an update..." ~text:init_update ~parse:(fun s -> update_of_string s, paths) ~callback

	method copy_features =
	  let paths = view#selection#get_selected_rows in
	  let fs = List.map (fun path -> log_of_string (self # feature_from_path path)) paths in
	  cb_copy := Features fs

	method private refresh_function_iter iter (s : string) : unit =
	  model#set ~row:iter ~column:feature (utf0 s);
	  model#set ~row:iter ~column:count "";
	  model#set ~row:iter ~column:lift_label "";
	  model#set ~row:iter ~column:lift_color str_white;
	  model#set ~row:iter ~column:row_color "blue";
	  model#set ~row:iter ~column:row_scale 1.;
	  model#set ~row:iter ~column:bg_color str_white;
	  for i = model#iter_n_children (Some iter) - 1 downto 0 do
	    ignore (model#remove (model#iter_children ~nth:i (Some iter)))
	  done

	method private refresh_tree_iter iter (s,i as si) : unit = Common.prof "refresh_tree_iter" (fun () ->
	  model#set ~row:iter ~column:feature (utf0 s);
	  model#set ~row:iter ~column:incr i;
	  model#set ~row:iter ~column:count (string_of_int i.incr_supp);
	  model#set ~row:iter ~column:lift_label (self # lift_label_from_incr si);
	  model#set ~row:iter ~column:lift_color (self # lift_color_from_incr si);
	  model#set ~row:iter ~column:row_color (self # row_color_from_incr si);
	  model#set ~row:iter ~column:row_scale (self # row_scale_from_incr si);
	  model#set ~row:iter ~column:bg_color (self # bg_color_from_ext history#current#int_preselect i.incr_ext))

	method refresh_count path =
	  let s = self # feature_from_path path in
	  let i0 = self # incr_from_path path in
	  let i = get_incr s i0 in
	  let tree_iter = model#get_iter path in
	  self # refresh_tree_iter tree_iter (s,i);
	  self # expand_view false true tree_iter path (s,i)
	    
	method refresh_view force parent_opt (v,i) sv = Common.prof "navig.refresh_view" (fun () ->
	  let page_start, page_size = sv#links_page_start, sv#links_page_size in
	  let incrs = ls_opt force (parent_opt = None) sv (v,i) in
	  let nb_incrs = List.length incrs in
	  let visible_incrs = Common.sub_list incrs (page_start - 1) page_size in
	  sv # set_expanded true;
	  let expd = (List.length visible_incrs) = 1 in
	  let n_children = model#iter_n_children parent_opt in
	  let pos = ref (-1) in
	  let get_iter () =
	    Pervasives.incr pos;
	    if !pos < n_children
	    then model#iter_children ~nth:!pos parent_opt
	    else model#append ?parent:parent_opt () in
	  if nb_incrs > page_size then
	    self # refresh_function_iter (get_iter ())
	      (if page_start = 1
	      then function_bottom
	      else function_up);
	  List.iter
	    (fun (s,i) ->
	      let iter' = get_iter () in
	      self # refresh_tree_iter iter' (s,i);
	      self # expand_view expd force iter' (model#get_path iter') (s,i))
	    visible_incrs;
	  if nb_incrs > page_size then
	    self # refresh_function_iter (get_iter ())
	      (if nb_incrs > page_start - 1 + page_size
	      then function_down
	      else function_top);
	  for i = n_children - 1 downto !pos + 1 do
	    ignore (model#remove (model#iter_children ~nth:i parent_opt))
	  done)

	method private expand_view expd force tree_iter path (v,i) =
	  model#set ~row:(model#append ~parent:tree_iter ()) ~column:feature "";
	  let st = history # current in
	  let views = st # views in
	  let sv_opt =
	    try
	      let sv = Hashtbl.find views v in
	      if sv # expanded
	      then Some sv
	      else None
	    with _ ->
	      if expd  (* if we must expand tree_iter *)
	      then begin (* then we create a stateview for tree_iter in expanded_views *)
		let sv =
		  try
		    match model#iter_parent tree_iter with
		    | None -> st # copy_view (self # state_view)
		    | Some iter -> st # copy_view (Hashtbl.find views (self # feature_from_tree_iter iter))
		  with _ -> (* should not happen *)
		    st # new_view in
		Hashtbl.add views v sv;
		Some sv end
	      else None in
	  match sv_opt with
	  | None ->
	      view # collapse_row path
	  | Some sv ->
	      ignore (self # refresh_view force (Some tree_iter) (v,i) sv);
	      view # expand_row path

	method refresh force =  (* when force is true, links are re-computed *)
	  view#selection#unselect_all ();
	  self#refresh_view force None
	    (v,{incr_lift = {lift_supp = -1; lift_scale = 1.; lift_color = str_white};
		incr_supp = -1;
		incr_scale = 1.;
		incr_color = str_white;
		incr_ext = Ext.empty}) (* the -1 means undefined *)
	    (history # current # get_view v)

	method refresh_preselect unselect =
	  if unselect && view#selection#get_selected_rows <> [] then begin
	    ignore_selection_changed <- true;
	    view#selection#unselect_all ()
	  end;
	  if check_int_preselect#active
	  then
	    let int_preselect = history # current # int_preselect in
	    model # foreach
	      (fun path iter ->
		let e : Ext.t =
		  try (model#get ~row:iter ~column:incr).incr_ext
		  with exn -> Ext.empty in (* mystery: fails with exception, but [e] has the right value ! *)
		model#set ~row:iter ~column:bg_color (self # bg_color_from_ext int_preselect e);
		false)
	  else
	    model # foreach
	      (fun path iter ->
		model#set ~row:iter ~column:bg_color str_white;
		false)

	method expand_row ?options ?sort_fun ?page_size path =
	  let v = self # feature_from_path path in
	  let i = self # incr_from_path path in
	  let sv = self # state_view_from_path path in
	  Common.iter_option (fun options -> sv # set_links_options options) options; 
	  Common.iter_option (fun sort_fun -> sv # set_links_sort_fun sort_fun) sort_fun;
	  Common.iter_option (fun page_size -> sv # set_links_page_size page_size) page_size;
	  self # refresh_view false (Some (model#get_iter path)) (v,i) sv;
	  view # expand_row path

	method collapse_row path =
	  let v = self # feature_from_path path in
	  let views = history # current # views in
	  (try (Hashtbl.find views v) # set_expanded false with Not_found -> ());
(*	  Hashtbl.remove expanded_views v; *)
	  view # collapse_row path

	method page_top parent_opt =
	  let v = Common.fold_option (fun iter -> self # feature_from_path (model#get_path iter)) v parent_opt in
	  let sv = Hashtbl.find history#current#views v in
	  let new_offset = 1 in
	  sv # set_links_page_start new_offset;
	  match parent_opt with
	  | None -> self # refresh false
	  | Some iter -> self # refresh_view false (Some iter) (v, self # incr_from_path (model#get_path iter)) sv

	method page_up parent_opt =
	  let v = Common.fold_option (fun iter -> self # feature_from_path (model#get_path iter)) v parent_opt in
	  let sv = Hashtbl.find history#current#views v in
	  let offset = sv # links_page_start in
	  let maxobj = sv # links_page_size in
	  let new_offset =
	    let no = offset - maxobj in
	    if no < 1 then 1 else no in
	  sv # set_links_page_start new_offset;
	  match parent_opt with
	  | None -> self # refresh false
	  | Some iter -> self # refresh_view false (Some iter) (v, self # incr_from_path (model#get_path iter)) sv

	method page_down parent_opt =
	  let v = Common.fold_option (fun iter -> self # feature_from_path (model#get_path iter)) v parent_opt in
	  let sv = Hashtbl.find history#current#views v in
	  let offset = sv # links_page_start in
	  let maxobj = sv # links_page_size in
	  let nbobj = Common.fold_option (fun l -> List.length l) 0 sv#links_opt in
	  let new_offset =
	    let no = offset + maxobj in
	    if no > nbobj then offset else no in
	  sv # set_links_page_start new_offset;
	  match parent_opt with
	  | None -> self # refresh false
	  | Some iter -> self # refresh_view false (Some iter) (v, self # incr_from_path (model#get_path iter)) sv

	method page_bottom parent_opt =
	  let v = Common.fold_option (fun iter -> self # feature_from_path (model#get_path iter)) v parent_opt in
	  let sv = Hashtbl.find history#current#views v in
	  let maxobj = sv # links_page_size in
	  let nbobj = Common.fold_option (fun l -> List.length l) 0 sv#links_opt in
	  let new_offset = ((nbobj - 1) / maxobj) * maxobj + 1 in
	  sv # set_links_page_start new_offset;
	  match parent_opt with
	  | None -> self # refresh false
	  | Some iter -> self # refresh_view false (Some iter) (v, self # incr_from_path (model#get_path iter)) sv

	initializer
	  view#selection#set_mode `MULTIPLE;
	  col_lift#set_sizing `AUTOSIZE;
(*
	  col_lift#set_cell_data_func
	    cell_lift
	    (fun model iter ->
	      let lift_color = model # get ~row:iter ~column:lift_color in
	      cell_lift#set_properties [`CELL_BACKGROUND_GDK lift_color]);
*)
	  col_count#set_sizing `AUTOSIZE;
	  col_count#set_cell_data_func
	    cell_count
	    (fun model iter ->
	      let row_color = model # get ~row:iter ~column:row_color in
	      let bg_color =
		if check_ext_color#active && row_color = str_black
		then 
		  try history # current # ext_colors # get_ext_color (model # get ~row:iter ~column:incr).incr_ext
		  with _ -> str_white
		else str_white in
	      cell_count#set_properties [`CELL_BACKGROUND bg_color]);
(*
	  col_feature#set_cell_data_func
	    cell_feature
	    (fun model iter ->
	      let bg_color = model # get ~row:iter ~column:bg_color in
	      cell_feature#set_properties [`CELL_BACKGROUND_GDK bg_color]);
*)
	  view#append_column col_lift;
	  view#append_column col_count;
	  view#append_column col_feature;
	  view#set_expander_column (Some col_feature);

	  view#selection#connect#after#changed ~callback:self#selection_changed;
	  view#connect#after#row_activated ~callback:self#row_activated;
	  view#connect#row_expanded ~callback:(fun _ path -> self # expand_row path; callbacks # row_expanded path);
	  view#connect#row_collapsed ~callback:(fun _ path -> self # collapse_row path; callbacks # row_collapsed path);
	  view#event#connect#button_press ~callback:self#button_press;
	  view#drag#source_set ~modi:[`BUTTON1] ~actions:[`MOVE]
	    [ { Gtk.target = "STRING"; Gtk.flags = []; Gtk.info = 0} ];
	  view#drag#connect#data_get ~callback:(fun _ sel ~info ~time ->
	    self#copy_features;
	    view#selection#unselect_all ();
	    sel#return "");
	  view#drag#dest_set ~actions:[`MOVE]
	    [ { Gtk.target = "STRING"; Gtk.flags = []; Gtk.info = 0} ];
	  view#drag#connect#data_received ~callback:(fun context ~x ~y sel ~info ~time ->
	    begin try
	      view#selection#unselect_all ();
	      ( match view#get_path_at_pos ~x ~y with
	      | Some (path,_,_,_) -> view#selection#select_path path
	      | None -> ()
	       );
	      callbacks # data_received ()
	    with _ -> () end;
	    context#finish ~success:true ~del:false ~time);
	  ()
      end

    class multi_tree_features ~pack =
      let nb = GPack.notebook ~scrollable:true ~tab_pos:`LEFT ~homogeneous_tabs:true ~packing:pack () in
      let ht = Hashtbl.create 10 in
      object (self)
	val mutable callbacks = new tree_features_callbacks

	method connect c =
	  callbacks <- c;
	  Hashtbl.iter (fun i tf -> tf # connect c) ht

	method get_tree i = Hashtbl.find ht i

	method current = Hashtbl.find ht (nb # current_page)

	method selection_changed = self # current # selection_changed

	method feature_from_path = self # current # feature_from_path

	method ext_from_path = self # current # ext_from_path

	method state_view = self # current # state_view

	method state_view_from_path = self # current # state_view_from_path

	method selected_features = self # current # selected_features

	method selected_update = self # current # selected_update

	method copy_features = self # current # copy_features

	method refresh_count = self # current # refresh_count

	method refresh = self # current # refresh

	method refresh_preselect = self # current # refresh_preselect

	method expand_row = self # current # expand_row

	method collapse_row = self # current # collapse_row

	method add_tab v =
	  let label = GMisc.label ~text:v () in
	  let page = new tree_features
	      ~pack:(fun x -> ignore(nb#append_page ~tab_label:(label :> GObj.widget) x))
	      v in
	  page # connect callbacks;
	  Hashtbl.add ht (Hashtbl.length ht) (* nb # page_num page#widget *) page

	method remove_tab =
	  if self # current # root <> "all"
	  then begin
	    let i = nb # current_page in
	    nb # remove_page i;
	    Hashtbl.remove ht i;
	    for k = i to Hashtbl.length ht - 1 do
	      Hashtbl.add ht k (Hashtbl.find ht (k+1));
	      Hashtbl.remove ht (k+1)
	    done
	  end

	initializer
	  let v = "all" in
(*	  let sv = history # current # view in *)
	  self # add_tab v;
	  nb # connect # switch_page ~callback:(fun i -> (self # get_tree i) # refresh true);
	  ()
      end

    class list_objects_callbacks =
      object
	method row_activated : int -> unit = fun oid -> ()
	method ctx_menu : extent_tab -> Ext.t -> GMenu.menu -> unit = fun t oids menu -> ()
	method refresh_preselect : bool -> unit = fun _ -> ()
      end

    class list_objects_texto t name label ~pack =
      let vbox_extent = GPack.vbox ~spacing:4 ~packing:pack () in
      let hbox_extent_buttons = GPack.hbox ~spacing:2 ~packing:vbox_extent#pack () in
      let button_pagetop = GButton.button ~label:"|<" ~relief:`NONE (*~stock:`GOTO_TOP*) ~packing:hbox_extent_buttons#pack () in
      let button_pageup = GButton.button ~label:"<" ~relief:`NONE (*~stock:`GO_UP*) ~packing:hbox_extent_buttons#pack () in
      let label_extent = GMisc.label ~text:"No results" ~packing:hbox_extent_buttons#pack () in
      let button_pagebot = GButton.button ~label:">|" ~relief:`NONE (*~stock:`GOTO_BOTTOM*) ~packing:(hbox_extent_buttons#pack ~from:`END) () in
      let button_pagedown = GButton.button ~label:">" ~relief:`NONE (*~stock:`GO_DOWN*) ~packing:(hbox_extent_buttons#pack ~from:`END) () in
      let sw_extent = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:(vbox_extent#pack ~expand:true ~fill:true) () in
      let cols_extent = new GTree.column_list in
      let id = cols_extent#add int in
      let descr = cols_extent#add string in
(*      let bg_color = cols_extent#add string in *)
      let model_extent = GTree.list_store cols_extent in
      let view_extent = GTree.view ~model:model_extent ~headers_visible:false ~rules_hint:false ~packing:sw_extent#add (*_with_viewport*) () in
      let col_descr = GTree.view_column ~title:"Description" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",descr (*; "cell_background",bg_color*)]) in

      object (self)
	val mutable callbacks = new list_objects_callbacks

	method connect c = callbacks <- c

	val mutable ignore_selection_changed = false
	val mutable previous_selection_extent = []
	val mutable selection_extent = []

	method tab = t

	method scroll_changed () =
	  let stab = history # current # extent # tab t in
	  stab # set_page_scroll sw_extent#vadjustment#value

	method selection_changed () =
	  previous_selection_extent <- selection_extent;
	  selection_extent <- view_extent#selection#get_selected_rows;
	  if ignore_selection_changed
	  then ignore_selection_changed <- false
	  else begin
	    let new_preselection =
	      match selection_extent (*if check_int_preselect#active then selection_extent else []*) with
	      | [] -> None
	      | paths ->
		  let oids =
		    List.fold_left
		      ~f:(fun res path ->
			Ext.add (model_extent#get ~row:(model_extent#get_iter path) ~column:id) res)
		      ~init:Ext.empty
		      paths in
		  Some oids in
	    let st = history # current in
	    if new_preselection <> st # int_preselect
	    then begin
	      st # set_ext_preselect new_preselection; (* *)
	      st # set_int_preselect new_preselection;
	      if check_int_preselect#active then callbacks # refresh_preselect true
(*	      if check_ext_preselect#active then self # refresh_preselect false *)
	    end
	  end

	method row_activated path =
	  let oid = model_extent#get ~row:(model_extent#get_iter path) ~column:id in
	  callbacks # row_activated oid

	method button_press ev =
	  let button = GdkEvent.Button.button ev in
	  if button = 3
	  then begin
	    let oids = self#selected_oids () in
	    let menu = GMenu.menu () in
	    callbacks # ctx_menu t oids menu;
	    menu#popup ~button:button ~time:(GdkEvent.Button.time ev);
	    true end
	  else false

	method selected_oids () =
	  match selection_extent with
	  | [] ->
	      let st = history # current in
	      if st # extent # locals_only
	      then st # locals
	      else st # extent # ext
	  | paths ->
	      List.fold_left
		~f:(fun res path ->
		  Ext.add (model_extent#get ~row:(model_extent#get_iter path) ~column:id) res)
		~init:Ext.empty
		paths

	method copy_objects () =
	  let oids = self#selected_oids () in
	  cb_copy := Objects oids

(*
	method private bg_color_from_oid ext_preselect oid =
	  if not check_ext_preselect#active || Common.fold_option (Ext.mem oid) true ext_preselect
	  then str_white
	  else str_lightgrey
*)

	val mutable q = Queue.create ()
	val mutable former_visible_ext = []

	method refresh () =
	  let st = history # current in
	  let ext_preselect = st # ext_preselect in
	  let ext0 =
	    if st # extent # locals_only
	    then st # locals
	    else st # extent # ext in
	  let ext =
	    Common.mapfilter
	      (fun (_,_,oid) ->
		let d = descr_of_oid oid in
		match t, d with
		| String, Source.String d -> Some (oid,d)
		| Picture, Source.Picture (_,_) -> None
		| _ -> None)
	      (List.sort compare (* order objects by source url, then ref, and finally oid *)
		 (Ext.fold
		    (fun res oid ->
		      try match Context.origin_obj oid with
		      | Context.Object.Internal _ -> ("", "", oid) :: res
		      | Context.Object.External (src,ref) -> (Context.Src.url src, ref, oid) :: res
		      with Not_found ->
			Printf.printf "Unknown oid: %d\n" oid; flush stdout;
			res)
		    [] ext0)) in
	  let nbobj = List.length ext in
	  label # set_text (name ^ " (" ^ string_of_int nbobj ^ ")");
	  let ste = st # extent in
	  let stab = ste # tab t in
	  stab # set_nbobj nbobj;
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let visible_ext = Common.sub_list ext (offset-1) maxobj in
	  let label_text =
	    if nbobj = 0
	    then "No result"
	    else Printf.sprintf "Results: %d - %d / %d" offset (min nbobj (offset + maxobj - 1)) nbobj in
	  label_extent#set_text label_text;
	  if visible_ext <> former_visible_ext then begin
	    former_visible_ext <- visible_ext;
	    Queue.clear q;
	    q <- Queue.create ();
	    model_extent#clear ();
	    ignore (List.fold_left
		      ~f:(fun pos (oid,d) ->
			let item = model_extent#append () in
			model_extent#set ~row:item ~column:id oid;
(*			model_extent#set ~row:item ~column:bg_color (self # bg_color_from_oid ext_preselect oid); *)
			let d' =
			  if d = utf ""
			  then utf (string_of_int oid)
			  else d in
			model_extent#set ~row:item ~column:descr (utf0 d');
			pos+1)
		      ~init:1 visible_ext);
	    Queue.add (fun () -> sw_extent # vadjustment # set_value stab#page_scroll) q;
	    ignore (do_while_idle q) end
	  else
	    view_extent#selection#unselect_all ();
	  nbobj

	method refresh_preselect unselect =
	  if unselect && view_extent#selection#get_selected_rows <> [] then begin
	    ignore_selection_changed <- true;
	    view_extent#selection#unselect_all ()
	  end;
	  if check_ext_preselect#active
	  then
	    let ext_preselect = history # current # ext_preselect in
	    model_extent # foreach
	      (fun path iter ->
		let oid = model_extent#get ~row:iter ~column:id in
		let preselected =
		  Common.fold_option (Ext.mem oid) false ext_preselect in
(*		model_extent#set ~row:iter ~column:bg_color (if preselected then str_white else str_lightgrey); *)
		if preselected then begin
		  ignore_selection_changed <- true;
		  view_extent#selection#select_path path
		end;
		false)
	  else begin
	    ignore_selection_changed <- true;
	    view_extent#selection#unselect_all ()
	  end
(*
	    model_extent # foreach
	      (fun path iter ->
		model_extent#set ~row:iter ~column:bg_color str_white;
		false)
*)

	method pagetop () =
	  let stab = history # current # extent # tab t in
	  stab # set_page_start 1;
	  ignore (self # refresh ())

	method pagedown () =
	  let stab = history # current # extent # tab t in
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (* s # extent # nbobj in *)
	  let new_offset =
	    let no = offset + maxobj in
	    if no > nbobj then offset else no in
	  stab # set_page_start new_offset;
	  ignore (self # refresh ())

	method pageup () =
	  let stab = history # current # extent # tab t in
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (*  # extent # nbobj in *)
	  let new_offset =
	    let no = offset - maxobj in
	    if no < 1 then 1 else no in
	  stab # set_page_start new_offset;
	  ignore (self # refresh ())

	method pagebot () =
	  let stab = history # current # extent # tab t in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (*  # extent # nbobj in *)
	  stab # set_page_start (nbobj - ((nbobj - 1) mod maxobj));
	  ignore (self # refresh ())

	initializer
	  view_extent#selection#set_mode `MULTIPLE;
	  col_descr#set_sizing `AUTOSIZE;
	  view_extent#append_column col_descr;
	  button_pagetop#connect#clicked ~callback:self#pagetop;
	  tooltips#set_tip button_pagetop#coerce ~text:"Go to the top of the extent";
	  button_pageup#connect#clicked ~callback:self#pageup;
	  tooltips#set_tip button_pageup#coerce ~text:"Go one page upward in the extent";
	  button_pagedown#connect#clicked ~callback:self#pagedown;
	  tooltips#set_tip button_pagedown#coerce ~text:"Go one page downward in the extent";
	  button_pagebot#connect#clicked ~callback:self#pagebot;
	  tooltips#set_tip button_pagebot#coerce ~text:"Go to the bottom of the extent";
	  view_extent#selection#connect#after#changed ~callback:self#selection_changed;
	  view_extent#connect#after#row_activated ~callback:(fun path vcol -> self#row_activated path);
	  view_extent#event#connect#button_press ~callback:self#button_press;
	  view_extent#drag#source_set ~modi:[`BUTTON1] ~actions:[`MOVE]
	    [ { Gtk.target = "STRING"; Gtk.flags = []; Gtk.info = 0} ];
	  view_extent#drag#connect#data_get ~callback:(fun dc sel ~info ~time ->
	    assert (List.length selection_extent = 1);
	    if List.mem (List.hd selection_extent) previous_selection_extent
	    then selection_extent <- previous_selection_extent;
	    self#copy_objects ();
	    sel#return "");
	  ()
      end

    class list_objects_picto t name label ~pack =
      let vbox_extent = GPack.vbox ~spacing:4 ~packing:pack () in
      let hbox_extent_buttons = GPack.hbox ~spacing:2 ~packing:vbox_extent#pack () in
      let button_pagetop = GButton.button ~label:"|<" ~relief:`NONE (*~stock:`GOTO_TOP*) ~packing:hbox_extent_buttons#pack () in
      let button_pageup = GButton.button ~label:"<" ~relief:`NONE (*~stock:`GO_UP*) ~packing:hbox_extent_buttons#pack () in
      let label_extent = GMisc.label ~text:"No results" ~packing:hbox_extent_buttons#pack () in
      let button_pagebot = GButton.button ~label:">|" ~relief:`NONE (*~stock:`GOTO_BOTTOM*) ~packing:(hbox_extent_buttons#pack ~from:`END) () in
      let button_pagedown = GButton.button ~label:">" ~relief:`NONE (*~stock:`GO_DOWN*) ~packing:(hbox_extent_buttons#pack ~from:`END) () in
      let sw_extent = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:(vbox_extent#pack ~expand:true ~fill:true) () in
      let cols_extent = new GTree.column_list in
      let id = cols_extent#add int in
      let descr : string GTree.column = cols_extent#add string in
      let thumb_name = cols_extent#add string in
      let thumb_pixbuf : GdkPixbuf.pixbuf GTree.column = cols_extent#add (Gobject.Data.gobject_by_name "GdkPixbuf") in (*gobject : GdkPixbuf.pixbuf Gobject.data_conv*)
(*      let bg_color = cols_extent#add string in *)
      let model_extent = GTree.list_store cols_extent in
      let view_extent = GTree.icon_view ~model:model_extent ~orientation:`VERTICAL ~selection_mode:`MULTIPLE ~packing:sw_extent#add () in
      object (self)
	val mutable callbacks = new list_objects_callbacks

	method connect c = callbacks <- c

	val mutable ignore_selection_changed = false
	val mutable previous_selection_extent = []
	val mutable selection_extent = []

	method tab = t

	method scroll_changed () =
	  let stab = history # current # extent # tab t in
	  stab # set_page_scroll sw_extent#vadjustment#value

	method selection_changed () =
	  previous_selection_extent <- selection_extent;
	  selection_extent <- view_extent#get_selected_items;
	  if ignore_selection_changed
	  then ignore_selection_changed <- false
	  else begin
	    let new_preselection =
	      match selection_extent (*if check_int_preselect#active then selection_extent else []*) with
	      | [] -> None
	      | paths ->
		  let oids =
		    List.fold_left
		      ~f:(fun res path ->
			Ext.add (model_extent#get ~row:(model_extent#get_iter path) ~column:id) res)
		      ~init:Ext.empty
		      paths in
		  Some oids in
	    let st = history # current in
	    if new_preselection <> st # int_preselect
	    then begin
	      st # set_ext_preselect new_preselection; (* *)
	      st # set_int_preselect new_preselection;
	      if check_int_preselect#active then callbacks # refresh_preselect true
(*	      if check_ext_preselect#active then self # refresh_preselect false *)
	    end
	  end

	method row_activated path =
	  let oid = model_extent#get ~row:(model_extent#get_iter path) ~column:id in
	  callbacks # row_activated oid

	method button_press ev =
	  let button = GdkEvent.Button.button ev in
	  if button = 3
	  then begin
	    let oids = self#selected_oids () in
	    let menu = GMenu.menu () in
	    callbacks # ctx_menu t oids menu;
	    menu#popup ~button:button ~time:(GdkEvent.Button.time ev);
	    true end
	  else false

	method selected_oids () =
	  match selection_extent with
	  | [] ->
	      let st = history # current in
	      if st # extent # locals_only
	      then st # locals
	      else st # extent # ext
	  | paths ->
	      List.fold_left
		~f:(fun res path ->
		  Ext.add (model_extent#get ~row:(model_extent#get_iter path) ~column:id) res)
		~init:Ext.empty
		paths

	method copy_objects () =
	  let oids = self#selected_oids () in
	  cb_copy := Objects oids

(*
	method private compute_pixbuf filename item =
	  let pixbuf = pixbuf_of_picture ~maxside:150 filename in
	  model_extent#set ~row:item ~column:thumb_pixbuf pixbuf
*)
(*
	method private bg_color_from_oid ext_preselect oid =
	  if not check_ext_preselect#active || Common.fold_option (Ext.mem oid) true ext_preselect
	  then str_white
	  else str_lightgrey
*)

	val mutable q = Queue.create ()
	val mutable former_visible_ext = []

	method refresh () =
	  let st = history # current in
	  let ext_preselect = st # ext_preselect in
	  let ext0 =
	    if st # extent # locals_only
	    then st # locals
	    else st # extent # ext in
	  let ext =
	    Common.mapfilter
	      (fun (_,_,oid) ->
		let d = descr_of_oid oid in
		match t, d with
		| String, Source.String _ -> None
		| Picture, Source.Picture (f,d) -> Some (oid,f,d)
		| _ -> None)
	      (List.sort compare (* order objects by source url, then ref, and finally oid *)
		 (Ext.fold
		    (fun res oid ->
		      try match Context.origin_obj oid with
		      | Context.Object.Internal _ -> ("", "", oid) :: res
		      | Context.Object.External (src,ref) -> (Context.Src.url src, ref, oid) :: res
		      with Not_found ->
			Printf.printf "Unknown oid: %d\n" oid; flush stdout;
			res)
		    [] ext0)) in
	  let nbobj = List.length ext in
	  label # set_text (name ^ " (" ^ string_of_int nbobj ^ ")");
	  let ste = st # extent in
	  let stab = ste # tab t in
	  stab # set_nbobj nbobj;
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let visible_ext = Common.sub_list ext (offset-1) maxobj in
	  let label_text =
	    if nbobj = 0
	    then "No result"
	    else Printf.sprintf "Results: %d - %d / %d" offset (min nbobj (offset + maxobj - 1)) nbobj in
	  label_extent#set_text label_text;
	  if visible_ext <> former_visible_ext then begin
	    former_visible_ext <- visible_ext;
	    Queue.clear q;
	    q <- Queue.create ();
	    model_extent#clear ();
	    sw_extent # vadjustment # set_value 0.;
	    ignore (List.fold_left
		      ~f:(fun pos (oid,f,d) ->
			let item = model_extent#append () in
			model_extent#set ~row:item ~column:id oid;
(*			model_extent#set ~row:item ~column:bg_color (self # bg_color_from_oid ext_preselect oid); *)
			model_extent#set ~row:item ~column:thumb_name f;
			(try model_extent#set ~row:item ~column:thumb_pixbuf (pixbuf_of_picture ~maxside:150 f) with _ -> ());
			model_extent#set ~row:item ~column:descr (utf0 d);
(*
			if pos <= 5
			then self # compute_pixbuf f item
			else Queue.add (fun () -> self # compute_pixbuf f item) q;
*)
			pos+1)
		      ~init:1 visible_ext);
	    Queue.add (fun () -> sw_extent # vadjustment # set_value stab#page_scroll) q;
	    ignore (do_while_idle q) end
	  else
	    view_extent#unselect_all ();
	  nbobj

	method refresh_preselect unselect =
	  if unselect && view_extent#get_selected_items <> [] then begin
	    ignore_selection_changed <- true;
	    view_extent#unselect_all ()
	  end;
	  if check_ext_preselect#active
	  then
	    let ext_preselect = history # current # ext_preselect in
	    model_extent # foreach
	      (fun path iter ->
		let oid = model_extent#get ~row:iter ~column:id in
		let preselected =
		  Common.fold_option (Ext.mem oid) false ext_preselect in
(*		model_extent#set ~row:iter ~column:bg_color (if preselected then str_white else str_lightgrey); *)
		if preselected then begin
		  ignore_selection_changed <- true;
		  view_extent#select_path path
		end;
		false)
	  else begin
	    ignore_selection_changed <- true;
	    view_extent#unselect_all ()
	  end
(*
	    model_extent # foreach
	      (fun path iter ->
		model_extent#set ~row:iter ~column:bg_color str_white;
		false)
*)

	method pagetop () =
	  let stab = history # current # extent # tab t in
	  stab # set_page_start 1;
	  ignore (self # refresh ())

	method pagedown () =
	  let stab = history # current # extent # tab t in
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (* s # extent # nbobj in *)
	  let new_offset =
	    let no = offset + maxobj in
	    if no > nbobj then offset else no in
	  stab # set_page_start new_offset;
	  ignore (self # refresh ())

	method pageup () =
	  let stab = history # current # extent # tab t in
	  let offset = stab # page_start in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (*  # extent # nbobj in *)
	  let new_offset =
	    let no = offset - maxobj in
	    if no < 1 then 1 else no in
	  stab # set_page_start new_offset;
	  ignore (self # refresh ())

	method pagebot () =
	  let stab = history # current # extent # tab t in
	  let maxobj = stab # page_size in
	  let nbobj = stab # nbobj in (*  # extent # nbobj in *)
	  stab # set_page_start (nbobj - ((nbobj - 1) mod maxobj));
	  ignore (self # refresh ())

	initializer
	  view_extent#set_text_column descr;
	  view_extent#set_pixbuf_column thumb_pixbuf;
(*
	  col_thumb#set_sizing `AUTOSIZE;
	  col_thumb#set_cell_data_func
	    cell_thumb
	    (fun model iter ->
	      try
(*
		let colormap = Gdk.Color.get_system_colormap () in
		let black = Gdk.Color.alloc ~colormap `BLACK in
*)
		match model#get ~row:iter ~column:thumb_pixbuf with
		| Some pixbuf ->
		    let height = GdkPixbuf.get_height pixbuf in
		    let width = GdkPixbuf.get_width pixbuf in
		    cell_thumb#set_properties [`HEIGHT (height+20); `PIXBUF pixbuf; `VISIBLE true]
		| None ->
		    raise Not_found
	      with _ -> cell_thumb#set_properties [`VISIBLE false]);
	  col_descr#set_sizing `AUTOSIZE;
*)
	  button_pagetop#connect#clicked ~callback:self#pagetop;
	  tooltips#set_tip button_pagetop#coerce ~text:"Go to the top of the extent";
	  button_pageup#connect#clicked ~callback:self#pageup;
	  tooltips#set_tip button_pageup#coerce ~text:"Go one page upward in the extent";
	  button_pagedown#connect#clicked ~callback:self#pagedown;
	  tooltips#set_tip button_pagedown#coerce ~text:"Go one page downward in the extent";
	  button_pagebot#connect#clicked ~callback:self#pagebot;
	  tooltips#set_tip button_pagebot#coerce ~text:"Go to the bottom of the extent";
	  view_extent#connect#selection_changed ~callback:self#selection_changed;
	  view_extent#connect#item_activated ~callback:self#row_activated;
	  view_extent#event#connect#button_press ~callback:self#button_press;
	  view_extent#drag#source_set ~modi:[`BUTTON1] ~actions:[`MOVE]
	    [ { Gtk.target = "STRING"; Gtk.flags = []; Gtk.info = 0} ];
	  view_extent#drag#connect#data_get ~callback:(fun dc sel ~info ~time ->
	    assert (List.length selection_extent = 1);
	    if List.mem (List.hd selection_extent) previous_selection_extent
	    then selection_extent <- previous_selection_extent;
	    self#copy_objects ();
	    sel#return "");
	  ()
      end

    class multi_list_objects ~pack =
      let nb = GPack.notebook ~scrollable:true ~homogeneous_tabs:true ~packing:pack () in
      let ht = Hashtbl.create 10 in
      let create_tab_texto t name =
	let label = GMisc.label ~text:name () in
	let page =
	  new list_objects_texto t name label
	    ~pack:(fun x -> ignore (nb#append_page ~tab_label:(label :> GObj.widget) x)) in
	Hashtbl.add ht (Hashtbl.length ht) page
      in
      let create_tab_picto t name =
	let label = GMisc.label ~text:name () in
	let page =
	  new list_objects_picto t name label
	    ~pack:(fun x -> ignore (nb#append_page ~tab_label:(label :> GObj.widget) x)) in
	Hashtbl.add ht (Hashtbl.length ht) page
      in
      object (self)
	val mutable callbacks = new list_objects_callbacks

	method connect c =
	  callbacks <- c;
	  Hashtbl.iter (fun i tf -> tf # connect c) ht

 	method get_list i = Hashtbl.find ht i

	method current = Hashtbl.find ht (nb # current_page)

	method current_tab = self # current # tab

	method scroll_changed = self # current # scroll_changed

	method selection_changed = self # current # selection_changed

	method row_activated = self # current # row_activated

	method button_press = self # current # button_press

	method selected_oids = self # current # selected_oids

	method copy_objects = self # current # copy_objects

	method refresh () = (* self # current # refresh () *)
	  let n0 = (self # get_list 0) # refresh () in
	  let n1 = (self # get_list 1) # refresh () in
	  let i = nb#current_page in
	  if n0 = 0 && n1 > 0 && i = 0 then nb#goto_page 1
	  else if n1 = 0 && n0 > 0 && i = 1 then nb#goto_page 0
	  else ()

	method refresh_preselect = self # current # refresh_preselect

	method pagetop = self # current # pagetop

	method pagedown = self # current # pagedown

	method pageup = self # current # pageup

	method pagebot = self # current # pagebot

	initializer
	  create_tab_picto Picture "Picto";
	  create_tab_texto String "Texto";
(*	  nb # connect # switch_page ~callback:(fun i -> (self # get_list i) # refresh ()); *)
	  ()
      end


    let paned_global = GPack.paned `VERTICAL ~packing:(vbox#pack ~expand:true ~fill:true) ()

    let wq = new text_query ~width_chars:30 ~pack:(paned_global#pack1 ~resize:true ~shrink:true)

    let paned_sw = GPack.paned `HORIZONTAL ~packing:(paned_global#pack2 ~resize:true ~shrink:true) ()

    let vbox = GPack.vbox ~spacing:4 ~packing:(paned_sw#pack1 ~resize:true ~shrink:true) ()

    let nl = new navigation_link ~pack:vbox#pack

    let navig = new multi_tree_features ~pack:(vbox#pack ~expand:true ~fill:true)

    let extent = new multi_list_objects ~pack:(paned_sw#pack2 ~resize:true ~shrink:true)


(* global *)

    let refresh force =
      history#current#extent#refresh;
      wq#refresh;
      navig#refresh force;
      extent#refresh ()

    let refresh_force () = refresh true
    let refresh_counts paths () = 
      if paths = []
      then navig # refresh true
      else List.iter navig#refresh_count paths

    let awaken f final = (* forcing the updating process to work :-) *)
      ignore (GMain.Idle.add (fun () ->
	let n = Context.nb_updates () in
	entry_nbupdates#set_text (utf (string_of_int n));
	if n=0 then final ();
	n > 0));
      ignore (f ())

    let import options ctxname =
      let get_bool valr = valr = "on" || valr = "yes" || valr = "true" in
      let ch = open_in_bin (ctxname ^ ".ctx") in
      (try while true do
	let line = input_line ch in
	if String.sub line 0 4 <> "set "
	then failwith "" (* rest to be done by Context.import *)
	else
	  try
	    let ok = Str.string_match (Str.regexp "set[ ]+\\([_a-zA-Z]+\\)[ ]+\\([-_a-zA-Z0-9]+\\)\\([ ]+at[ ]+\\(.+\\)\\)?[ ]*$") line 0 in
	    if not ok then failwith "";
	    let attr = Str.matched_group 1 line in
	    let valr = Str.matched_group 2 line in
	    let constr = try Str.matched_group 4 line with Not_found -> "" in
	    let v = try string_of_feat (log_of_string constr) with _ -> "all" in (* view name *)
	    match attr with
	    | "variable_font_size" -> check_scale#set_active (get_bool valr)
	    | "foreground_color" -> check_feat_color#set_active (get_bool valr)
	    | "total_count" -> check_lift#set_active (get_bool valr)
	    | "total_count_grey_level" -> check_lift_color#set_active (get_bool valr)
	    | "equal_extent_colors" -> check_ext_color#set_active (get_bool valr)
	    | "objects_preselection" -> check_ext_preselect#set_active (get_bool valr)
	    | "features_preselection" -> check_int_preselect#set_active (get_bool valr)
	    | "local_objects_only" -> history#home#extent#set_locals_only (get_bool valr)
	    | "texto_page_size" -> (history#home#extent#tab String)#set_page_size (int_of_string valr)
	    | "picto_page_size" -> (history#home#extent#tab Picture)#set_page_size (int_of_string valr)
	    | "feature_sorting" -> (history#home#get_view v)#set_links_sort_fun 
		  (if valr = "logical" then sort_by_feature
		  else if valr = "decreasing" then sort_by_count
		  else failwith "")
	    | "feature_conceptual_ordering" -> (history#home#get_view v)#set_links_conceptual_ordering (get_bool valr)
	    | "feature_minimum_support" -> (history#home#get_view v)#set_links_minimum_support
		  (try Some (int_of_string valr) with _ -> None)
	    | "feature_page_size" -> (history#home#get_view v)#set_links_page_size (int_of_string valr)
	    | "feature_expanded" -> (history#home#get_view v)#set_to_be_expanded (get_bool valr)
	    | _ -> failwith ""
	  with
	  | Failure "" -> failwith ""
	  | exn ->
	      prerr_endline ("Invalid set command: " ^ line);
	      prerr_endline (Printexc.to_string exn)
      done with _ -> ());
      close_in ch;
      Context.import options ctxname

    let export options ctxname =
      if confirm_overwrite (ctxname ^ ".ctx")
      then Context.export options ctxname

    let init filename =
      let f = try Filename.chop_extension filename with _ -> filename in
      Context.init ~log:!log f

    let load filename =
      let f = try Filename.chop_extension filename with _ -> filename in
      if Filename.check_suffix filename "lis" then
	try Context.load ~readonly:!readonly f
	with _ ->
	  if Sys.file_exists (f ^ ".ctx")
	  then
	    if question "The .lis file could not be open because of version mismatch, \
                         but there is a .ctx associated file. Do you want to recover your data \
                         from this file (it may take a while) ?"
	    then awaken (fun () -> import [] f) refresh_force
	    else ()
	  else error_message "The .lis file could not be open because of version mismatch,\
                              and there is no .ctx associated file, sorry."
      else error_message "This file has not extension .lis. Camelis cannot open it."

    let save filename =
      try
	let f = try Filename.chop_extension filename with _ -> filename in
	Context.save f
(*
	Context.wq_incr (fun _ -> Query.top); (* in order to save everything with export *)
	Context.export [] f;
	Context.wq_back (); (* restore current query *)
*)
      with
      | Sys_error msg -> error_message msg
      | Failure msg -> error_message msg
      | e -> error_message (Printexc.to_string e)

    let close () =
      Context.init ""

    let init_source url options wq_as_home =
      let home = if wq_as_home then Context.descr_of_query (Context.wq ()) else Context.Object.Descr.empty in
      awaken (fun () -> try Context.init_source url options home with _ -> ()) refresh_force

    let init_well url =
      let q = Context.wq () in
      Context.init_well url q


(* Event handling *)

    let menu_paste_intr () =
      match !cb_copy with
      | Nothing -> ()
      | Features fs -> ()
      | Objects oids ->
	  navig#selected_update
	    ~callback:(fun (u, paths) ->
	      awaken
		(fun () -> Ext.iter (fun oid -> Context.update_obj oid (Context.Intr u)) oids)
		(refresh_counts paths)) ()
      | Query q ->
	  navig#selected_update
	    ~callback:(fun (u, paths) ->
	      awaken
		(fun () -> ignore (Context.init_rule q (Context.Intr u)))
		(refresh_counts paths)) ()
      | Sources srcs -> ()
      | Wells wlls -> ()

    let menu_paste_aux get_update =
      match !cb_copy with
      | Nothing -> ()
      | Features fs ->
	  get_update
	    ~callback:(fun (u,_) ->
	      let gs = match Context.Root.Update.to_descr u with [] -> [Context.Feat.top ()] | gs -> gs in
	      List.iter (fun f -> List.iter (fun g -> Context.add_axiom f g) gs) fs;
	      refresh true) ()
      | Objects oids ->
	  get_update
	    ~callback:(fun (u, paths) ->
	      awaken
		(fun () -> Ext.iter (fun oid -> Context.update_obj oid (Context.Extr u)) oids)
		(refresh_counts paths)) ()
      | Query q ->
	  get_update
	    ~callback:(fun (u, paths) ->
	      awaken
		(fun () -> ignore (Context.init_rule q (Context.Extr u)))
		(refresh_counts paths)) ()
      | Sources srcs ->
	  get_update
	    ~callback:(fun (u, paths) ->
	      awaken
		(fun () -> List.iter (fun src -> try Context.move_source src u with _ -> ()) srcs)
		(refresh_counts paths)) ()
      | Wells wlls ->
	  let q = Context.wq () in
	  awaken
	    (fun () -> List.iter (fun wll -> Context.move_well wll q) wlls)
	    refresh_force

    let menu_paste ?neg () =
      menu_paste_aux (navig#selected_update ~interactive:false ?neg)

    let menu_paste_to () =
      menu_paste_aux (navig#selected_update ~interactive:true ~neg:false)

(*
      let default_update =
	let u, paths = navig#selected_update () in
	if paths = []
	then None
	else Some (string_of_update u) in
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Paste to..."
	  () in
      let vbox = w#vbox in
      let entry = GEdit.entry ?text:default_update ~width_chars:80 ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      let action () =
	try
	  menu_paste_aux (fun () -> update_of_string entry#text, []);
	  w#destroy ()
	with
	| Stream.Error msg
	| Failure msg -> syntax_error_logic msg in
      entry#connect#activate ~callback:action;
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:action;
      w#show ()
*)

(*
    let menu_replace () =
      let feature = string_of_update (fst (navig#selected_update ())) in
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Replace..."
	  () in
      let vbox = w#vbox in
      let entry = GEdit.entry ~text:feature ~width_chars:80 ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      let action () =
	try
	  menu_paste_aux (fun () -> update_of_string ("not " ^ feature ^ " and " ^ entry#text), []);
	  w#destroy ()
	with
	| Stream.Error msg
	| Failure msg -> syntax_error_logic msg in
      entry#connect#activate ~callback:action;
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:action;
      w#show ()
*)

    let menu_new () =
      file_dialog ~title:"Create a LIS context" ~filter:filter_lis ?filename:camelisdata
	~callback:(fun name ->
	  if confirm_overwrite name then begin
	    init name;
	    window#set_title (utf (title ()));
	    history # init;
	    refresh true end) ()

    let menu_open () =
      file_dialog ~title:"Open a LIS context" ~filter:filter_lis ?filename:camelisdata
	~callback:(fun name ->
	  load name;
	  window#set_title (utf (title ()));
	  history # init;
	  awaken (fun () -> ()) refresh_force)
	()

    let menu_saveas () =
      file_dialog ~title:"Save a LIS context" ~filter:filter_lis ?filename:camelisdata
	~callback:(fun name ->
	  if confirm_overwrite name then begin
	    save name;
	    load name;
	    window#set_title (utf (title ()))
	  end) ()

    let menu_save () =
      match Context.filename () with
      | "" -> menu_saveas ()
      | f -> save f

    let menu_close () =
      close ();
      window#set_title (utf (title ()));
      refresh true

    let dialog_ctx_options f =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Options for context import/export"
	  () in
      let vbox = w#vbox in
      let check_sources = GButton.check_button ~label:"sources (src)" ~active:true ~packing:vbox#add () in
      let check_terms = GButton.check_button ~label:"axioms (axiom)" ~active:true ~packing:vbox#add () in
      let check_extrinsic = GButton.check_button ~label:"extrinsic data (mk, mv)" ~active:true ~packing:vbox#add () in
      let check_features = GButton.check_button ~label:"features (show, hide)" ~active:true  ~packing:vbox#add () in
      let check_wells = GButton.check_button ~label:"wells (wll)" ~active:true ~packing:vbox#add () in
      let check_rules = GButton.check_button ~label:"rules (rule)" ~active:true ~packing:vbox#add () in
      let check_actions = GButton.check_button ~label:"actions (action)" ~active:true ~packing:vbox#add () in
      let hbox_file = GPack.hbox ~packing:vbox#pack () in
      let entry_file = GEdit.entry ~width_chars:60 ~packing:hbox_file#pack () in
      let button_file = GButton.button ~stock:`FIND ~packing:hbox_file#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      button_file#connect#clicked ~callback:
	(file_dialog ~title:"Import context" ~filename:!last_dir ~filter:filter_ctx
	   ~callback:(fun name ->
	     last_dir := Filename.dirname name ^ dir_sep;
	     entry_file#set_text (utf name)));
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:(fun () ->
	try
	  let options =
	    (if check_sources#active then [`Sources] else []) @
	    (if check_terms#active then [`Axioms] else []) @
	    (if check_extrinsic#active then [`Extrinsic] else []) @
	    (if check_features#active then [`Features] else []) @
	    (if check_wells#active then [`Wells] else []) @
            (if check_rules#active then [`Rules] else []) @
            (if check_actions#active then [`Actions] else []) in
	  let filename = loc entry_file#text in
	  if Filename.check_suffix filename "ctx"
	  then f options (Filename.chop_suffix filename "ctx")
	  else raise Not_found;
	  w#destroy ()
	with _ -> ());
      w#show ()

    let menu_import_ctx () =
      dialog_ctx_options import

    let menu_import_file () =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Options for file import"
	  () in
      let vbox = w#vbox in
      let hbox_file = GPack.hbox ~packing:vbox#pack () in
      let entry_file = GEdit.entry ~width_chars:60 ~packing:hbox_file#pack () in
      let button_file = GButton.button ~stock:`FIND ~packing:hbox_file#add () in
      let check_recursive = GButton.check_button ~label:"recursive traversal of directories" ~active:true ~packing:vbox#add () in
      let check_parts = GButton.check_button ~label:"extract file parts when applicable" ~active:false ~packing:vbox#add () in
      let check_suffixes = GButton.check_button ~label:"filter files from their suffix (dot-separated list)" ~active:false ~packing:vbox#add () in
      let entry_suffixes = GEdit.entry ~editable:false ~width_chars:40 ~packing:vbox#add () in
      let check_wq_as_home = GButton.check_button ~label:"put objects in current query" ~active:false ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      check_suffixes#connect#toggled ~callback:(fun () -> entry_suffixes#set_editable check_suffixes#active);
      button_file#connect#clicked ~callback:
	(file_dialog ~title:"Import file" ~filename:!last_dir
	   ~callback:(fun name ->
	     last_dir := Filename.dirname name ^ dir_sep;
	     entry_file#set_text (utf name)));
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:(fun () ->
	try
	  let options =
	    (if check_parts#active then [`Parts] else []) @
	    (if check_recursive#active then [`Recursive] else []) @
	    (if check_suffixes#active then [`Suffixes (Str.split (Str.regexp "\\.") entry_suffixes#text)] else []) in
	  if entry_file#text = "" then raise Not_found;
	  init_source (loc entry_file#text) options check_wq_as_home#active;
	  w#destroy ()
	with _ -> ());
      w#show ()

    let menu_import_update () =
      awaken
	(fun () ->
	  Context.update_monitor ();
	  Context.fold_sources_all (fun url () -> Context.update_source url) ())
	refresh_force

    let menu_import_list () =
      let ht = Hashtbl.create 100 in (* for recording expanded dirs *)
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Source tree"
	  () in
      let vbox = w#vbox in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:vbox#add () in
      let cols = new GTree.column_list in
      let dir = cols#add string in
      let descr = cols#add string in
      let row_color = cols#add string in
      let query = cols#add string in
      let model = GTree.tree_store cols in
      let view = GTree.view ~model ~headers_visible:true ~height:300 ~rules_hint:true ~packing:sw#add () in
      view#selection#set_mode `MULTIPLE;
      let col_dir = GTree.view_column ~title:"Location" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",dir; "foreground",row_color]) in
      let col_query = GTree.view_column ~title:"Home" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",query]) in
      view#append_column col_dir;
      view#append_column col_query;
      let rec add_src url par last current =
	match last, current with
	| _, [] ->
	    model#set ~row:par ~column:descr url;
	    model#set ~row:par ~column:row_color
	      (if Sys.file_exists url then str_darkgreen else str_darkred);
	    model#set ~row:par ~column:query (utf0 (string_of_descr (Context.home_source url)));
	    []
	| [], dir2::l2 ->
	    let item = model#append ~parent:par () in
	    model#set ~row:item ~column:dir dir2;
	    model#set ~row:item ~column:row_color str_black;
	    model#set ~row:item ~column:descr "";
	    let next = add_src url item [] l2 in
	    (dir2,item)::next
	| (dir1,item1)::l1, dir2::l2 ->
	    if dir1 = dir2
	    then
	      let next = add_src url item1 l1 l2 in
	      (dir1,item1)::next
	    else begin
	      let item = model#append ~parent:par () in
	      model#set ~row:item ~column:dir dir2;
	      model#set ~row:item ~column:row_color str_black;
	      model#set ~row:item ~column:descr "";
	      let next = add_src url item [] l2 in
	      (dir2,item)::next end in
      let refresh () =
	let urls = Context.fold_sources_all
	    (fun url res -> Common.insert (<=) url res)
	    [] in
(*	let urls = List.sort compare (Context.get_sources_all ()) in *)
	model#clear ();
	let root = model#append () in
	model#set ~row:root ~column:dir dir_sep;
	model#set ~row:root ~column:row_color str_black;
	model#set ~row:root ~column:descr "";
	ignore (List.fold_left
	  ~f:(fun acc url ->
	    add_src url root acc (Str.split (Str.regexp "/") url))
	  ~init:[]
	  urls);
	model#foreach
	  (fun path iter ->
	    let dirname = model#get ~row:iter ~column:dir in
	    if Hashtbl.mem ht dirname then view#expand_row path;
	    false) in
      let selected_srcs () =
	List.fold_left
	  ~f:(fun res path ->
	    let d = model#get ~row:(model#get_iter path) ~column:descr in
	    if d = "" then res else d::res)
	  ~init:[]
	  view#selection#get_selected_rows in

      let action_area = w#action_area in
      let button_refresh = GButton.button ~stock:`REFRESH ~packing:action_area#add () in
      let button_add = GButton.button ~stock:`ADD ~packing:action_area#add () in
      let button_move = GButton.button ~stock:(`STOCK "Move") ~packing:action_area#add () in
      let button_paste = GButton.button ~stock:`PASTE ~packing:action_area#add () in
      let button_delete = GButton.button ~stock:`DELETE ~packing:action_area#add () in
      let button_clean = GButton.button ~stock:`CLEAR ~packing:action_area#add () in
      let button_update = GButton.button ~stock:(`STOCK "Update") ~packing:action_area#add () in
      let button_close = GButton.button ~stock:`CLOSE ~packing:action_area#add () in

      view#connect#row_expanded
	~callback:(fun iter path ->
	  let dirname = model#get ~row:iter ~column:dir in
	  Hashtbl.replace ht dirname ());
      view#connect#row_collapsed
	~callback:(fun iter path ->
	  let dirname = model#get ~row:iter ~column:dir in
	  Hashtbl.remove ht dirname);
      button_refresh#connect#clicked ~callback:refresh;
      button_add#connect#clicked ~callback:(fun () ->
	menu_import_file ();
	refresh ());
      button_move#connect#clicked ~callback:(fun () ->
	( match selected_srcs () with
	| [url] ->
	    file_dialog ~title:"Move source" ~filename:url
	      ~callback:(fun new_url ->
		last_dir := Filename.dirname new_url ^ dir_sep;
		if confirm_overwrite new_url then
		  awaken
		    (fun () -> Context.relocate_source url new_url)
		    refresh)
	      ()
	| [] -> error_message "A source must be selected."
	| _ -> error_message "Only one source must be selected.");
	refresh ());
      button_paste#connect#clicked ~callback:(fun () ->
	cb_copy := Sources (selected_srcs ());
	menu_paste ());
      button_delete#connect#clicked ~callback:(fun () ->
	awaken
	  (fun () -> List.iter (fun src -> try Context.final_source src with _ -> ()) (selected_srcs ()))
	  refresh);
      button_clean#connect#clicked ~callback:(fun () ->
	let lost_sources =
	  Context.fold_sources_all
	    (fun url res -> if Sys.file_exists url then res else url::res)
	    [] in
	awaken
	  (fun () -> List.iter (fun src -> try Context.final_source src with _ -> ()) lost_sources)
	  refresh);
      button_update#connect#clicked ~callback:(fun () ->
	awaken
	  (fun () -> List.iter (fun src -> try Context.update_source src with _ -> ()) (selected_srcs ()))
	  refresh);
      button_close#connect#clicked ~callback:w#destroy;
      refresh ();
      w#show ()

    let menu_export_ctx options =
      dialog_ctx_options export

    let menu_export_as_binary () =
      file_dialog ~title:"Export as binary context" ~filter:filter_bin
	~callback:(fun name ->
	  if confirm_overwrite name then begin
	    let oids = extent # selected_oids () in
	    let props = List.map
		(fun path -> query_of_string (navig # feature_from_path path))
		navig#selected_features in
	    Context.export_as_binary oids props name
	  end)
	()

    let menu_export_file () =
      let cs = contents_of_extent history#current#extent#ext (* Context.extent () *) in
      let kinds = Context.Wll.kinds cs in
      match kinds with
      | [] -> info_message "No export available for the selected objects."
      | (ext,label)::ks ->
	  let w = GWindow.dialog
	      ~destroy_with_parent:true
	      ~title:"Exporting a file"
	      () in
	  let vbox = w#vbox in
	  let label1 = GMisc.label ~text:"Select the type of the file to be exported:" ~xalign:0. ~packing:vbox#add () in
	  let (radio1,ext1) = GButton.radio_button ~label ~packing:vbox#add (), ext in
	  let group = radio1#group in
	  let l =
	    (radio1,ext1) ::
	    List.map
	      (fun (ext,label) ->
		(GButton.radio_button ~group ~label ~packing:vbox#add (), ext))
	      ks in
	  let label2 = GMisc.label ~text:"\nEnter the location of the file to be exported:" ~xalign:0. ~packing:vbox#add () in
	  let hbox_file = GPack.hbox ~packing:vbox#pack () in
	  let entry_file = GEdit.entry ~width_chars:60 ~packing:hbox_file#pack () in
	  let button_file = GButton.button ~stock:`FIND ~packing:hbox_file#add () in
	  let action_area = w#action_area in
	  let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
	  let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
	  button_file#connect#clicked ~callback:(fun () ->
	    try
	      let ext = try Some (snd (List.find ~f:(fun (radio,ext) -> radio#active) l)) with _ -> None in
	      file_dialog ~title:"Export file" ~filename:!last_dir
		?filter:(Common.map_option (fun s -> "*." ^ s) ext)
		~callback:(fun name ->
		  last_dir := Filename.dirname name ^ dir_sep;
		  if confirm_overwrite name then
		    entry_file#set_text (utf name)) ()
	    with Not_found ->
	      info_message "Select a file type");
	  button_cancel#connect#clicked ~callback:w#destroy;
	  button_ok#connect#clicked ~callback:(fun () ->
	    try
	      match entry_file#text with
	      | "" -> raise Not_found
	      | url ->
		  if confirm_overwrite url then begin
		    init_well url;
		    w#destroy () end
	    with e ->
	      error_message (Printexc.to_string e));
	  w#show ()

    let menu_export_list () =
      let ht = Hashtbl.create 100 in (* for recording expanded dirs *)
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"List wells"
	  () in
      let vbox = w#vbox in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:vbox#add () in
      let cols = new GTree.column_list in
      let dir = cols#add string in
      let descr = cols#add string in
      let query = cols#add string in
      let row_color = cols#add string in
      let model = GTree.tree_store cols in
      let view = GTree.view ~model ~headers_visible:true ~height:300 ~rules_hint:true ~packing:sw#add () in
      view#selection#set_mode `MULTIPLE;
      let col_dir = GTree.view_column ~title:"Location" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",dir; "foreground",row_color]) in
      let col_query = GTree.view_column ~title:"Query" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",query]) in
      view#append_column col_dir;
      view#append_column col_query;

      let rec add_wll (url,q) par last current =
	match last, current with
	| _, [] ->
	    model#set ~row:par ~column:descr url;
	    model#set ~row:par ~column:query (utf0 (string_of_query q));
	    model#set ~row:par ~column:row_color
	      (if Sys.file_exists url then str_darkgreen else str_darkred);
	    []
	| [], dir2::l2 ->
	    let item = model#append ~parent:par () in
	    model#set ~row:item ~column:dir dir2;
	    model#set ~row:item ~column:row_color str_black;
	    model#set ~row:item ~column:descr "";
	    model#set ~row:item ~column:query "";
	    let next = add_wll (url,q) item [] l2 in
	    (dir2,item)::next
	| (dir1,item1)::l1, dir2::l2 ->
	    if dir1 = dir2
	    then
	      let next = add_wll (url,q) item1 l1 l2 in
	      (dir1,item1)::next
	    else begin
	      let item = model#append ~parent:par () in
	      model#set ~row:item ~column:dir dir2;
	      model#set ~row:item ~column:row_color str_black;
	      model#set ~row:item ~column:descr "";
	      model#set ~row:item ~column:query "";
	      let next = add_wll (url,q) item [] l2 in
	      (dir2,item)::next end in
      let refresh () =
	let wlls = List.sort compare (List.map ~f:(fun (q,url) -> (url,q)) (Context.get_wells_all ()))  in
	model#clear ();
	let root = model#append () in
	model#set ~row:root ~column:dir dir_sep;
	model#set ~row:root ~column:row_color str_black;
	model#set ~row:root ~column:descr "";
	model#set ~row:root ~column:query "";
	ignore (List.fold_left
	  ~f:(fun acc (url,q) ->
	    add_wll (url,q) root acc (Str.split (Str.regexp "/") url))
	  ~init:[]
	  wlls);
	model#foreach
	  (fun path iter ->
	    let dirname = model#get ~row:iter ~column:dir in
	    if Hashtbl.mem ht dirname then view#expand_row path;
	    false) in
      let selected_wlls () =
	List.fold_left
	  ~f:(fun res path ->
	    let d = model#get ~row:(model#get_iter path) ~column:descr in
	    if d = "" then res else d::res)
	  ~init:[]
	  view#selection#get_selected_rows in

      let action_area = w#action_area in
      let button_refresh = GButton.button ~stock:`REFRESH ~packing:action_area#add () in
      let button_add = GButton.button ~stock:`ADD ~packing:action_area#add () in
      let button_move = GButton.button ~stock:(`STOCK "Move") ~packing:action_area#add () in
      let button_paste = GButton.button ~stock:`PASTE ~packing:action_area#add () in
      let button_delete = GButton.button ~stock:`DELETE ~packing:action_area#add () in
      let button_close = GButton.button ~stock:`CLOSE ~packing:action_area#add () in

      view#connect#row_expanded
	~callback:(fun iter path ->
	  let dirname = model#get ~row:iter ~column:dir in
	  Hashtbl.replace ht dirname ());
      view#connect#row_collapsed
	~callback:(fun iter path ->
	  let dirname = model#get ~row:iter ~column:dir in
	  Hashtbl.remove ht dirname);
      button_refresh#connect#clicked ~callback:refresh;
      button_add#connect#clicked ~callback:(fun () ->
	menu_export_file ();
	refresh ());
      button_move#connect#clicked ~callback:(fun () ->
	( match selected_wlls () with
	| [url] ->
	    file_dialog ~title:"Move well" ~filename:url
	      ~callback:(fun name ->
		last_dir := Filename.dirname name ^ dir_sep;
		if confirm_overwrite name then 
		  Context.relocate_well url name) ()
	| [] -> error_message "A well must be selected."
	| _ -> error_message "Only one well must be selected.");
	refresh ());
      button_paste#connect#clicked ~callback:(fun () ->
	cb_copy := Wells (selected_wlls ());
	menu_paste ());
      button_delete#connect#clicked ~callback:(fun () ->
	List.iter Context.final_well (selected_wlls ());
	refresh ());
      button_close#connect#clicked ~callback:w#destroy;
      refresh ();
      w#show ()

    let menu_stat () =
      let st = Context.get_stat () in
      let nb_obj = st.Context.nb_obj in
      let nb_feat = st.Context.nb_node in
      let size_ctx = st.Context.sum_ext in
      let size_mem = st.Context.sum_ext_words in
      let feats_per_obj = if nb_obj = 0 then None else Some (size_ctx / nb_obj) in
      let msg = String.concat "\n" [
	"number of objects: " ^  string_of_int nb_obj;
	"number of features: " ^ string_of_int nb_feat;
	"context size: " ^ string_of_int size_ctx ^ " object-feature (" ^ string_of_int size_mem ^ " words)";
	"average number of features per object: " ^ Common.fold_option string_of_int "undefined" feats_per_obj
      ] in
      info_message msg

    let menu_quit () =
      if question "Do you really want to quit?"
      then begin
	Context.init "";
	GMain.quit ();
	Printf.printf "Profiling...\n";
	Hashtbl.iter
	  (fun f (n,t,m) -> Printf.printf "%s: %d calls, %.1f seconds\n" f n t)
	  Common.tbl_prof
      end

    let menu_home () =
      cd_abs "all";
      extent#scroll_changed ();
      history # after_cd_home;
      refresh true

    let menu_current_state () =
      cd_abs (Syntax.stringizer (Query.print (Context.home_monitor ())));
      extent#scroll_changed ();
      history # after_cd;
      refresh true

    let menu_back () =
      extent#scroll_changed ();
      history # backward;
      refresh false

    let menu_fwd () =      
      extent#scroll_changed ();
      history # forward;
      refresh false

    let menu_refresh () =
      refresh true

    let entry_wq_activated path =
      try
	extent#scroll_changed ();
	cd_abs path;
	history # after_cd;
	refresh true
      with Stream.Error msg -> syntax_error_logic msg

    type link_type = Rel | Abs

    type link =
	{ link_type : link_type;
	  global_transf : string -> string;
	    local_transf : string -> string
	}

    let link rel fg fl = {link_type = rel; global_transf = fg; local_transf = fl}

    let f_id incr = incr
    let f_and incr = incr
    let f_not incr = "not " ^ incr
    let f_compl incr = "compl"
    let f_if incr = "if"
    let f_unif incr = "unif"
    let f_ge incr = Str.global_replace (Str.regexp "=") ">=" incr
    let f_le incr = Str.global_replace (Str.regexp "=") "<=" incr

    let row_activated f paths =
      let features = List.map (fun path -> f.local_transf (navig#feature_from_path path)) paths in
      let union_incr = 
	match features with
	| [] -> "none"
	| [x] -> x
	| _ -> "(" ^ String.concat " or " features ^ ")" in
      let incr = f.global_transf union_incr in
      ( match f.link_type with
      | Rel -> cd incr
      | Abs -> cd_abs incr);
      extent#scroll_changed ();
      history # after_cd;
      refresh true

    let menu_cd (f : link) () =
      let paths = navig # selected_features in
      row_activated f paths

    let menu_expand_as_tab () =
      let paths = navig # selected_features in
      List.iter
	(fun path ->
	  let v = navig # feature_from_path path in
(*	  let sv = navig # state_view_from_path path in *)
	  navig # expand_row path;
	  navig # add_tab v;
	  navig # collapse_row path)
	paths

    let menu_collapse_current_tab () =
      navig # remove_tab

    let menu_expand () =
      let paths = navig # selected_features in
      let sv =
	match paths with
	| [] -> navig # state_view
	| path::_ -> navig # state_view_from_path path in
      let option x = List.mem x sv # links_options in
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Expanding a view with options"
	  () in
      let vbox = w#vbox in
      let check_lexicalorder = GButton.check_button ~label:"logical (vs. decreasing count) sorting of increments"
	  ~active:(sv # links_sort_fun==sort_by_feature) ~packing:vbox#add () in
(* deprecated
      let check_mostspec = GButton.check_button ~label:"most specific refinements" ~active:(option `MostSpec) ~packing:vbox#add () in
      let check_valonly = GButton.check_button ~label:"object descriptors only" ~active:(option `ValOnly) ~packing:vbox#add () in
      let check_noviews = GButton.check_button ~label:"strict refinements" ~active:(option `NoViews) ~packing:vbox#add () in
*)
      let check_maxonly = GButton.check_button ~label:"conceptual (vs. logical) ordering" ~active:(option `MaxOnly) ~packing:vbox#add () in
      let suppmin = List.fold_left ~f:(fun res -> function `SuppMin x -> Some (string_of_int x) | _ -> res) ~init:None sv # links_options in
      let check_suppmin = GButton.check_button ~label:"minimum support of increments" ~active:(suppmin<>None)  ~packing:vbox#add () in
      let entry_suppmin = GEdit.entry ?text:suppmin ~editable:(suppmin<>None) ~width_chars:10 ~packing:vbox#add () in
      let page_size = let ps = sv # links_page_size in if ps = max_int then None else Some (string_of_int ps) in
      let check_page_size = GButton.check_button ~label:"maximum number of increments displayed" ~active:(page_size<>None) ~packing:vbox#add () in
      let entry_page_size = GEdit.entry ?text:page_size ~editable:(page_size<>None) ~width_chars:10 ~packing:vbox#add () in
(* deprecated
      let pattern = List.fold_left ~f:(fun res -> function `Pattern x -> Some x | _ -> res) ~init:None sv # links_options in
      let check_pattern = GButton.check_button ~label:"regular expression as an increment filter" ~active:(pattern<>None) ~packing:vbox#add () in
      let entry_pattern = GEdit.entry ?text:pattern ~editable:(pattern<>None) ~width_chars:40 ~packing:vbox#add () in
*)
      check_suppmin#connect#toggled ~callback:(fun () -> entry_suppmin#set_editable check_suppmin#active);
      check_page_size#connect#toggled ~callback:(fun () -> entry_page_size#set_editable check_page_size#active);
(*      check_pattern#connect#toggled ~callback:(fun () -> entry_pattern#set_editable check_pattern#active); *)

      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:(fun () ->
	try
	  let options =
(*
	    (if check_mostspec#active then [`MostSpec] else []) @
	    (if check_valonly#active then [`ValOnly] else []) @
	    (if check_noviews#active then [`NoViews] else []) @
*)
	    (if check_maxonly#active then [`MaxOnly] else []) @
	    (if check_suppmin#active then [`SuppMin (int_of_string entry_suppmin#text)] else []) in
(*	    (if check_pattern#active then [`Pattern entry_pattern#text] else []) in  *)
	  let sort_fun =
	    if check_lexicalorder#active
	    then sort_by_feature
	    else sort_by_count in
	  let page_size =
	    if check_page_size#active
	    then int_of_string entry_page_size#text
	    else max_int in
	  let chsv sv =
	    sv # set_links_options options;
	    sv # set_links_sort_fun sort_fun;
	    sv # set_links_page_size page_size in
	  ( match paths with
	  | [] ->
	      chsv sv;
	      navig # refresh true
	  | _ ->
	      List.iter 
		(fun path ->
		  sv # set_links_opt None;
		  sv # set_locals_opt None;
		  navig # expand_row ~options:options ~sort_fun:sort_fun ~page_size:page_size path)
		paths
	   );
	  if history # current # extent # locals_only then extent # refresh ();
	  w#destroy ()
	with _ -> ());
      w#show ()

    let menu_collapse () =
      let paths = navig # selected_features in
      List.iter navig#collapse_row paths;
      if history # current # extent # locals_only then extent # refresh ()

    let menu_extent_options t () =
      let se = history # current # extent in
      let stab = se # tab t in
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Extent window options"
	  () in
      let vbox = w#vbox in
      let check_locals_only = GButton.check_button ~label:"local objects only" ~active:se#locals_only ~packing:vbox#add () in
      let label_page_start = GMisc.label ~text:"Page start" ~xalign:0. ~packing:vbox#add () in
      let entry_page_start = GEdit.entry ~text:(string_of_int stab#page_start) ~width_chars:10 ~packing:vbox#add () in
      let label_page_size = GMisc.label ~text:"Page size" ~xalign:0. ~packing:vbox#add () in
      let entry_page_size = GEdit.entry ~text:(string_of_int stab#page_size) ~width_chars:10 ~packing:vbox#add () in

      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:(fun () ->
	try
	  se # set_locals_only check_locals_only#active;
	  let page_start = int_of_string entry_page_start#text in
	  if page_start >= 1 && page_start <= stab#nbobj then stab # set_page_start page_start;
	  let page_size = int_of_string entry_page_size#text in
	  if page_size >= 1 then stab # set_page_size page_size;
	  extent#refresh ();
	  w#destroy ()
	with _ -> ());
      w#show ()

    let menu_show () =
      let paths = navig # selected_features in
      let feature_opt = match paths with [] -> None | path::_ -> Some (navig#feature_from_path path) in
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Inserting a feature"
	  () in
      let vbox = w#vbox in
      let entry = GEdit.entry ?text:feature_opt ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      let action () =
	try
	  Context.add_feature (log_of_string entry#text);
	  w#destroy ();
	  refresh true
	with Stream.Error msg -> syntax_error_logic msg in
      entry#connect#activate ~callback:action;
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:action;
      w#show ()

    let menu_hide () =
      let paths = navig # selected_features in
      List.iter
	(fun path ->
	  navig#collapse_row path;
	  Context.del_feature (log_of_string (navig#feature_from_path path)))
	paths;
      refresh true

    let menu_join () =
      let paths = navig # selected_features in
      let features = List.map navig#feature_from_path paths in
      match features with
      | [] -> ()
      | f::fs ->
	  let w = GWindow.dialog
	      ~destroy_with_parent:true
	      ~title:"Inserting a join feature"
	      () in
	  let vbox = w#vbox in
	  let entry = GEdit.entry ~text:f ~packing:vbox#add () in
	  let action_area = w#action_area in
	  let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
	  let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
	  button_cancel#connect#clicked ~callback:w#destroy;
	  button_ok#connect#clicked ~callback:(fun () ->
	    try
	      let f0 = log_of_string entry#text in
	      List.iter (fun f -> Context.add_axiom (log_of_string f) f0) features;
	      awaken (fun () -> Context.add_feature f0) refresh_force;
	      w#destroy ()
	    with Stream.Error msg -> syntax_error_logic msg);
	  w#show ()

    let menu_replace () =
      let paths = navig # selected_features in
      match paths with
      | [path] -> 
	  let f = navig # feature_from_path path in
	  let w = GWindow.dialog
	      ~destroy_with_parent:true
	      ~title:"Replacing a feature"
	      () in
	  let vbox = w#vbox in
	  let entry = GEdit.entry ~text:f ~packing:vbox#add () in
	  let action_area = w#action_area in
	  let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
	  let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
	  let action () =
	    try
	      let f' = entry#text in
	      Context.replace_feature (log_of_string f) (log_of_string f');
	      history # current # replace_view f f';
	      refresh true;
	      w#destroy ()
	    with
	    | Stream.Error msg -> syntax_error_logic msg
	    | Failure msg -> error_message msg in
	  entry#connect#activate ~callback:action;
	  button_cancel#connect#clicked ~callback:w#destroy;
	  button_ok#connect#clicked ~callback:action;
	  w#show ()
      | _ -> ()


    let menu_rule_list () =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"List rules"
	  () in
      let vbox = w#vbox in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:vbox#add () in
      let cols = new GTree.column_list in
      let rule = cols#add (caml : Context.rule Gobject.data_conv) in
      let kind = cols#add string in
      let update = cols#add string in
      let query = cols#add string in
      let model = GTree.list_store cols in
      let view = GTree.view ~model ~height:300 ~rules_hint:true ~packing:sw#add () in
      view#selection#set_mode `MULTIPLE;
      let col_kind = GTree.view_column ~title:"kind" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",kind]) in
      let col_update = GTree.view_column ~title:"update" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",update]) in
      let col_query = GTree.view_column ~title:"query" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",query]) in
      view#append_column col_kind;
      view#append_column col_update;
      view#append_column col_query;
      let refresh () =
	model#clear ();
	List.iter
	  (fun r ->
	    let item = model#append () in
	    model#set ~row:item ~column:rule r;
	    model#set ~row:item ~column:query (utf0 (string_of_query (Context.query_rule r)));
	    match Context.update_rule r with
	    | Context.Intr u ->
		model#set ~row:item ~column:kind (utf "paste intr");
		model#set ~row:item ~column:update (utf0 (string_of_update u))
	    | Context.Extr u ->
		model#set ~row:item ~column:kind (utf "paste");
		model#set ~row:item ~column:update (utf0 (string_of_update u))
	    | Context.Rem ->
		model#set ~row:item ~column:kind (utf "delete"))
	  (Context.get_rules ()) in
      let selected_rules () : Context.rule list =
	List.fold_left
	  ~f:(fun res path ->
	    let item = model#get_iter path in
	    let r = model#get ~row:item ~column:rule in
	    r::res)
	  ~init:[]
	  view#selection#get_selected_rows in

      let action_area = w#action_area in
      let button_refresh = GButton.button ~stock:`REFRESH ~packing:action_area#add () in
      let button_delete = GButton.button ~stock:`DELETE ~packing:action_area#add () in
      let button_close = GButton.button ~stock:`CLOSE ~packing:action_area#add () in
      button_refresh#connect#clicked ~callback:refresh;
      button_delete#connect#clicked ~callback:(fun () ->
	List.iter Context.final_rule (selected_rules ());
	refresh ());
      button_close#connect#clicked ~callback:w#destroy;
      refresh ();
      w#show ()

    let ctx_menu_clicked nb_selected ctx_menu =
      let ctx_menu_expand = GMenu.menu_item ~label:"Expand..." ~packing:ctx_menu#append () in
      let ctx_menu_collapse = GMenu.menu_item ~label:"Collapse" ~packing:ctx_menu#append () in
      let ctx_sep2 = GMenu.separator_item ~packing:ctx_menu#append () in
      let ctx_menu_expand_as_tab = GMenu.menu_item ~label:"Expand as tab" ~packing:ctx_menu#append () in
      let ctx_menu_collapse_current_tab = GMenu.menu_item ~label:"Collapse current tab" ~packing:ctx_menu#append () in
      let ctx_sep3 = GMenu.separator_item ~packing:ctx_menu#append () in
      let ctx_menu_show = GMenu.menu_item ~label:"Add..." ~packing:ctx_menu#append () in
      let ctx_menu_join = GMenu.menu_item ~label:"Join as..." ~packing:ctx_menu#append () in
      let ctx_menu_replace = GMenu.menu_item ~label:"Replace..." ~show:(nb_selected = 1) ~packing:ctx_menu#append () in
      let ctx_menu_hide = GMenu.menu_item ~label:"Delete" ~packing:ctx_menu#append () in
      let ctx_sep4 = GMenu.separator_item ~packing:ctx_menu#append () in
      let ctx_menu_copy_features = GMenu.menu_item ~label:"Copy" ~packing:ctx_menu#append () in
      let ctx_menu_copy_paste_features = GMenu.menu_item ~label:"Copy and paste..." ~packing:ctx_menu#append () in
      let ctx_menu_paste_features = GMenu.menu_item ~label:"Paste" ~packing:ctx_menu#append () in
      let ctx_menu_paste_not_features = GMenu.menu_item ~label:"Paste not" ~packing:ctx_menu#append () in
      let ctx_menu_paste_to = GMenu.menu_item ~label:"Paste to..." ~show:(nb_selected > 0) ~packing:ctx_menu#append () in
      
      ctx_menu_expand#connect#activate ~callback:menu_expand;
      ctx_menu_collapse#connect#activate ~callback:menu_collapse;
      ctx_menu_expand_as_tab#connect#activate ~callback:menu_expand_as_tab;
      ctx_menu_collapse_current_tab#connect#activate ~callback:menu_collapse_current_tab;
      ctx_menu_show#connect#activate ~callback:menu_show;
      ctx_menu_join#connect#activate ~callback:menu_join;
      ctx_menu_replace#connect#activate ~callback:(fun () -> if nb_selected = 1 then menu_replace ());
      ctx_menu_hide#connect#activate ~callback:menu_hide;
      ctx_menu_copy_features#connect#activate ~callback:(fun () -> navig#copy_features);
      ctx_menu_copy_paste_features#connect#activate ~callback:(fun () -> navig#copy_features; menu_paste_to ());
      ctx_menu_paste_features#connect#activate ~callback:menu_paste;
      ctx_menu_paste_not_features#connect#activate ~callback:(menu_paste ~neg:true);
      ctx_menu_paste_to#connect#activate ~callback:menu_paste_to;
      ()

    let menu_list_action_args oid () =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"List action arguments"
	  () in
      let vbox = w#vbox in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:vbox#add () in
      let cols = new GTree.column_list in
      let mime = cols#add string in
      let role = cols#add string in
      let location = cols#add string in
      let model = GTree.list_store cols in
      let view = GTree.view ~model ~width:300 ~height:200 ~rules_hint:true ~packing:sw#add () in
      view#selection#set_mode `NONE;
      let col_mime = GTree.view_column ~title:"Mime type" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",mime]) in
      let col_role = GTree.view_column ~title:"Role" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",role]) in
      let col_location = GTree.view_column ~title:"Value" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",location]) in
      view#append_column col_mime;
      view#append_column col_role;
      view#append_column col_location;
      let refresh () =
	model#clear ();
	List.iter
	  (fun arg ->
	    let item = model#append () in
	    model#set ~row:item ~column:mime (utf arg.Source.mime);
	    model#set ~row:item ~column:role (utf arg.Source.role);
	    model#set ~row:item ~column:location (utf arg.Source.location))
	  (Context.action_args oid) in

      let action_area = w#action_area in
      let button_close = GButton.button ~stock:`CLOSE ~packing:action_area#add () in
      button_close#connect#clicked ~callback:w#destroy;
      refresh ();
      w#show ()


    let menu_make () =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Add object"
	  () in
      let vbox = w#vbox in
      let label_title = GMisc.label ~text:"Title" ~xalign:0. ~packing:vbox#add () in
      let entry_title = GEdit.entry ~packing:vbox#add () in
      let label_file = GMisc.label ~text:"Picture (optional, JPG file)" ~xalign:0. ~packing:vbox#add () in
      let hbox_file = GPack.hbox ~packing:vbox#pack () in
      let entry_file = GEdit.entry ~width_chars:60 ~packing:hbox_file#pack () in
      let button_file = GButton.button ~stock:`FIND ~packing:hbox_file#add () in
      let label_cmd = GMisc.label ~text:"Shell command (optional)" ~xalign:0. ~packing:vbox#add () in
      let entry_cmd = GEdit.entry ~packing:vbox#add () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      let action () =
	let title = loc0 entry_title#text in
	if title <> "" then begin
	  let file = loc0 entry_file#text in
	  let pre = if file = "" then Source.String title else Source.Picture (file,title) in
	  let cmd = loc0 entry_cmd#text in
	  let args = if cmd = "" then [] else [{Source.mime="cmd"; Source.role="Execute"; Source.location=cmd}] in
	  let d = Context.Object.Descr.empty in
	  navig#selected_update
	    ~callback:(fun (e,_) ->
	      awaken (fun () -> Context.make_obj pre args d e) refresh_force;
	      w#destroy ()) ()
	end in
      entry_title#connect#activate ~callback:action;
      button_file#connect#clicked ~callback:
	(file_dialog ~title:"Select JPG file" ~filename:!last_dir ~filter:filter_jpg
	   ~callback:(fun name ->
	     last_dir := Filename.dirname name ^ dir_sep;
	     entry_file#set_text (utf name)));
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:action;
      w#show ()

    let menu_delete_objects () =
      let oids = extent#selected_oids () in
      if question "Are you sure you want to delete the selected objects ?"
      then begin
	cb_copy := Nothing;
	awaken
	  (fun () ->
	    Ext.iter (fun oid -> try Context.update_obj oid Context.Rem with _ -> ()) oids;
	    Context.fold_sources_under (fun url () -> Context.final_source url) () oids)
	  refresh_force
      end

    let menu_new_state () =
      awaken
	(fun () -> Context.move_monitor (Context.wq ()))
	refresh_force

    let menu_delete_query () =
      if question "Are you sure you want to create a rule deleting every instance of the current query ?"
      then begin
	cb_copy := Nothing;
	awaken
	  (fun () -> Context.init_rule (Context.wq ()) Context.Rem)
	  refresh_force
      end

    let menu_intent oids =
      cd_intent oids;
      extent#scroll_changed ();
      history # after_cd;
      refresh true

    let query_menu_clicked query query_menu =
      let query_menu_factory = new GMenu.factory query_menu in
      let cmd_new_state = query_menu_factory#add_item "New state" in
      let _ = query_menu_factory#add_separator () in
      let cmd_copy = query_menu_factory#add_item "Copy" in
      let cmd_copy_paste = query_menu_factory#add_item "Copy and paste to..." in
      let cmd_delete = query_menu_factory#add_item "Delete" in
      cmd_new_state#connect#activate ~callback:menu_new_state;
      cmd_copy#connect#activate ~callback:(fun () -> wq # copy_query);
      cmd_copy_paste#connect#activate ~callback:(fun () -> wq # copy_query; menu_paste_to ());
      cmd_delete#connect#activate ~callback:menu_delete_query;
      ()

    let extent_menu_clicked t oids objects_menu =
      let objects_menu_factory = new GMenu.factory objects_menu in
      let cmds = Hashtbl.create 10 in
      if Ext.is_empty oids
      then ()
      else
	List.iter
	  (fun ac ->
	    Ext.iter
	      (fun oid ->
		let action_args = Context.action_args oid in
		  List.iter
		  (fun (label,code) ->
		    let codes0, nobj0 =
		      try Hashtbl.find cmds label
		      with Not_found -> [], 0 in
		    Hashtbl.replace cmds label (code::codes0,1+nobj0))
		  (Context.apply_action ac action_args))
	      oids)
	  (Context.get_actions ());
      let cmds_object : (GMenu.menu_item * (unit -> unit)) list =
	Hashtbl.fold
	  (fun label (codes,nbobj) res ->
	    let label_nbobj =
	      if nbobj = 1
	      then label
	      else label ^ " (x " ^ string_of_int nbobj ^ ")" in
	    let code () =
	      if nbobj = 1 || question "This action will be performed many times. Do you still want to procede ?"
	      then List.iter (fun f -> f ()) codes in
	    (objects_menu_factory#add_item label_nbobj, code) :: res)
	  cmds
	    [] in 
      if cmds_object <> [] then ignore (objects_menu_factory#add_separator ());
      let cmds_export_temp : (GMenu.menu_item * (unit -> unit)) list =
	let cs = contents_of_extent oids in
	let actions =
	  List.fold_left
	    ~f:(fun res (ext,label) ->
	      let tempfile = Filename.temp_file "camelis" ("."^ext) in
		res @ Context.temp_well_actions tempfile cs)
	    ~init:[]
	    (Context.Wll.kinds cs) in
	List.map
	  ~f:(fun (label,code) ->
	    (objects_menu_factory#add_item label, code))
	  actions in
      if cmds_export_temp <> [] then ignore (objects_menu_factory#add_separator ());
      
      if Ext.cardinal oids = 1 then begin
	match Ext.elements oids with
	| [] -> ()
	| oid::_ ->
	    let cmd_list_args = objects_menu_factory#add_item "List action arguments..." in
	    ignore (objects_menu_factory#add_separator ());
	    ignore (cmd_list_args#connect#activate ~callback:(menu_list_action_args oid))
      end;
      
      let cmd_make = objects_menu_factory#add_item "Add" in
      let cmd_copy = objects_menu_factory#add_item "Copy" in
      let cmd_paste_extr = objects_menu_factory#add_item "Paste" in
      let cmd_copy_paste_to = objects_menu_factory#add_item "Copy and paste to..." in
(*	let cmd_paste_intr = objects_menu_factory#add_item "Paste intrinsic" in*)
      let cmd_delete = objects_menu_factory#add_item "Delete" in
      let _ = objects_menu_factory#add_separator () in
      let cmd_intent = objects_menu_factory#add_item "Intent" in
      let _ = objects_menu_factory#add_separator () in
      let cmd_extent_options = objects_menu_factory#add_item "Options..." in
      List.iter
	(fun (item,code) ->
	  ignore (item#connect#activate ~callback:code))
	cmds_object;
      List.iter
	(fun (item,code) ->
	  ignore (item#connect#activate ~callback:code))
	cmds_export_temp;
      cmd_make#connect#activate ~callback:menu_make;
      cmd_copy#connect#activate ~callback:(fun () -> extent#copy_objects ());
      cmd_paste_extr#connect#activate ~callback:menu_paste;
      cmd_copy_paste_to#connect#activate ~callback:(fun () -> extent#copy_objects (); menu_paste_to ());
(*	cmd_paste_intr#connect#activate ~callback:menu_paste_intr;*)
      cmd_delete#connect#activate ~callback:(fun () -> menu_delete_objects ());
      cmd_intent#connect#activate ~callback:(fun () -> menu_intent (extent#selected_oids ()));
      cmd_extent_options#connect#activate ~callback:(menu_extent_options t);
      ()


(* actions *)

    let dialog_edit_action title ac f =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title
	  () in
      let vbox = w#vbox in
      let label_intro = GMisc.label
	  ~text:"Note: the fields 'name' and 'command' can contain MIME type variables,\n\
	         like $(image/jpeg) or $(url/text/html).\n\
                 They must conform to action arguments associated to objects by sources."
          ~xalign:0. ~packing:vbox#pack () in
      let label_name = GMisc.label ~text:"\nName of the action" ~xalign:0. ~packing:vbox#pack () in
      let entry_name = GEdit.entry ~width_chars:60 ~text:(utf ac.Context.action_name) ~packing:vbox#pack () in
      let label_cmd = GMisc.label ~text:"\nThe command to be executed"
	  ~xalign:0. ~packing:vbox#pack () in
      let hbox_cmd = GPack.hbox ~packing:(vbox#pack ~expand:true) () in
      let entry_cmd = GEdit.entry ~width_chars:60 ~text:(utf ac.Context.action_cmd) ~packing:hbox_cmd#pack () in
      let button_cmd = GButton.button ~stock:`FIND ~label:"Choose an application..." ~packing:hbox_cmd#pack () in
      let label_query = GMisc.label ~text:"\nThe query defining the scope of the action" ~xalign:0. ~packing:vbox#pack () in
      let entry_query = GEdit.entry ~width_chars:60 ~text:(utf0 (Syntax.stringizer (Query.print ac.Context.action_query)))
	  ~packing:vbox#pack () in
      let action_area = w#action_area in
      let button_cancel = GButton.button ~stock:`CANCEL ~packing:action_area#add () in
      let button_ok = GButton.button ~stock:`OK ~packing:action_area#add () in
      button_cmd#connect#clicked ~callback:
	(file_dialog ~title:"Choosing an application" ~filename:!last_dir
	   ~callback:(fun name ->
	     last_dir := Filename.dirname name ^ dir_sep;
	     entry_cmd#set_text (utf (name ^ " $(file) &"))));
      button_cancel#connect#clicked ~callback:w#destroy;
      button_ok#connect#clicked ~callback:(fun () ->
	try
	  ac.Context.action_query <- Query.parse (Syntax.from_string (loc0 entry_query#text));
	  ac.Context.action_name <- loc entry_name#text;
	  ac.Context.action_cmd <- loc entry_cmd#text;
	  f ();
	  w#destroy ()
	with _ ->
	  entry_query#select_region ~start:0 ~stop:(entry_query#text_length - 1));
      w#show ()


    let menu_add_action ?init () =
      let ac =
	match init with
	| Some ac -> ac
	| None -> {
	    Context.action_name = "";
	    Context.action_cmd = "";
	    Context.action_query = Context.wq ();
	  } in
      dialog_edit_action "Adding a new action" ac (fun () -> Context.add_action ac)

    let menu_list_actions () =
      let w = GWindow.dialog
	  ~destroy_with_parent:true
	  ~title:"Listing actions"
	  () in
      let vbox = w#vbox in
      let sw = GBin.scrolled_window ~shadow_type:`ETCHED_IN ~hpolicy:`AUTOMATIC
	  ~vpolicy:`AUTOMATIC ~packing:vbox#add () in
      let cols = new GTree.column_list in
      let action = cols#add (caml : Context.action Gobject.data_conv) in
      let name = cols#add string in
      let cmd = cols#add string in
      let query = cols#add string in
      let model = GTree.list_store cols in
      let view = GTree.view ~model ~rules_hint:true ~height:300 ~packing:sw#add () in
      view#selection#set_mode `SINGLE;
      let col_name = GTree.view_column ~title:"name" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",name]) in
      let col_cmd = GTree.view_column ~title:"command" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",cmd]) in
      let col_query = GTree.view_column ~title:"query" ()
	  ~renderer:(GTree.cell_renderer_text[], ["text",query]) in
      view#append_column col_name;
      view#append_column col_cmd;
      view#append_column col_query;
      let refresh () =
	model#clear ();
	List.iter
	  (fun ac ->
	    let item = model#append () in
	    model#set ~row:item ~column:action ac;
	    model#set ~row:item ~column:name (utf ac.Context.action_name);
	    model#set ~row:item ~column:cmd (utf ac.Context.action_cmd);
	    model#set ~row:item ~column:query (utf0 (string_of_query ac.Context.action_query)))
	  (Context.get_actions ()) in
      let selected_actions () : Context.action list =
	List.fold_left
	  ~f:(fun res path ->
	    let item = model#get_iter path in
	    let ac = model#get ~row:item ~column:action in
	    ac::res)
	  ~init:[]
	  view#selection#get_selected_rows in
      let action_area = w#action_area in
      let button_refresh = GButton.button ~stock:`REFRESH ~packing:action_area#add () in
      let button_add = GButton.button ~stock:`ADD ~packing:action_area#add () in
      let button_edit = GButton.button ~stock:(`STOCK "Edit") ~packing:action_area#add () in
      let button_delete = GButton.button ~stock:`DELETE ~packing:action_area#add () in
      let button_close = GButton.button ~stock:`CLOSE ~packing:action_area#add () in
      button_refresh#connect#clicked ~callback:refresh;
      button_add#connect#clicked ~callback:(fun () ->
	let init =
	  try
	    let ac = List.hd (selected_actions ()) in
	    Some ({ac with Context.action_name = ac.Context.action_name})
	      (* a copy a ac *)
          with _ -> None in
	menu_add_action ?init ();
	refresh ());
      button_edit#connect#clicked ~callback:(fun () ->
	try
	  dialog_edit_action "Editing the action" (List.hd (selected_actions ()) : Context.action) (fun () -> ());
	  refresh ()
	with _ -> ());
      button_delete#connect#clicked ~callback:(fun () ->
	try
	  Context.remove_action (List.hd (selected_actions ()));
	  refresh ()
	with _ -> ());
      button_close#connect#clicked ~callback:w#destroy;
      refresh ();
      w#show ()


(* Building the GUI *)

    let main () =
      Arg.parse
	[ ("-readonly", Arg.Unit (fun () -> readonly := true), "makes open contexts read only");
	  ("-log", Arg.Unit (fun () -> log := true), "creates a log when creating a new context")]
	(fun f ->
	  if Filename.check_suffix f "lis" then
	    if !lis_file <> ""
	    then raise (Arg.Bad "Only one .lis file can be loaded.")
	    else lis_file := f
	  else if Filename.check_suffix f "ctx" then
	    ctx_files := f::!ctx_files
	  else
	    raise (Arg.Bad "Only one .lis and .ctx files can be loaded."))
	"glis.exe [-readonly] [-log] [file.lis] [file1.ctx ... filen.ctx]";
      Gc.set { (Gc.get())
	     with
	       Gc.major_heap_increment = 512*1024;
	       Gc.space_overhead = 50;
	       Gc.max_overhead = 50;
               Gc.stack_limit = 8 * 1024 * 1024 };

      (* callbacks *)
      window#event#connect#delete ~callback:(fun _ -> menu_quit (); true) (*fun () -> Context.init ""; GMain.quit ()*);

      cmd_new#connect#activate ~callback:menu_new;
      cmd_open#connect#activate ~callback:menu_open;
      cmd_save#connect#activate ~callback:menu_save;
      cmd_save#add_accelerator ~group:accel_group ~modi:[`CONTROL] _s;
      cmd_saveas#connect#activate ~callback:menu_saveas;
      cmd_close#connect#activate ~callback:menu_close;
      cmd_import_from_ctx#connect#activate ~callback:menu_import_ctx;
      cmd_import_from_file#connect#activate ~callback:menu_import_file;
      cmd_import_list#connect#activate ~callback:menu_import_list;
      cmd_import_update#connect#activate ~callback:menu_import_update;
      cmd_import_update#add_accelerator ~group:accel_group ~modi:[`CONTROL] _u;
      cmd_export_to_ctx#connect#activate ~callback:menu_export_ctx;
      cmd_export_to_bin#connect#activate ~callback:menu_export_as_binary;
      cmd_export_file#connect#activate ~callback:menu_export_file;
      cmd_export_list#connect#activate ~callback:menu_export_list;
      cmd_stat#connect#activate ~callback:menu_stat;
      cmd_quit#connect#activate ~callback:menu_quit;

      cmd_home#connect#activate ~callback:menu_home;
      cmd_current_state#connect#activate ~callback:menu_current_state;
      cmd_back#connect#activate ~callback:menu_back;
      cmd_back#add_accelerator ~group:accel_group ~modi:[`CONTROL] _b;
      cmd_fwd#connect#activate ~callback:menu_fwd;
      cmd_fwd#add_accelerator ~group:accel_group ~modi:[`CONTROL] _f;
      cmd_refresh#connect#activate ~callback:menu_refresh;
      cmd_refresh#add_accelerator ~group:accel_group ~modi:[`CONTROL] _r;
(*
      cmd_select#connect#activate ~callback:(menu_cd (link Rel f_and f_id));
      cmd_exclude#connect#activate ~callback:(menu_cd (link Rel f_not f_id));
      cmd_absolute#connect#activate ~callback:(menu_cd (link Abs f_and f_id));
      cmd_absolute_abs#connect#activate ~callback:(menu_cd (link Abs f_not f_id));
*)
      cmd_complement#connect#activate ~callback:(menu_cd (link Rel f_compl f_id));
      cmd_if#connect#activate ~callback:(menu_cd (link Rel f_if f_id));
      cmd_unif#connect#activate ~callback:(menu_cd (link Rel f_unif f_id));
      cmd_intent#connect#activate ~callback:(fun () -> menu_intent (extent#selected_oids ()));
      check_scale#connect#activate ~callback:(fun () -> refresh false); (* only graphical refresh *)
      check_feat_color#connect#activate ~callback:(fun () -> refresh false); (* only graphical refresh *)
      check_lift#connect#activate ~callback:(fun () -> refresh false); (* only graphical refresh *)
      check_lift_color#connect#activate ~callback:(fun () -> refresh false); (* only graphical refresh *)
      check_ext_color#connect#activate ~callback:(fun () -> refresh false); (* only graphical refresh *)
      check_ext_preselect#connect#clicked ~callback:(fun () -> extent # refresh_preselect false); (* only preselection refresh *)
      check_int_preselect#connect#clicked ~callback:(fun () -> navig # refresh_preselect false); (* only preselection refresh *)
      cmd_expand#connect#activate ~callback:menu_expand;
      cmd_collapse#connect#activate ~callback:menu_collapse;
      cmd_extent_options#connect#activate ~callback:(fun () -> menu_extent_options extent#current_tab ());

      cmd_show#connect#activate ~callback:menu_show;
      cmd_show#add_accelerator ~group:accel_group ~modi:[`CONTROL] _a;
      cmd_join#connect#activate ~callback:menu_join;
      cmd_copy_features#connect#activate ~callback:(fun () -> navig # copy_features);
      cmd_paste_features#connect#activate ~callback:menu_paste;
      cmd_hide#connect#activate ~callback:menu_hide;
      cmd_hide#add_accelerator ~group:accel_group ~modi:[`CONTROL] _h;
      cmd_copy_query#connect#activate ~callback:(fun () -> wq # copy_query);
      cmd_copy_paste_query#connect#activate ~callback:(fun () -> wq # copy_query; menu_paste_to ());
      cmd_delete_query#connect#activate ~callback:menu_delete_query;
      cmd_rule_list#connect#activate ~callback:menu_rule_list;

      cmd_new_state#connect#activate ~callback:menu_new_state;
      cmd_new_state#add_accelerator ~group:accel_group ~modi:[`CONTROL] _e;
      cmd_make#connect#activate ~callback:menu_make;
      cmd_make#add_accelerator ~group:accel_group ~modi:[`CONTROL] _o;
      cmd_copy#connect#activate ~callback:(fun () -> extent#copy_objects ());
      cmd_paste_extr#connect#activate ~callback:menu_paste;
      cmd_copy_paste_to#connect#activate ~callback:(fun () -> extent#copy_objects (); menu_paste_to ());
      cmd_paste_extr#add_accelerator ~group:accel_group ~modi:[`CONTROL] _p;
(*      cmd_paste_intr#connect#activate ~callback:menu_paste_intr;*)
      cmd_delete#connect#activate ~callback:(fun () -> menu_delete_objects ());

      cmd_add_action#connect#activate ~callback:menu_add_action;
      cmd_list_action#connect#activate ~callback:menu_list_actions;

      cmd_about#connect#activate ~callback:(fun () -> info_message
	  ~title:"About Camelis"
(*	  ~icon:(pixbuf_of_picture ~maxside:30 "/local/ferre/prog/ocaml/camelis/bin/lis.gif") *)
	  "Camelis 1.4.3\n\n\
           An implementation of Logical Information Systems (LIS)\n\
           developped by Sebastien Ferre (ferre@irisa.fr),\n\
           and distributed under GPL license (http://www.irisa.fr/LIS/ferre/camelis/).");

      button_home#connect#clicked ~callback:menu_home;
      tooltips#set_tip button_home#coerce ~text:"Go to the root query covering all object";
      button_back#connect#clicked ~callback:menu_back;
      tooltips#set_tip button_back#coerce ~text:"Go backward in the history of queries";
      button_fwd#connect#clicked ~callback:menu_fwd;
      tooltips#set_tip button_fwd#coerce ~text:"Go forward in the history of queries";
      button_refresh#connect#clicked ~callback:menu_refresh;
      tooltips#set_tip button_refresh#coerce ~text:"Refresh the navigation and extent windows";
      button_save#connect#clicked ~callback:menu_save;
      tooltips#set_tip button_save#coerce ~text:"Save the context";
(*      button_make#connect#clicked ~callback:menu_make;*)
      button_paste#connect#clicked ~callback:menu_paste;
      tooltips#set_tip button_paste#coerce ~text:"Add the copied entities into the selected features (default is the features of the current query)";
      button_update#connect#clicked ~callback:menu_import_update;
      tooltips#set_tip button_update#coerce ~text:"Update the context w.r.t. imported files";
      tooltips#set_tip entry_nbupdates#coerce ~text:"This displays the number of updates to be performed";

      wq#connect_activated ~callback:entry_wq_activated;
      wq#connect_ctx_menu ~callback:query_menu_clicked;

      nl#connect
	(object
	  method cd ~mode ~negation ~comparison =
	    let fg = if negation then f_not else f_id in
	    let fl = match comparison with ">=" -> f_ge | "<=" -> f_le | _ -> f_id in
	    let incr =
	      match mode with
	      | Zoom -> link Rel fg fl
	      | Pivot -> link Abs fg fl in
	    menu_cd incr ()
	end);

      navig#connect
	(object
	  method row_activated path =
	    row_activated (link Rel f_and f_id) [path]

	  method row_expanded path =
	    if history # current # extent # locals_only
	    then extent # refresh ()

	  method row_collapsed path =
	    if history # current # extent # locals_only
	    then extent # refresh ()

	  method refresh_preselect = extent # refresh_preselect

	  method ctx_menu = ctx_menu_clicked

	  method data_received () = menu_paste ()
	end);

      extent#connect
	(object
	  method row_activated oid = menu_intent (Ext.singleton oid)

	  method ctx_menu = extent_menu_clicked

	  method refresh_preselect = navig # refresh_preselect
	end);

      window#set_default_size ~width:800 ~height:600;
(*
      let font = window#misc#pango_context#font_description in
      print_int (Pango.Font.get_size font); print_newline ();
      Pango.Font.set_size font (4 * Pango.Font.get_size font / 3);
      Pango.Font.set_weight font `BOLD;
      window#misc#modify_font font;
*)
      window#add_accel_group accel_group;
      window#show ();
      paned_global#set_position ((7 * paned_global#min_position + 1 * paned_global#max_position) / 8); (* position aux 1/8 *)
      paned_sw#set_position ((3 * paned_sw#min_position + 2 * paned_sw#max_position) / 5); (* position aux 2/5 *)
      GMain.Idle.add (fun () ->
	if !lis_file <> ""
	then begin
	  if Sys.file_exists !lis_file
	  then load !lis_file  (* opening *)
	  else init !lis_file; (* creating *)
	  window#set_title (utf (title ()));
	  history # init end
	else begin
	  if !ctx_files = [] then (* default setting to an initializing ctx file *)
	    match Sys.os_type with
	    | "Unix" -> ctx_files := [ Filename.concat (Sys.getenv "HOME") ".camelis.ctx" ]
	    | "Cygwin" -> ctx_files := [ Filename.concat (Filename.dirname Sys.argv.(0)) "init.ctx" ]
	    | _ -> ()
	end;
	List.iter
	  (fun file -> import [] (Filename.chop_extension file))
(*
	    try Context.import [] (Filename.chop_extension file)
	    with e ->
	      print_string "Importing: "; print_endline file;
	      print_endline (Printexc.to_string e))
*)
	  (List.rev !ctx_files);
	awaken (fun () -> ()) menu_home;
	false);
      GMain.main ()

  end
