
type v =
  | None
  | TermAsAttr of string
  | StringList of string list
  | Integer of int
  | Date of (int * string * int) (* day, month, year *)
  | Time of (int * int * int) (* hours, minutes, seconds *)
  | Composed of string * v
  | Verbatim of string
	
type t = (string * v) list

let init = []

let add : (string * v) -> t -> t =
  fun (f,v) lf ->
    (f,v)::lf

open Token

let utf s =
  if Glib.Utf8.validate s
  then s
  else try Glib.Convert.locale_to_utf8 s with _ -> s

let preview_utf = function
  | Source.String s -> Source.String (utf s)
  | Source.Picture (f,s) -> Source.Picture (f, utf s)

let print_attr a =
  let a_utf = utf a in
  if Str.string_match (Str.regexp "[a-z][a-zA-Z0-9_]*$") a_utf 0
  then Ident a_utf
  else Term a_utf

let rec print_val v =
  match v with
  | None -> []
  | TermAsAttr t -> [Term (utf t)]
  | StringList [] -> []
  | StringList (x::l) ->
      let s = List.fold_left (fun res s -> res ^ ", " ^ s) x l in
      [Ident "is"; PP_tilda; String (utf s)]
  | Integer n ->
      [Equal; PP_tilda] @ Syntax.toks_of_int n
  | Date (d,m,y) ->
      let m' = String.lowercase (String.sub m 0 3) in
      [Ident "date"; PP_tilda; Equal; PP_tilda; Nat d; PP_tilda; Ident m'; PP_tilda] @ Syntax.toks_of_int y
  | Time (h,m,s) ->
      [Ident "time"; PP_tilda; Equal; PP_tilda; Nat h; Colon; Nat m; Colon; Nat s]
  | Composed (b,v) ->
      print_attr b :: PP_tilda :: print_val v
  | Verbatim s ->
      Syntax.list_of_stream (Syntax.from_string s)

let print (a,v) =
  if a = ""
  then print_val v
  else print_attr a :: PP_tilda :: print_val v

let prints = function
  | [] -> [Star]
  | f::fs ->
      List.fold_left
	(fun res f -> res @ (Comma :: PP_tilda :: print f))
	(print f)
	fs
