
type descr = string * (Fields.t * Source.preview * Source.action_arg list)

module Make
(Param : sig
  val filter : Source.url -> bool
  val may_have_changed_since : Source.url -> float -> bool
  val get_objs : string -> bool -> (string * (Fields.t * Source.preview * Source.action_arg list)) list
end) : Source.T =
  struct
    include Source.Default

    type obj = {
	mutable preview : Source.preview;
	mutable action_args : Source.action_arg list;
	mutable digest : string
      }

    type t = {
	mutable url : string;
	mutable options : Source.src_option list;
	mutable last_update : float;
	mutable objs : (ref,obj) Hashtbl.t;
      }

    let make url =
      if Param.filter url
      then {
	url = url;
	options = [];
	last_update = 0.;
	objs = Hashtbl.create 13;
      }
      else raise (Invalid_argument "Setobj.make: invalid url")

    let url s = s.url

    let eq s1 s2 = s1.url = s2.url

    let parts s =
      Hashtbl.fold (fun ref _ res -> LSet.add ref res) s.objs (LSet.empty ())

    let preview s ref =
      (Hashtbl.find s.objs ref).preview

    let action_args s ref =
      (Hashtbl.find s.objs ref).action_args

    let init s options =
      s.options <- options;
      s.last_update <- Unix.time ();
      try
	List.map
	  (fun (ref, (lf,pre,args)) ->	    
	    let d = Fields.prints lf in
	    let pre = Fields.preview_utf pre in
	    let obj = {
	      preview = pre;
	      action_args = args;
	      digest = Digest.string (Syntax.stringizer d)} in
	    Hashtbl.add s.objs ref obj;
	    Source.Make (s.url, ref, d))
	  (Param.get_objs s.url (List.mem `Parts options))
      with _ -> []

    let update s options =
      if s.options = options && not (Param.may_have_changed_since s.url s.last_update)
      then begin
	s.last_update <- Unix.time ();
	[] end
      else begin
	s.last_update <- Unix.time ();
	s.options <- options;
	try
	  let objs = Param.get_objs s.url (List.mem `Parts options) in
	  let ups =
	    List.fold_left
	      (fun res (ref, (lf,pre,args)) ->
		let d = Fields.prints lf in
		let pre = Fields.preview_utf pre in
		let digest = Digest.string (Syntax.stringizer d) in
		try
		  let obj = Hashtbl.find s.objs ref in
		  obj.preview <- Fields.preview_utf pre;
		  obj.action_args <- args;
		  if obj.digest <> digest
		  then begin (* updating modified references *)
		    obj.digest <- digest;
		    Source.UpdateIntrSrc (s.url, ref, d) :: res end
		  else res
		with Not_found -> (* making new references *)
		  let obj = {
		    preview = pre;
		    action_args = args;
		    digest = digest} in
		  Hashtbl.add s.objs ref obj;
		  Source.Make (s.url, ref, d) :: res)
	      [] objs in
	  Hashtbl.fold (* removing old references *)
	    (fun ref obj res ->
	      if List.mem_assoc ref objs
	      then res
	      else begin
		Hashtbl.remove s.objs ref;
		Source.RemoveSrc (s.url, ref) :: res
	      end)
	    s.objs ups
	with _ -> [] (* safety *)
      end

    let final s =
      Hashtbl.fold
	(fun ref obj res ->
	  Source.RemoveSrc (s.url, ref)::res)
	s.objs []

    let move s new_url =
      s.url <- new_url;
      []

  end
