
let is_local url = Sys.file_exists url

let is_on_web url =
  let code = try Sys.command ("wget -q --spider " ^ url) with _ -> (-1) in
  code = 0

let may_have_changed_since url t =
  if is_local url
  then (Unix.stat url).Unix.st_mtime > t
  else true

(*
let path_of_command cmd url =
  let path = Filename.temp_file "cmd" ".txt" in
  let code = Sys.command (cmd ^ " \"" ^ url ^ "\" > \"" ^ path ^ "\"") in
  path
*)

let wget url path =
  let code = Sys.command ("wget -q -O " ^ path ^ " \"" ^ url ^ "\"") in
  if code <> 0 then invalid_arg ("Url.path_of_url: invalid url: " ^ url)

let path_of_url url =
  if is_local url
  then url
  else
    let path = Filename.temp_file "url" ".htm" in
    wget url path;
    path

let mime url m =
  if is_local url
  then m
  else "url/" ^ m

type attr =
  | Url       (* full url *)
  | Protocol  (* protocol name *)
  | Host      (* host name *)
  | Dir       (* directories containing the file *)
  | Name      (* the file basename *)
  | Ext       (* the file extent *)
  | Size  (* the file size in bytes *)
  | MTime (* last modification time *)
  | CTime (* last status change time *)
  | ATime (* last acces time *)
  | User  (* file owner name *)
  | Group (* file group name *)
  | Perm  (* file access permissions *)
  | Noattr

open Token

let field_of_string a l_a s =
  let v =
    List.fold_right
      (fun a v -> Fields.Composed (a,v))
      l_a
      (Fields.StringList [s]) in
  (a,v)

let field_of_int a l_a n =
  let v =
    List.fold_right
      (fun a v -> Fields.Composed (a,v))
      l_a
      (Fields.Integer n) in
  (a,v)

let field_of_tm a l_a tm =
  let day = tm.Unix.tm_mday in
  let year = 1900 + tm.Unix.tm_year in
  let months = [|"jan";"feb";"mar";"apr";"may";"jun";"jul";"aug";"sep";"oct";"nov";"dec"|] in
  let month = months.(tm.Unix.tm_mon) in
  let v =
    List.fold_right
      (fun a v -> Fields.Composed (a,v))
      l_a
      (Fields.Date (day,month,year)) in
  (a,v)

(* returns the description of a path *)
let fields_of_path mainattr attrs path =
  let base = Filename.basename path in
  let b = String.length base in
  let name = try let n = Filename.chop_extension base in if n="" then base else n with _ -> base in
  let n = String.length name in
  let ext_opt = if n<b then Some (String.lowercase (String.sub base (n+1) (b - n - 1))) else None in
  let dirs = try List.tl (List.rev (Str.split (Str.regexp "/+") path)) with _ -> [] in
  let st = Unix.stat path in
  let size = st.Unix.st_size in
  let mtm = Unix.localtime st.Unix.st_mtime in
  let ctm = Unix.localtime st.Unix.st_ctime in
  let atm = Unix.localtime st.Unix.st_atime in
  let user = (Unix.getpwuid st.Unix.st_uid).Unix.pw_name in
  let group = (Unix.getgrgid st.Unix.st_gid).Unix.gr_name in
  let perm = st.Unix.st_perm in
  let t attr f =
    if attrs = [] || List.mem attr attrs
    then f ()
    else [] in
  let lf =
    t Url (fun () -> [field_of_string mainattr ["path"] (Filename.dirname path)]) @
    t Ext (fun () -> Common.fold_option (fun ext -> [field_of_string mainattr ["ext"] ext]) [] ext_opt) @
    t Name (fun () -> [field_of_string mainattr ["name"] name]) @
    t Dir (fun () -> List.map (fun dir -> field_of_string mainattr ["dir"] dir) dirs) @
    t Size (fun () -> [field_of_int mainattr ["size"] size]) @
    t MTime (fun () -> [field_of_tm mainattr ["last_modif"] mtm]) @
    t CTime (fun () -> [field_of_tm mainattr ["last_status_change"] ctm]) @
    t MTime (fun () -> [field_of_tm mainattr ["last_access"] atm]) @
    t User (fun () -> [field_of_string mainattr ["user"] user]) @
    t Group (fun () -> [field_of_string mainattr ["group"] group]) @
    t Perm (fun () -> [field_of_int mainattr ["perm"] perm])
  in
  lf

(* returns the description of a path *)
let fields_of_web mainattr attrs url : Fields.t =
  if Str.string_match (Str.regexp "\\([^:]+\\)://\\([^/]+\\)?\\(/?[^#?]*\\).*") url 0 then
    let protocol = Str.matched_group 1 url in
    let host = try Str.matched_group 2 url with Not_found -> "" in
    let path = Str.matched_group 3 url in
    let hosts = Str.split (Str.regexp "\\.") host in
    let base = Filename.basename path in
    let b = String.length base in
    let name = try Filename.chop_extension base with _ -> base in
    let n = String.length name in
    let ext_opt = if n<b then Some (String.lowercase (String.sub base (n+1) (b - n - 1))) else None in
    let dirs = try List.tl (List.rev (Str.split (Str.regexp "/+") path)) with _ -> [] in
    let t attr f =
      if attrs = [] || List.mem attr attrs
      then f ()
      else [] in
    let lf =
      t Url (fun () -> [field_of_string mainattr [] url]) @
      t Protocol (fun () -> [field_of_string mainattr ["protocol"] protocol]) @
      t Host (fun () -> List.map (fun h -> field_of_string mainattr ["host"] h) hosts) @
      (match ext_opt with None -> [] | Some ext -> t Ext (fun () -> [field_of_string mainattr ["ext"] ext])) @
      t Name (fun () -> [field_of_string mainattr ["name"] name]) @
      t Dir (fun () -> List.map (fun dir -> field_of_string mainattr ["dir"] dir) dirs) in
    lf
  else
    [field_of_string mainattr [] url]

let fields_of_url mainattr attrs url =
  if is_local url
  then fields_of_path mainattr attrs url
  else fields_of_web mainattr attrs url

let args_of_url url =
  if is_local url
  then [{Source.mime="dir"; Source.role="directory"; Source.location=Filename.dirname url}]
  else []

let preview_of_url url =
  if is_local url
  then Filename.basename url
  else url

