(** Well: file lists as a plain text file. *)

module Make
(Param : sig
  val kinds : (string * string) list
  val url_arg : string (* mime of the object argument giving the url *)
  val key_arg : string option (* mime of the object argument giving the sorting key *)
  val mime : string (* mime of the export *)
  val role : string (* role of the export *)
  val head : string
  val tail : string
  val item : string -> string
  val url_compare : string -> string -> int
end) : Well.T = struct

  type t = {
      mutable path : string (* path of the url list *)
    }

  let kinds cs =
    if List.exists
	(fun args ->
	  List.exists (fun arg -> Source.mime_match Param.url_arg arg.Source.mime) args &&
	  match Param.key_arg with
	  | None -> true
	  | Some k -> List.exists (fun arg -> Source.mime_match k arg.Source.mime) args)
	cs
    then Param.kinds
    else []

  let action_args w = [
    {Source.mime = Param.mime; Source.role = Param.role; Source.location = w.path}
  ]

  open Token

  let get_urls cs =
    Common.mapfilter
      (fun args ->
	try
	  let url = (List.find (fun arg -> Source.mime_match Param.url_arg arg.Source.mime) args).Source.location in
	  let key =
	    match Param.key_arg with
	    | None -> ""
	    | Some k -> (List.find (fun arg -> Source.mime_match k arg.Source.mime) args).Source.location in
	  Some (key, url)
	with Not_found ->
	  None)
      cs

  let output_urllist ch urls =
    output_string ch Param.head;
    List.iter
      (fun (key,url) ->
	output_string ch (Param.item url))
      (List.sort (fun (k1,p1) (k2,p2) -> Param.url_compare (k1^p1) (k2^p2)) urls);
    output_string ch Param.tail

  let eq w1 w2 = w1.path = w2.path
      
  let make url =
    if List.exists (fun (suf,_) -> Filename.check_suffix url suf) Param.kinds
    then { path = url }
    else raise (Invalid_argument ("Urllist.make: invalid path: " ^ url))

  let url w = w.path

  let init w cs = (* raise Sys_error *)
    let ch = open_out w.path in
    output_urllist ch (get_urls cs);
    close_out ch

  let final w _cs = (* raise Sys_error *)
    Sys.remove w.path
      
  let update w cs = (* raise Sys_error *)
    let ch = open_out w.path in
    output_urllist ch (get_urls cs);
    close_out ch
	
  let move w new_url =
    Sys.rename w.path new_url;
    w.path <- new_url
end
